Package: G4SNVHunter
Type: Package
Title: Evaluating SNV-Induced Disruption of G-Quadruplex Structures
Version: 1.3.0
Authors@R: person(given = "Rongxin", 
                  family = "Zhang", 
                  email = "rongxinzhang@outlook.com", 
                  role = c("cre", "aut"),
                  comment = c('ORCID' = "0000-0002-1643-1185"))
Description: 
    G-quadruplexes (G4s) are unique nucleic acid secondary structures 
    predominantly found in guanine-rich regions and have been shown to be 
    involved in various biological regulatory processes. 
    G4SNVHunter is an R package designed to rapidly identify genomic sequences 
    with G4-forming propensity and to accurately screen user-provided 
    single nucleotide variants—as well as other small-scale variants such as 
    indels and MNVs—for their potential to destabilize these structures.
    This allows researchers to then screen these critical variants for deeper 
    study, digging into how they might influence biological functions—think 
    gene regulation, for instance—by impairing G4 formation propensity.
Encoding: UTF-8
LazyData: false
Depends: R (>= 4.3.0)
Imports: Biostrings, stats, GenomicRanges, IRanges, Rcpp, RcppRoll,
        data.table, Seqinfo, S4Vectors, ggplot2, cowplot, progress,
        ggseqlogo, viridis, ggpointdensity, tools,
        SummarizedExperiment, VariantAnnotation, dplyr, openxlsx,
        tidyr, magrittr, ggdensity
LinkingTo: Rcpp
RoxygenNote: 7.3.2
License: MIT + file LICENSE
Suggests: knitr, BiocStyle, rmarkdown, BiocManager,
        BSgenome.Hsapiens.UCSC.hg19, DT, rtracklayer, testthat (>=
        3.0.0), RBGL
VignetteBuilder: knitr
biocViews: Epigenetics, SNP
News: NEWS.md
URL: https://github.com/rongxinzh/G4SNVHunter
BugReports: https://github.com/rongxinzh/G4SNVHunter/issues
Config/testthat/edition: 3
git_url: https://git.bioconductor.org/packages/G4SNVHunter
git_branch: devel
git_last_commit: ec8c408
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-11-04
NeedsCompilation: yes
Packaged: 2025-11-04 22:37:31 UTC; biocbuild
Author: Rongxin Zhang [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-1643-1185>)
Maintainer: Rongxin Zhang <rongxinzhang@outlook.com>

# Generated by roxygen2: do not edit by hand

export(G4HunterDetect)
export(G4HunterScore)
export(G4VarImpact)
export(SNVImpactG4)
export(checkSNV)
export(exportG4)
export(exportMutG4)
export(filterSNVImpact)
export(filterVarImpact)
export(loadSequence)
export(loadVariant)
export(plotG4Info)
export(plotImpactSeq)
export(plotImpactedG4)
export(plotSNVImpact)
export(plotVarImpact)
import(GenomicRanges)
import(IRanges)
import(Rcpp)
import(RcppRoll)
import(S4Vectors)
import(Seqinfo)
import(cowplot)
import(ggplot2)
import(ggpointdensity)
import(ggseqlogo)
import(progress)
import(viridis)
importFrom(Biostrings,BString)
importFrom(Biostrings,DNAString)
importFrom(Biostrings,DNAStringSet)
importFrom(Biostrings,readDNAStringSet)
importFrom(Biostrings,replaceAt)
importFrom(Biostrings,reverseComplement)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,GRangesList)
importFrom(GenomicRanges,strand)
importFrom(IRanges,IRanges)
importFrom(IRanges,findOverlaps)
importFrom(IRanges,start)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,metadata)
importFrom(S4Vectors,queryHits)
importFrom(S4Vectors,subjectHits)
importFrom(Seqinfo,seqlevels)
importFrom(SummarizedExperiment,rowRanges)
importFrom(VariantAnnotation,isDelins)
importFrom(VariantAnnotation,isIndel)
importFrom(VariantAnnotation,isSNV)
importFrom(VariantAnnotation,isSubstitution)
importFrom(VariantAnnotation,isTransition)
importFrom(VariantAnnotation,readVcf)
importFrom(cowplot,draw_label)
importFrom(cowplot,ggdraw)
importFrom(cowplot,plot_grid)
importFrom(data.table,":=")
importFrom(data.table,.SD)
importFrom(data.table,CJ)
importFrom(data.table,fread)
importFrom(data.table,fwrite)
importFrom(dplyr,mutate)
importFrom(ggdensity,geom_hdr)
importFrom(ggplot2,scale_color_viridis_c)
importFrom(ggpointdensity,geom_pointdensity)
importFrom(grid,arrow)
importFrom(grid,unit)
importFrom(magrittr,"%>%")
importFrom(openxlsx,addWorksheet)
importFrom(openxlsx,createWorkbook)
importFrom(openxlsx,saveWorkbook)
importFrom(openxlsx,writeData)
importFrom(stats,setNames)
importFrom(tidyr,unnest)
importFrom(tools,file_ext)
useDynLib(G4SNVHunter)

# G4SNVHunter <img src="./vignettes/images/logo.png" align = "right" width = "150" />

G4SNVHunter is an R package leveraging the G4Hunter algorithm to systematically 
identify single nucleotide variants (SNVs), 
along with other small-scale variants such as indels and MNVs,
that have the potential to disrupt G-quadruplex (G4) formation propensity.


# Installation
### Option 1: Install from GitHub

You can install the package directly from GitHub,
```r
# install.packages("devtools")
devtools::install_github("rongxinzh/G4SNVHunter")
```

To run the sample code in our [vignette](
https://rongxinzh.github.io/G4SNVHunter/G4SNVHunter.html
), set the `dependencies` parameter to `TRUE`,
```r
# install.packages("devtools")
devtools::install_github("rongxinzh/G4SNVHunter", dependencies = TRUE)
```

**NOTE**

* Your R version must be &#8805; 4.3.
* If you are using Windows, please install 
[Rtools](https://ohdsi.github.io/Hades/rSetup.html#Installing_RTools) 
before proceeding.

### Option 2: Install from Bioconductor

The package is also available on Bioconductor as a development (devel) version. 
To install it, follow these steps,
```r
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install(version = 'devel')

BiocManager::install("G4SNVHunter")
```

If you want to run the sample code in our [vignette](
https://rongxinzh.github.io/G4SNVHunter/G4SNVHunter.html
), 
set `dependencies` to `TRUE`,

```r
BiocManager::install("G4SNVHunter", dependencies = TRUE)
```

**NOTE**

* The Bioconductor installation (`version = 'devel'`) requires the latest 
version of R.
* Make sure to initialize Bioconductor devel using 
`BiocManager::install(version = 'devel')` before installing the package.

# Quick Start

First, you need to library the G4SNVHunter package,

```r
library(G4SNVHunter)
```

In this example, we need to use the human genome sequence of chromosome 16 
(hg19 version).

```r
if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
    BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
}
library(BSgenome.Hsapiens.UCSC.hg19)

# Load sequence for chromosome 16 (hg19)
hg19 <- BSgenome.Hsapiens.UCSC.hg19

chr16_seq <- DNAStringSet(hg19$chr16)

# Chromosome names are needed for analysis
names(chr16_seq) <- "chr16"

```

We have prepared some sample variant data, which can be loaded as,

```r
vcf_file <- system.file("extdata", 
                        "example_variants_chr16.vcf", 
                        package = "G4SNVHunter")
variants <- loadVariant(vcf_file, file_type = "vcf")
seqlevels(variants) <- paste0("chr", seqlevels(variants))
```

Before assessing the impact of SNV on G4, we need to predict the G4-prone 
regions,

```r
chr16_G4 <- G4HunterDetect(chr16_seq)
```

Then, the impact of variants on the formation of G4 structures 
can be evaluated as,

```r
result <- G4VarImpact(G4 = chr16_G4, 
                      variants = variants, 
                      ref_col = "REF",
                      alt_col = "ALT")
```

Since not all variants can significantly impact the formation of G4s, we need 
to filter the output. While the filtering parameters can be set flexibly, 
a set of recommended parameters is:

```r
filtered_var_eff <- filterVarImpact(result, 
                                    raw_score_threshold = 1.5,
                                    mut_score_threshold = 1.2)

print(filtered_var_eff)
```

Generally, G4 structures with an absolute G4Hunter score above 1.5 are 
considered to have a high confidence in forming stable structures. In contrast, 
those with scores below 1.2, or more conservatively, below 1.0, are typically 
less likely to form stable structures.

# Vignette

For full use of our package, please refer to our [vignette](
https://rongxinzh.github.io/G4SNVHunter/G4SNVHunter.html)
(highly recommended). 

# Bug Reports

If you encounter any issues, have questions, or would like to make suggestions, 
feel free to report them at our [bug tracker](
https://github.com/rongxinzh/G4SNVHunter/issues).

# Contact

For additional inquiries, contact us at: 
Email: rongxinzhang@outlook.com

#' Plot the Impact of SNVs on G4Hunter Scores
#'
#' This function is deprecated and will be removed in a future version.
#'
#' This function generates two plots for visualizing the impact of SNVs on
#' G4 formation:
#' 1. A scatter plot with density shading comparing the original G4Hunter score
#' and the mutant G4Hunter score.
#' 2. A density plot showing the distribution of score changes between the
#' original and mutant G4 sequences.
#'
#' @param gr A \code{GRanges} object containing the G4Hunter scores and
#' associated metadata. The object should include the following columns:
#' \itemize{
#'   \item \code{G4.info.score}: Numeric vector of the original G4Hunter
#'   scores.
#'   \item \code{mut.score}: Numeric vector of the G4Hunter scores after
#'   mutation.
#'   \item \code{score.diff}: Numeric vector of the differences between the
#'   original and mutant G4Hunter scores.
#' }
#'
#' @param p_colors A vector of four colors used for plotting.
#'
#' @return A combined plot (using \code{plot_grid}) containing two subplots:
#'         - Density scatter plot comparing original vs mutant G4Hunter scores.
#'         - Density plot of the G4Hunter score differences.
#'
#' @seealso \code{\link{SNVImpactG4}} for evaluating the impact of SNVs on G4
#' formation, and \code{\link{filterSNVImpact}} for filtering G4s that are
#' significantly affected by SNVs.
#'
#' @import ggplot2
#' @import ggpointdensity
#' @import viridis
#' @import cowplot
#' @export
#'
#' @examples
#'
#' if (!requireNamespace("BiocManager", quietly = TRUE)) {
#'   install.packages("BiocManager")
#' }
#'
#' if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
#'   BiocManager::install("GenomicRanges")
#' }
#'
#' if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
#'   BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
#' }
#'
#' library(GenomicRanges)
#' library(BSgenome.Hsapiens.UCSC.hg19)
#'
#' hg19 <- BSgenome.Hsapiens.UCSC.hg19
#' chr21_seq <- DNAStringSet(hg19$chr21)
#' # Chromosome name is needed
#' names(chr21_seq) <- "chr21"
#'
#' G4 <- G4HunterDetect(chr21_seq)
#'
#' # Load SNVs
#' data(snp_gr)
#'
#' res_snp <- SNVImpactG4(G4, snp_gr, alt_col = "alt")
#' plotSNVImpact(res_snp)
#' @section Deprecated:
#' This function is no longer supported.
#' Use \code{\link{plotVarImpact}} instead.
plotSNVImpact <- function(gr,
                          p_colors = c("#b22d2d", "#6ca4d6",
                                       "#2d69b0", "#1f77b4")) {

  .Deprecated("plotVarImpact")

  if (length(p_colors) != 4) {
    stop("'p_colors' must be a vector of exactly four colors.")
  }

  df <- as.data.frame(gr)

  max_range <- max(abs(df$G4.info.score), abs(df$mut.score))
  min_range <- min(abs(df$G4.info.score), abs(df$mut.score))

  p1 <- ggplot(data = df, mapping = aes(x = abs(.data$G4.info.score),
                                        y = abs(.data$mut.score))) +
    geom_pointdensity() +
    scale_color_viridis() +
    geom_abline(intercept = 0, slope = 1, linewidth = 1,
                linetype = "dashed", color = p_colors[1]) +
    scale_x_continuous(limits = c(min_range, max_range)) +
    scale_y_continuous(limits = c(min_range, max_range)) +
    labs(
      title = "Density Plot of G4Hunter Score",
      x = "Original G4 Score",
      y = "Mutant G4 Score"
    ) +
    coord_fixed() +
    theme_bw()

  p2 <- ggplot(df, aes(x = .data$score.diff)) +
    geom_density(fill = p_colors[2], alpha = 0.5, color = p_colors[3]) +
    geom_vline(xintercept = 0, color = p_colors[4],
               linewidth = 1, linetype = "dashed") +
    labs(
      title = "Distribution of G4Hunter Score Changes",
      x = "Score Changes",
      y = "Density Estimate"
    ) +
    theme_bw()

  combined_plot <- plot_grid(p1, p2, ncol = 2, labels = "AUTO", align = 'v')

  title <- ggdraw() +
    draw_label(
      "Changes in G4Hunter Score",
      fontface = 'bold',
      x = 0.5, hjust = 0.5
    )

  final_plot <- plot_grid(title, combined_plot,
                          ncol = 1, rel_heights = c(0.1, 1))

  return(final_plot)
}

#' Visualize the variants in G4 sequence
#'
#' This function is deprecated and will be removed in a future version.
#'
#' This function plot sequence logos to visualize sequence variants caused by
#' SNVs or SNPs, with the location of the variants highlighted by rectangles
#' and arrows.
#'
#' @param filtered_gr A \code{GRanges} object containing sequence data and
#' G4Hunter scores. The object must have metadata columns named
#' \code{G4.info.score}, \code{mut.score}, \code{G4.info.sequence}, and
#' \code{mut.G4.seq}.
#' @param ncol An integer specifying the number of columns in the output plot
#' grid. Default is 1.
#'
#' @return A plot that displays the grid of sequence logos, showing the
#' differences between the original and mutated sequences.
#'
#' @seealso \code{\link{SNVImpactG4}} for evaluating the impact of SNVs on G4
#' formation, and \code{\link{filterSNVImpact}} for filtering G4s that are
#' significantly affected by SNVs.
#'
#' @import ggplot2
#' @import ggseqlogo
#' @import cowplot
#' @export
#'
#' @examples
#'
#' if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
#'   BiocManager::install("GenomicRanges")
#' }
#'
#' library(GenomicRanges)
#'
#' seq <- data.frame(chr = c("seq1", "seq2"),
#'                   seq = c("ATTTGGGAGGGAGGGAGGGATGATGAAAATTTTATTTATTTTATTTA",
#'                           "TTTATACTATTCCCTTACCCTCCCATCCCCATACGGCATCTAGATC"))
#' seq_gr <- loadSequence(seq)
#' G4 <- G4HunterDetect(seq_gr)
#'
#' snv_gr <- GRanges(seqnames = c("seq1", "seq2"),
#'                   ranges = IRanges(start = c(18, 23), width = 1),
#'                   ref = c("G", "C"),
#'                   alt = c("C", "G"))
#'
#' effect <- SNVImpactG4(G4, snv_gr, alt_col = "alt")
#' plotImpactSeq(effect, ncol = 2)
#' @section Deprecated:
#' This function is no longer supported.
#' Use \code{\link{plotImpactedG4}} instead.
plotImpactSeq <- function(filtered_gr,
                          ncol = 1) {

  .Deprecated("plotImpactedG4")

  if (!inherits(filtered_gr, "GRanges")) {
    stop("'filtered_gr' must be a GRanges object")
  }

  if (length(filtered_gr) == 0) {
    stop("Input GRanges object 'filtered_gr' is empty.")
  }

  if (!is.numeric(ncol) || ncol <= 0 || floor(ncol) != ncol) {
    stop("'ncol' must be a positive integer.")
  }

  fig_list <- lapply(seq_along(filtered_gr), function(i) {

    original_score <- filtered_gr$G4.info.score[i]
    mutated_score <- filtered_gr$mut.score[i]

    label <- paste0("Original G4Hunter score: ", original_score,
                    " \nMutated G4Hunter score: ", mutated_score)

    original_seq <- filtered_gr$G4.info.sequence[i]
    mutated_seq <- filtered_gr$mut.G4.seq[i]

    original_seq_char <- strsplit(original_seq, "")[[1]]
    mutated_seq_char <- strsplit(mutated_seq, "")[[1]]

    diff_positions <- which(original_seq_char != mutated_seq_char)

    aln_ori_mat <- data.frame(char = original_seq_char,
                              group = rep(paste0("Original\n(Score:",
                                                 original_score, ")"),
                                          length(original_seq_char)),
                              position = seq_along(original_seq_char),
                              mut = rep('no', length(original_seq_char)))

    aln_mut_mat <- data.frame(char = mutated_seq_char,
                              group = rep(paste0("Mutated\n(Score:",
                                                 mutated_score, ")"),
                                          length(mutated_seq_char)),
                              position = seq_along(mutated_seq_char),
                              mut = rep('no', length(mutated_seq_char)))

    aln_mut_mat$mut[diff_positions] <- "yes"

    aln_mat <- rbind(aln_ori_mat, aln_mut_mat)

    p1 <- ggseqlogo(setNames(list(c(original_seq, mutated_seq)), label),
                    method = 'prob') +
       annotate('rect', xmin = diff_positions - 0.5,
                xmax = diff_positions + 0.5, ymin = -0.05, ymax = 1.05,
                alpha = .2, col='black', fill='yellow') +
       geom_segment(aes(x = diff_positions, xend = diff_positions, y = -0.2,
                        yend = -0.1),
                    color = '#b22d2d', size = 0.8,
                    arrow = arrow(type = "closed",
                                  length = unit(0.2, "inches"))) +
       theme(axis.text.x = element_blank())

    p2 <- ggplot(aln_mat, aes(.data$position, .data$group)) +
      geom_text(aes(label = .data$char, color = .data$mut, size = .data$mut)) +
      scale_x_continuous(breaks = seq_len(10), expand = c(0.105, 0)) +
      xlab('') + ylab('') +
      scale_color_manual(values = c('black', '#b22d2d')) +
      scale_size_manual(values = c(5, 6)) +
      theme_logo() +
      theme(legend.position = 'none', axis.text.x = element_blank())

    return(plot_grid(p1, p2, ncol = 1, align = 'v'))
  })

  combined_plot <- plot_grid(plotlist = fig_list, labels = "AUTO",
                             ncol = ncol, align = 'v')

  return(combined_plot)
}

#' Export G4s and Associated Variant Information to File
#'
#' This function exports a \code{GRanges} object containing predicted G4
#' regions affected by variants to a TXT, CSV, or XLSX file. The output
#' includes both the original and mutated G-rich sequences, along with
#' detailed information about the associated variants.
#'
#' @param mut_G4 A \code{GRanges} object returned by \code{filterVarImpact}.
#' Both the \code{G4.info.sequence} and the \code{mutated.G4.seq} columns
#' must be included.
#' @param filename A \code{character} string specifying the output file path.
#' The file extension must be one of:
#' \code{.txt}, \code{.csv}, or \code{.xlsx}.
#' @param include_metadata A \code{logical} value.
#' Whether to include global metadata (e.g., G4 prediction parameters) in the
#' output.
#' Default is \code{TRUE}.
#' @param revcomp_G4_seq A \code{logical} value. Whether to
#' reverse-complement G4 sequences on the antisense (negative) strand.
#' Default is \code{TRUE}.
#' @param revcomp_mutG4_seq A \code{logical} value. Whether to
#' reverse-complement mutated G4 sequences on the antisense (negative) strand.
#' Default is \code{TRUE}.
#'
#' @return Invisibly returns a \code{data.frame} object.
#'
#' @export
#'
#' @importFrom Biostrings reverseComplement DNAStringSet
#' @importFrom tools file_ext
#' @importFrom data.table fwrite
#' @importFrom openxlsx createWorkbook addWorksheet writeData saveWorkbook
#' @importFrom S4Vectors metadata
#'
#' @examples
#' library(GenomicRanges)
#' fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
#' seq <- loadSequence(seq_path = fa_path)
#' # Predict G4s
#' G4_detected <- G4HunterDetect(seq)
#' # create mut granges object
#' mut <- data.frame(
#'   chr = c("seq1", "seq5"),
#'   pos = c(81, 11),
#'   ref = c("GGGTAGGG", "A"),
#'   alt = c("G", "AGGGGGGGGGGGGGGGG")
#' )
#'
#' mut <- GRanges(
#'   seqnames = mut$chr,
#'   ranges = IRanges(start = mut$pos, end = mut$pos),
#'   strand = "*",
#'   ref = mut$ref,
#'   alt = mut$alt
#' )
#'
#' mut_G4 <- G4VarImpact(G4_detected, mut, ref_col = "ref", alt_col = "alt")
#' filtered_mut_G4 <- filterVarImpact(mut_G4, score_diff_threshold = -0.2)
#' exportMutG4(filtered_mut_G4, "./result_mut.txt")
#' exportMutG4(filtered_mut_G4, "./result_mut.xlsx", include_metadata = FALSE)
#' exportMutG4(filtered_mut_G4, "./result_mut.csv", revcomp_mutG4_seq = FALSE)
#'
#' # remove all exported files
#' unlink("./result_mut.txt")
#' unlink("./result_mut.xlsx")
#' unlink("./result_mut.csv")
exportMutG4 <- function(mut_G4 = NULL,
                        filename = NULL,
                        include_metadata = TRUE,
                        revcomp_G4_seq = TRUE,
                        revcomp_mutG4_seq = TRUE) {

  validateExportMutG4Inputs(mut_G4,
                            filename,
                            include_metadata,
                            revcomp_G4_seq,
                            revcomp_mutG4_seq)

  df <- as.data.frame(mut_G4, stringsAsFactors = FALSE)

  if (revcomp_G4_seq) {
    if (!"G4.info.sequence" %in% names(df)) {
      stop("The 'mut_G4' object must contain a 'G4.info.sequence' column.")
    }

    df$G_rich_sequence <- df$G4.info.sequence

    antisense_idx <- which(df$strand == "-")
    if (length(antisense_idx) > 0) {
      df$G_rich_sequence[antisense_idx] <- as.character(
        Biostrings::reverseComplement(
          Biostrings::DNAStringSet(df$G4.info.sequence[antisense_idx])
        )
      )
    } else {
      message("All G4s are on the positive strand; ",
              "reverse complementation for G4 sequences was not applied.")
    }
  }

  if (revcomp_mutG4_seq) {
    if (!"mutated.G4.seq" %in% names(df)) {
      stop("The 'mut_G4' object must contain a 'mutated.G4.seq' column.")
    }

    df$G_rich_mut_sequence <- df$mutated.G4.seq

    antisense_idx <- which(df$strand == "-")
    if (length(antisense_idx) > 0) {
      df$G_rich_mut_sequence[antisense_idx] <- as.character(
        Biostrings::reverseComplement(
          Biostrings::DNAStringSet(df$mutated.G4.seq[antisense_idx])
        )
      )
    } else {
      message("All G4s are on the positive strand; ",
              "reverse complementation for mutated G4 ",
              "sequences was not applied.")
    }
  }

  if (include_metadata) {
    meta_cols <- names(metadata(mut_G4))
    if (length(meta_cols) > 0) {
      metadata_str <- paste(meta_cols,
                            metadata(mut_G4)[meta_cols],
                            sep = ":",
                            collapse = ";")
      metadata_str <- paste0("#", metadata_str)
    } else {
      message("[Note] No metadata columns available in 'mut_G4'.")
    }
  }

  ext <- tolower(file_ext(filename))

  if (ext %in% c("txt", "csv")) {

    if (exists("metadata_str")) {
      writeLines(metadata_str, filename)
      append_mode <- TRUE
    } else {
      append_mode <- FALSE
    }

    data.table::fwrite(
      df,
      file = filename,
      sep = ifelse(ext == "csv", ",", "\t"),
      quote = FALSE,
      col.names = TRUE,
      append = append_mode
    )
  } else {

    wb <- openxlsx::createWorkbook()
    openxlsx::addWorksheet(wb, sheetName = "G4SNVHunter_Output")

    if (exists("metadata_str")) {
      openxlsx::writeData(wb, sheet = "G4SNVHunter_Output",
                          metadata_str, startRow = 1, colNames = FALSE)
      data_start_row <- length(metadata_str) + 1
    } else {
      data_start_row <- 1
    }

    openxlsx::writeData(
      wb,
      sheet = "G4SNVHunter_Output",
      df,
      startRow = data_start_row,
      colNames = TRUE
    )

    openxlsx::saveWorkbook(wb, filename, overwrite = TRUE)
  }

  invisible(df)
}

#' validateExportMutG4Inputs
#'
#' This function validates the input arguments for the \code{exportMutG4} function.
#'
#' @param mut_G4 A \code{GRanges} object returned by \code{filterVarImpact}.
#' Both the \code{G4.info.sequence} and the \code{mutated.G4.seq} columns
#' must be included.
#' @param filename A \code{character} string specifying the output file path.
#' The file extension must be one of:
#' \code{.txt}, \code{.csv}, or \code{.xlsx}.
#' @param include_metadata A \code{logical} value.
#' Whether to include global metadata (e.g., G4 prediction parameters) in the
#' output.
#' Default is \code{TRUE}.
#' @param revcomp_G4_seq A \code{logical} value. Whether to
#' reverse-complement G4 sequences on the antisense (negative) strand.
#' Default is \code{TRUE}.
#' @param revcomp_mutG4_seq A \code{logical} value. Whether to
#' reverse-complement mutated G4 sequences on the antisense (negative) strand.
#' Default is \code{TRUE}.
#'
#' @return NULL
#' @importFrom tools file_ext
#' @noRd
validateExportMutG4Inputs <- function(mut_G4,
                                      filename,
                                      include_metadata,
                                      revcomp_G4_seq,
                                      revcomp_mutG4_seq) {

  if (!inherits(mut_G4, "GRanges")) {
    stop("'mut_G4' must be a GRanges object.")
  }

  if (length(mut_G4) == 0) {
    stop("'mut_G4' cannot be NULL. Please check your input.")
  }

  if (!is.character(filename) || length(filename) != 1) {
    stop("'filename' should be a single character string.")
  }

  valid_exts <- c("txt", "csv", "xlsx")
  ext <- tolower(file_ext(filename))
  if (!ext %in% valid_exts) {
    stop(sprintf("File extension must be one of: %s",
                 paste(valid_exts, collapse = ", ")))
  }

  if (!is.logical(include_metadata) || length(include_metadata) != 1) {
    stop("'include_metadata' must be a single logical value")
  }

  if (!is.logical(revcomp_G4_seq) || length(revcomp_G4_seq) != 1) {
    stop("'revcomp_G4_seq' must be a single logical value")
  }

  if (!is.logical(revcomp_mutG4_seq) || length(revcomp_mutG4_seq) != 1) {
    stop("'revcomp_mutG4_seq' must be a single logical value")
  }

}

fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
seq <- loadSequence(seq_path = fa_path)
G4_detected <- G4HunterDetect(seq)

mut <- GRanges(
  seqnames = c("seq1", "seq5"),
  ranges = IRanges(start = c(81, 11), end = c(88, 11)),
  strand = "*",
  ref = c("GGGTAGGG", "A"),
  alt = c("G", "AGGGGGGGGGGGGGGGG")
  )

mut_G4 <- G4VarImpact(G4_detected, mut, ref_col = "ref", alt_col = "alt")
filtered_mut_G4 <- filterVarImpact(mut_G4, score_diff_threshold = -0.3)

test_that("exportMutG4 errors on invalid mut_G4 input", {
  expect_error(exportMutG4(NULL, "test.txt"),
               "'mut_G4' must be a GRanges object.")
  expect_error(exportMutG4(GRanges(), "test.txt"),
               "'mut_G4' cannot be NULL. Please check your input.")
})

test_that("exportMutG4 errors on invalid filename input", {
  expect_error(exportMutG4(filtered_mut_G4, NULL),
               "'filename' should be a single character string")
  expect_error(exportMutG4(filtered_mut_G4, 123),
               "'filename' should be a single character string")
  expect_error(exportMutG4(filtered_mut_G4, c("a.txt", "b.txt")),
               "'filename' should be a single character string")
  expect_error(exportMutG4(filtered_mut_G4, "test.pdf"),
               "File extension must be one of: txt, csv, xlsx")
})

test_that("exportMutG4 errors when required columns are missing", {
  missing_gr <- filtered_mut_G4
  mcols(missing_gr)$G4.info.sequence <- NULL
  expect_error(
    exportMutG4(missing_gr, "test.txt", revcomp_G4_seq = TRUE),
    "The 'mut_G4' object must contain a 'G4.info.sequence' column"
  )

  missing_gr <- filtered_mut_G4
  mcols(missing_gr)$mutated.G4.seq <- NULL
  expect_error(
    exportMutG4(missing_gr, "test.txt", revcomp_mutG4_seq = TRUE),
    "The 'mut_G4' object must contain a 'mutated.G4.seq' column"
  )
})

test_that("exportMutG4 correctly exports TXT files", {
  txt_file <- tempfile(fileext = ".txt")
  expect_s3_class(exportMutG4(filtered_mut_G4, txt_file), "data.frame")
  expect_true(file.exists(txt_file))

  lines <- readLines(txt_file)
  expect_true(grepl("^#threshold:1.5", lines[1]))

  unlink(txt_file)
})

test_that("exportMutG4 correctly exports CSV files", {
  csv_file <- tempfile(fileext = ".csv")
  expect_s3_class(exportMutG4(filtered_mut_G4, csv_file,
                              include_metadata = FALSE), "data.frame")
  expect_true(file.exists(csv_file))

  df <- read.csv(csv_file)
  expect_true("G_rich_sequence" %in% colnames(df))

  unlink(csv_file)
})

test_that("exportMutG4 correctly exports XLSX files", {
  xlsx_file <- tempfile(fileext = ".xlsx")
  expect_s3_class(exportMutG4(filtered_mut_G4, xlsx_file,
                              revcomp_G4_seq = FALSE), "data.frame")
  expect_true(file.exists(xlsx_file))

  wb <- openxlsx::loadWorkbook(xlsx_file)
  expect_true("G4SNVHunter_Output" %in% names(wb))

  unlink(xlsx_file)
})

#' Plot the Impact of Variants on G4 Formation
#'
#' This function generates two panels to visualize the impact of variants on
#' G4 formation:
#' \enumerate{
#'   \item A 2D density plot of absolute maximum G4Hunter scores before and
#'   after mutation.
#'   \item A density plot showing the distribution of score differences.
#' }
#'
#' @param gr A \code{GRanges} object containing G4Hunter scores and
#' associated metadata.
#' Required metadata columns include:
#' \itemize{
#'   \item \code{G4.info.max_score}: Numeric vector of
#'   original maximum G4Hunter scores.
#'   \item \code{mutated.max_score}: Numeric vector of
#'   maximum G4Hunter scores after mutation.
#'   \item \code{score.diff}: Numeric vector of score differences.
#' }
#'
#' @param p_colors A character vector of five colors used for plotting.
#' The vector must include:
#' \describe{
#'   \item{\code{"fill_2d"}}{Fill Color for the 2D density plot.}
#'   \item{\code{"diagonal"}}{Color for the diagonal line in the 2D
#'   density plot.}
#'   \item{\code{"density_fill"}}{Fill color for the density plot.}
#'   \item{\code{"density_line"}}{Line color for the density plot.}
#'   \item{\code{"vline"}}{Color for the vertical line at zero
#'   in the density plot.}
#' }
#'
#' @return A \code{ggplot} object combining two subplots:
#'   \enumerate{
#'     \item 2D density plot of the absolute maximum G4Hunter scores
#'     before and after mutation.
#'     \item Density plot of the maximum G4Hunter score differences.
#'   }
#'
#' @seealso \code{\link{G4VarImpact}} for computing the effects of variants
#' on G4 formation,
#' and \code{\link{filterVarImpact}} for screening G4s whose formation
#' propensity may be impaired by variants.
#'
#' @import ggplot2
#' @importFrom ggpointdensity geom_pointdensity
#' @importFrom cowplot plot_grid ggdraw draw_label
#' @importFrom ggplot2 scale_color_viridis_c
#' @importFrom ggdensity geom_hdr
#' @export
#'
#' @examples
#'
#' if (!requireNamespace("BiocManager", quietly = TRUE)) {
#'   install.packages("BiocManager")
#' }
#'
#' if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
#'   BiocManager::install("GenomicRanges")
#' }
#'
#' if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
#'   BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
#' }
#'
#' library(GenomicRanges)
#' library(BSgenome.Hsapiens.UCSC.hg19)
#'
#' hg19 <- BSgenome.Hsapiens.UCSC.hg19
#' chr21_seq <- DNAStringSet(hg19$chr21)
#' # Chromosome name is needed
#' names(chr21_seq) <- "chr21"
#'
#' G4 <- G4HunterDetect(chr21_seq)
#'
#' # Load SNVs
#' data(snp_gr)
#'
#' res_snp <- G4VarImpact(G4, snp_gr, ref_col = "ref", alt_col = "alt")
#' plotVarImpact(res_snp)
plotVarImpact <- function(gr,
                          p_colors = c(
                            fill_2d = "#376597",
                            diagonal = "#b22d2d",
                            density_fill = "#6ca4d6",
                            density_line = "#2d69b0",
                            vline = "#1f77b4"
                          )) {
  required_names <- c("fill_2d", "diagonal", "density_fill",
                      "density_line", "vline")
  if (!all(required_names %in% names(p_colors))) {
    stop("'p_colors' must be a named vector with elements: ",
         paste(required_names, collapse = ", "))
  }

  df <- as.data.frame(gr)

  max_range <- max(abs(df$G4.info.max_score), abs(df$mutated.max_score))
  min_range <- min(abs(df$G4.info.max_score), abs(df$mutated.max_score))

  p1 <- ggplot(df, aes(x = abs(.data$G4.info.max_score),
                       y = abs(.data$mutated.max_score))) +
    geom_hdr(fill = p_colors["fill_2d"]) +
    geom_abline(intercept = 0, slope = 1, linewidth = 0.8,
                linetype = "dashed", color = p_colors["diagonal"]) +
    scale_x_continuous(limits = c(min_range, max_range), expand = c(0, 0)) +
    scale_y_continuous(limits = c(min_range, max_range), expand = c(0, 0)) +
    labs(
      title = "2D Density Plot of\nMax G4Hunter Scores",
      x = "Original Score (absolute)\n",
      y = "Mutated Score (absolute)"
    ) +
    coord_fixed() +
    theme_bw(base_size = 12) +
    theme(
      plot.title = element_text(face = "bold", hjust = 0.5),
      axis.title = element_text(face = "bold")
    )

  p2 <- ggplot(df, aes(x = .data$score.diff)) +
    geom_density(fill = p_colors["density_fill"],
                 color = p_colors["density_line"],
                 alpha = 0.5, linewidth = 0.8) +
    geom_vline(xintercept = 0, color = p_colors["vline"],
               linewidth = 0.8, linetype = "dashed") +
    labs(
      title = "Distribution of Changes in\nMax G4Hunter Scores",
      x = "Change in Max G4Hunter Score\nabs(Mutated) - abs(Original)",
      y = "Density"
    ) +
    theme_bw(base_size = 12) +
    theme(
      plot.title = element_text(face = "bold", hjust = 0.5),
      axis.title = element_text(face = "bold")
    )

  combined_plot <- plot_grid(p1, p2, ncol = 2,
                             labels = c("a", "b"),
                             label_size = 14)

  title <- ggdraw() +
    draw_label(
      "Mutational Impact on G4 Formation",
      fontface = 'bold',
      size = 14,
      x = 0.5, hjust = 0.5
    )

  final_plot <- plot_grid(title, combined_plot,
                          ncol = 1, rel_heights = c(0.1, 1))

  return(final_plot)
}

#' Sequence Logo of the G4 Sequence and Its Mutant Counterpart
#'
#' This function visualizes and compares the G4 and its mutant counterpart
#' using sequence logos.
#'
#' @param mut_G4 A \code{GRanges} object containing the mutant G4.
#' If multiple G4s are provided, only the first is plotted.

#' @param keep_gstrand Logical. If \code{TRUE} (default), negative
#' strand (C-rich) G4 will be displayed in its complementary reversed form.
#'
#' @return A ggplot2-based object containing the G4 sequence logo.
#'
#' @import ggplot2
#' @import ggseqlogo
#' @importFrom grid arrow unit
#' @importFrom cowplot plot_grid
#' @importFrom GenomicRanges strand
#'
#' @examples
#' if (!requireNamespace("BiocManager", quietly = TRUE)) {
#'   install.packages("BiocManager")
#' }
#'
#' if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
#'   BiocManager::install("GenomicRanges")
#' }
#'
#' if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
#'   BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
#' }
#'
#' library(GenomicRanges)
#' library(BSgenome.Hsapiens.UCSC.hg19)
#'
#' hg19 <- BSgenome.Hsapiens.UCSC.hg19
#' chr21_seq <- DNAStringSet(hg19$chr21)
#' # Chromosome name is needed
#' names(chr21_seq) <- "chr21"
#'
#' G4 <- G4HunterDetect(chr21_seq)
#'
#' # Load SNVs
#' data(snp_gr)
#'
#' res_snp <- G4VarImpact(G4, snp_gr, ref_col = "ref", alt_col = "alt")
#' plotImpactedG4(res_snp[1])
#'
#' @export
plotImpactedG4 <- function(mut_G4 = NULL, keep_gstrand = TRUE) {

  if (!inherits(mut_G4, "GRanges")) {
    stop("'mut_G4' must be a GRanges object")
  }

  if (length(mut_G4) == 0) {
    stop("Input GRanges object 'mut_G4' is empty.")
  }

  if (length(mut_G4) > 1) {
    warning("Multiple records detected in the 'mut_G4' object. ",
            "Only the first record will be plotted.\n",
            "Please use a loop function to plot all records.")
  }

  mut_G4 <- mut_G4[1]

  version =
    ifelse("mutated.max_score" %in% colnames(mcols(mut_G4)),
           "new", "initial")

  if (version == "initial") {
    stop("This function requires G4SNVHunter version > 1.0.0.
         Please upgrade your package.")
  }

  original_score <- mut_G4$G4.info.max_score[1]
  mutated_score <- mut_G4$mutated.max_score[1]

  label <- paste0("Original max G4Hunter score: ", original_score,
                  "\nMutated max G4Hunter score: ", mutated_score)

  mutated_seq <- toupper(mut_G4$mutated.G4.anno.seq[1])
  mutated_seqs <- parseVarSeq(mutated_seq)

  if (as.character(strand(mut_G4)) == "-" & keep_gstrand) {
    mutated_seqs <- c(comRevBase(mutated_seqs[1]), comRevBase(mutated_seqs[2]))
  }

  original_seq_char <- strsplit(mutated_seqs[1], "")[[1]]
  mutated_seq_char <- strsplit(mutated_seqs[2], "")[[1]]

  aln_ori_mat <- data.frame(char = original_seq_char,
                            group = rep(paste0("Ref seq\n(Max score:",
                                               ifelse(keep_gstrand,
                                                      abs(original_score),
                                                      original_score), ")"),
                                        length(original_seq_char)),
                            position = seq_along(original_seq_char),
                            mut = rep('no', length(original_seq_char)))

  aln_mut_mat <- data.frame(char = mutated_seq_char,
                            group = rep(paste0("Mutated seq\n(Max score:",
                                               ifelse(keep_gstrand,
                                                      abs(mutated_score),
                                                      mutated_score), ")"),
                                        length(mutated_seq_char)),
                            position = seq_along(mutated_seq_char),
                            mut = rep('no', length(mutated_seq_char)))

  diff_positions <- which(original_seq_char != mutated_seq_char)
  aln_ori_mat$mut[diff_positions] <- "yes"
  aln_mut_mat$mut[diff_positions] <- "yes"

  aln_mat <- rbind(aln_ori_mat, aln_mut_mat)
  fig <- plotSeqlogo(mutated_seqs, label = label, aln_mat = aln_mat)

  return(fig)

}

#' parseVarSeq
#'
#' This function splits a variant sequence into original and
#' mutated sequences.
#'
#' @param variant_seq A \code{character} string representing a G4 sequence
#' with embedded mutation annotation(s)
#'
#' @return A \code{character} vector of length 2 containing the original and
#' mutated sequences.
#' @noRd
parseVarSeq <- function(variant_seq) {
  seq1 <- character(0)
  seq2 <- character(0)

  tokens <- regmatches(variant_seq,
                       gregexpr("\\[[^]]*\\]|[^[]+",
                                variant_seq))[[1]]

  for (token in tokens) {
    if (grepl("^\\[.*\\]$", token)) {
      content <- substr(token, 2, nchar(token)-1)

      if (grepl(">", content)) {
        parts <- strsplit(content, ">")[[1]]
        left <- parts[1]
        right <- parts[2]

        len_left <- nchar(left)
        len_right <- nchar(right)

        if (len_left > len_right) {
          right <-
            paste0(right, paste(rep("-", len_left - len_right), collapse=""))
        } else if (len_right > len_left) {
          left <-
            paste0(left, paste(rep("-", len_right - len_left), collapse=""))
        }

        seq1 <- c(seq1, left)
        seq2 <- c(seq2, right)
      }

    } else {
      seq1 <- c(seq1, token)
      seq2 <- c(seq2, token)
    }
  }

  result1 <- paste(seq1, collapse="")
  result2 <- paste(seq2, collapse="")

  return(c(result1, result2))
}

#' comRevBase
#'
#' This function returns the reverse complement of a nucleotide sequence.
#'
#' @param seq A \code{character} string representing a DNA or RNA sequence.
#'
#' @return A \code{character} string of the reverse complement sequence.
#'
#' @noRd
comRevBase <- function(seq) {

  if (grepl("U", seq, ignore.case = TRUE)) {
    tmp_seq <- chartr("AUGCTaugct", "UACGAuacga", seq)  # RNA
  } else {
    tmp_seq <- chartr("ATGCUatgcu", "TACGAtacga", seq)  # DNA
  }

  tmp_seq <- paste(rev(strsplit(tmp_seq, "")[[1]]), collapse = "")
  return(tmp_seq)
}

#' plotSeqlogo
#'
#' This function generates a sequence logo plot, with mutations highlighted.
#'
#' @param seqs A \code{character} vector of original and mutated G4 sequences.
#' @param label A \code{character} string of the seqplot title.
#' @param aln_mat A \code{data.frame} containing variant
#' annotation information.
#' @param alphabet A \code{character} vector of nucleotide types to be counted.
#'
#' @return A ggplot object containing the sequence logo plot.
#' @noRd
plotSeqlogo <- function(seqs,
                        label = "",
                        aln_mat = "",
                        alphabet = c("A", "T", "G", "C", "N", "U", "-")) {
  seq_mat <- do.call(rbind, strsplit(seqs, split = ""))

  ppm <- vapply(
    X = seq_len(ncol(seq_mat)),
    FUN = function(i) {
      table(factor(seq_mat[, i], levels = alphabet))
    },
    FUN.VALUE = numeric(length(alphabet))
  )
  rownames(ppm) <- alphabet
  ppm <- ppm / nrow(seq_mat)

  num_nonzero_per_col <-
    vapply(seq_len(ncol(ppm)), function(i) sum(ppm[, i] > 0), numeric(1))
  mut_cols <- which(num_nonzero_per_col > 1)

  if (length(mut_cols) > 0) {
    mut_blocks <- split(mut_cols, cumsum(c(1, diff(mut_cols) != 1)))
  } else {
    mut_blocks <- list()
  }

  p <- ggseqlogo(ppm, method = "custom",
                 seq_type = ifelse(any(grepl("U", seqs, ignore.case = TRUE)),
                                   "rna", "dna")) +
    ylab("probability") +
    labs(title = label) +
    theme(axis.text.x = element_blank(),
          plot.title = element_text(hjust = 0.5))

  for (block in mut_blocks) {
    p <- p + annotate("rect",
                      xmin = min(block) - 0.5,
                      xmax = max(block) + 0.5,
                      ymin = -0.05,
                      ymax = 1.05,
                      alpha = 0.15,
                      col='black',
                      fill = "yellow")
  }

  segment_df <- data.frame(
    x = mut_cols,
    xend = mut_cols,
    y = -0.2,
    yend = -0.1
  )

  p <- p + geom_segment(
    data = segment_df,
    aes(x = .data$x, xend = .data$xend, y = .data$y, yend = .data$yend),
    color = '#b22d2d',
    linewidth = 0.8,
    arrow = arrow(type = "closed", length = unit(0.1, "inches")))

  p2 <- ggplot(aln_mat, aes(.data$position, .data$group)) +
    geom_text(aes(label = .data$char, color = .data$mut, size = .data$mut)) +
    scale_x_continuous(breaks = seq_len(10), expand = c(0.105, 0)) +
    xlab('') + ylab('') +
    scale_color_manual(values = c('black', '#b22d2d')) +
    scale_size_manual(values = c(5, 6)) +
    theme_logo() +
    theme(legend.position = 'none', axis.text.x = element_blank())

  return(plot_grid(p, p2, ncol = 1, align = 'v'))
}

#' G4HTranslate
#'
#' Translates a DNA sequence into a numeric score vector based on G and C runs.
#'
#' @param sequence A string representing the DNA sequence.
#' @return A numeric vector with scores corresponding to \code{G} and \code{C}
#' runs in the sequence.
#' @import Rcpp
#' @noRd
#' @useDynLib G4SNVHunter
G4HTranslate <- function(sequence) {
  .Call(`_G4SNVHunter_G4HTranslate`, sequence)
}

#' findMaxVal
#'
#' Finds the maximum G4Hunter score (absolute) in a specified range of scores.
#'
#' @param start Integer, the start index of the range.
#' @param end Integer, the end index of the range.
#' @param scores Numeric vector of scores.
#' @return Numeric value representing the maximum absolute score within the
#' specified range.
#' @noRd
findMaxVal <- function(start, end, scores) {
  range_scores <- scores[start:end]
  max_value <- range_scores[which.max(abs(range_scores))]

  return(max_value)
}

#' maxScores
#'
#' Calculates the maximum G4Hunter scores in merged regions.
#'
#' @param gr A \code{GRanges} object containing the G4 ranges and their scores.
#' @param merged_gr A \code{GRanges} object with merged G4s.
#' @return Numeric vector of maximum G4Hunter scores for each merged G4.
#' @import GenomicRanges
#' @import IRanges
#' @noRd
maxScores <- function(gr, merged_gr) {

  overlaps <- findOverlaps(gr, merged_gr)

  gr_scores <- mcols(gr)$max_score

  result <- tapply(
    gr_scores[queryHits(overlaps)], subjectHits(overlaps),
    function(x) x[which.max(abs(x))]
  )

  output <- rep(NA_real_, length(merged_gr))
  output[as.integer(names(result))] <- result

  return(output)
}

#' extendAndTrimRanges
#'
#' Extends and trims genomic ranges based on sequence data and specific base
#' runs (G or C).
#'
#' @param gr A \code{GRanges} object with genomic ranges to be extended and
#' trimmed.
#' @param sequence Character string representing the sequence from which ranges
#' are derived.
#' @param include_sequences Logical value (\code{TRUE}/\code{FALSE}) indicating
#' whether to include the predicted G4 sequences in the output.
#' Default is \code{TRUE}.
#'
#' @return A \code{GRanges} object with extended and trimmed ranges.
#'
#' @import GenomicRanges
#' @import IRanges
#' @noRd
extendAndTrimRanges <- function(gr, sequence, include_sequences = FALSE) {
  seq_len <- nchar(sequence)
  sequence_vec <- strsplit(sequence, "")[[1]]

  pos_indices <- which(as.character(strand(gr)) == "+")
  neg_indices <- which(as.character(strand(gr)) == "-")

  extendLeft <- function(start_pos, base) {
    while (start_pos > 1 && sequence_vec[start_pos] == base &&
           sequence_vec[start_pos - 1] == base) {
      start_pos <- start_pos - 1
    }
    return(start_pos)
  }

  extendRight <- function(end_pos, base) {
    while (end_pos < seq_len && sequence_vec[end_pos] == base &&
           sequence_vec[end_pos + 1] == base) {
      end_pos <- end_pos + 1
    }
    return(end_pos)
  }

  trimLeft <- function(start_pos, base) {
    while (start_pos <= seq_len && sequence_vec[start_pos] != base) {
      start_pos <- start_pos + 1
    }
    return(start_pos)
  }

  trimRight <- function(end_pos, base) {
    while (end_pos > 0 && sequence_vec[end_pos] != base) {
      end_pos <- end_pos - 1
    }
    return(end_pos)
  }

  if (length(pos_indices) > 0) {
    pos_ranges <- ranges(gr)[pos_indices]

    pos_starts <- start(pos_ranges)
    pos_ends <- end(pos_ranges)

    new_pos_starts <- vapply(pos_starts, extendLeft,
                             base = "G", FUN.VALUE = numeric(1))
    new_pos_ends <- vapply(pos_ends, extendRight,
                           base = "G", FUN.VALUE = numeric(1))

    new_pos_starts <- vapply(new_pos_starts, trimLeft,
                             base = "G", FUN.VALUE = numeric(1))
    new_pos_ends <- vapply(new_pos_ends, trimRight,
                           base = "G", FUN.VALUE = numeric(1))

    extended_pos_ranges <- GRanges(
      seqnames = seqnames(gr)[pos_indices],
      ranges = IRanges(start = new_pos_starts, end = new_pos_ends),
      strand = strand(gr)[pos_indices],
      max_score = mcols(gr)$max_score[pos_indices]
    )
  } else {
    extended_pos_ranges <- GRanges()
  }

  if (length(neg_indices) > 0) {
    neg_ranges <- ranges(gr)[neg_indices]

    neg_starts <- start(neg_ranges)
    neg_ends <- end(neg_ranges)

    new_neg_starts <- vapply(neg_starts, extendLeft,
                             base = "C", FUN.VALUE = numeric(1))
    new_neg_ends <- vapply(neg_ends, extendRight,
                           base = "C", FUN.VALUE = numeric(1))

    new_neg_starts <- vapply(new_neg_starts, trimLeft,
                             base = "C", FUN.VALUE = numeric(1))
    new_neg_ends <- vapply(new_neg_ends, trimRight,
                           base = "C", FUN.VALUE = numeric(1))

    extended_neg_ranges <- GRanges(
      seqnames = seqnames(gr)[neg_indices],
      ranges = IRanges(start = new_neg_starts, end = new_neg_ends),
      strand = strand(gr)[neg_indices],
      max_score = mcols(gr)$max_score[neg_indices]
    )
  } else {
    extended_neg_ranges <- GRanges()
  }

  extended_ranges <- c(extended_pos_ranges, extended_neg_ranges)

  valid_ranges <- extended_ranges[start(extended_ranges) <=
                                    end(extended_ranges)]

  extended_sequences <- mapply(function(start, end) {
    substr(sequence, start, end)
  }, start(valid_ranges), end(valid_ranges))

  new_scores <- vapply(extended_sequences, function(seq) {
    # mean(G4HTranslate(seq))
    G4HunterScore(seq)
  }, FUN.VALUE = numeric(1))

  mcols(valid_ranges)$score <- new_scores

  if (include_sequences) {
    mcols(valid_ranges)$sequence <- extended_sequences
  }

  return(valid_ranges)
}

#' G4HunterAnalysis
#'
#' Perform G4Hunter analysis on a given DNA sequence to identify G4 regions.
#'
#' @param sequence A character string representing the DNA sequence.
#' @param chr A character string indicating the sequence identifier.
#' @param threshold A numeric value specifying the threshold for the G4Hunter
#' score (absolute value). Default is 1.5.
#' @param window_size An integer specifying the window size for G4Hunter
#' prediction. Default is 25.
#' @param include_sequences A logical value (\code{TRUE}/\code{FALSE})
#' indicating whether to include the predicted G4 sequences in the output.
#' Default is \code{TRUE}.
#' @param strands A character string (\code{"b"} for both strands and
#' \code{"p"} for positive strand only) indicating which strand to consider.
#' Default is \code{"b"}.
#'
#' @return A \code{GRanges} object with detected G4 regions and their scores.
#'
#' @import GenomicRanges
#' @import RcppRoll
#' @noRd
G4HunterAnalysis <- function(sequence, chr, threshold = 1.5, window_size = 25,
                             include_sequences, strands) {
  score <- G4HTranslate(sequence)
  running_mean_score <- roll_mean(score, window_size,
                                  fill = NA, align = "left")

  results <- list()

  if (strands == "b") {
    regions_above_threshold <- which(abs(running_mean_score) > threshold)
  } else {
    regions_above_threshold <- which(running_mean_score > threshold)
  }

  # Merge
  if (length(regions_above_threshold) > 0) {
    start_positions <- c(
      regions_above_threshold[1],
      regions_above_threshold[which(diff(regions_above_threshold) > 1) + 1]
      )
    end_positions <- c(
      regions_above_threshold[which(diff(regions_above_threshold) > 1)],
      regions_above_threshold[length(regions_above_threshold)]
      )

    max_values <- vapply(seq_along(start_positions), function(i) {
      start <- start_positions[i]
      end <- end_positions[i]
      findMaxVal(start, end, running_mean_score)
    }, numeric(1))

    gr <- GRanges(seqnames = chr,
                  ranges = IRanges(start = start_positions,
                                   end = end_positions + window_size - 1),
                  max_score = max_values,
                  strand = ifelse(max_values > 0, "+", "-"))
    merged_gr <- reduce(gr)

    max_scores_values <- maxScores(gr, merged_gr)

    merged_gr_with_scores <- GRanges(
      seqnames = seqnames(merged_gr),
      ranges = ranges(merged_gr),
      strand = strand(merged_gr),
      max_score = max_scores_values
    )

    G4_detect <- extendAndTrimRanges(merged_gr_with_scores,
                                     sequence, include_sequences)
    if (include_sequences) {
      mcols(G4_detect) <-
        mcols(G4_detect)[, c("score", "max_score", "sequence")]
    } else {
      mcols(G4_detect) <- mcols(G4_detect)[, c("score", "max_score")]
    }

    return(G4_detect)
  } else {
    return(NULL)
  }
}

#' validateG4HunterDetectInputs
#'
#'
#' This function validates the inputs for the G4Hunter analysis.
#'
#' @param sequences A \code{DNAStringSet} object containing the input sequences
#' to be analyzed.
#' @param threshold A numeric value specifying the threshold for the G4Hunter
#' score (absolute value). Default is 1.5.
#' @param window_size An integer specifying the window size for G4Hunter
#' prediction. Default is 25.
#' @param include_sequences A logical value (\code{TRUE}/\code{FALSE})
#' indicating whether to include the predicted G4 sequences in the output.
#' Default is \code{TRUE}.
#' @param strands A character string (\code{"b"} for both strands and
#' \code{"p"} for positive strand only) indicating which strand to consider.
#' Default is \code{"b"}.
#'
#' @return NULL
#'
#' @noRd
validateG4HunterDetectInputs <- function(sequences,
                                         threshold,
                                         window_size,
                                         include_sequences,
                                         strands) {

  if (!inherits(sequences, "DNAStringSet")) {
    stop("'sequences' must be a DNAStringSet object.")
  }

  if (!is.numeric(threshold) || length(threshold) != 1) {
    stop("'threshold' must be a single numeric value.")
  }

  if (threshold > 4 || threshold < 1) {
    stop("'threshold' is not reasonable. It should be between 1 and 4.")
  }

  if (!is.numeric(window_size) || length(window_size) != 1) {
    stop("'window_size' must be a single numeric value.")
  }

  if (window_size %% 1 != 0 || window_size < 10 || window_size > 50) {
    stop("'window_size' must be an integer between 10 and 50.")
  }

  if (!is.logical(include_sequences)) {
    stop("'include_sequences' must be a logical value (TRUE or FALSE).")
  }

  if (!(strands %in% c("b", "p"))) {
    stop("'strands' should be 'b' or 'p', but got '", strands, "' instead.")
  }

}

test_that("test G4HTranslate function", {

  seq1 <- "GGGTGAATGGATGCCCGCCCCCCATGGTG"
  expected_result1 <-c(3, 3, 3, 0, 1, 0, 0, 0, 2, 2, 0, 0, 1, -3, -3, -3,
                       1, -4, -4, -4, -4, -4, -4, 0, 0, 2, 2, 0, 1)

  expect_equal(G4HTranslate(seq1), expected_result1)

  seq2 <- "TTTGGGGG"
  expected_result2 <-c(0, 0, 0, 4, 4, 4, 4, 4)

  expect_equal(G4HTranslate(seq2), expected_result2)

  seq3 <- "T"
  expected_result3 <-c(0)

  expect_equal(G4HTranslate(seq3), expected_result3)

})

sequence <- DNAStringSet(c(
  "TAGGCATCGAGGGGCTGAGGGCTGGGCTGAGGGCTGGGGCATGG"
))
names(sequence) <- c("seq1")

test_that("test validateG4HunterDetectInputs function", {

  expect_null(validateG4HunterDetectInputs(sequence, 1.5, 25, TRUE, "b"))
  expect_null(validateG4HunterDetectInputs(sequence, 2, 15, FALSE, "p"))

  expect_error(validateG4HunterDetectInputs("ATGCTAG", 0.99, 25, FALSE, "b"),
               "'sequences' must be a DNAStringSet object.")
  expect_error(validateG4HunterDetectInputs(1.2, 0.99, 25, FALSE, "b"),
               "'sequences' must be a DNAStringSet object.")
  expect_error(validateG4HunterDetectInputs(sequence, 0.99, 25, FALSE, "b"),
               "'threshold' is not reasonable. It should be between 1 and 4.")
  expect_error(validateG4HunterDetectInputs(sequence, 6, 25, TRUE, "b"),
               "'threshold' is not reasonable. It should be between 1 and 4.")
  expect_error(validateG4HunterDetectInputs(sequence, 1.5, 20.2, TRUE, "b"),
               "'window_size' must be an integer between 10 and 50.")
  expect_error(validateG4HunterDetectInputs(sequence, 1.5, 25, "yes", "b"),
               paste0("'include_sequences' must be a logical value ",
               "\\(TRUE or FALSE\\)."))
  expect_error(validateG4HunterDetectInputs(sequence, 1.5, 25, TRUE, "x"),
               "'strands' should be 'b' or 'p', but got 'x' instead.")
  expect_error(validateG4HunterDetectInputs(sequence, 1.5, 125, TRUE, "b"),
               "'window_size' must be an integer between 10 and 50.")
})

#' Calculate the G4Hunter Score for a Given Sequence
#'
#' This function calculates the G4Hunter score for a given nucleotide sequence,
#' which reflects the ability of that sequence to form a G4 structure.
#'
#' @param seq A single character string representing the nucleotide sequence.
#' Must contain only the characters \code{A}, \code{T}, \code{C}, \code{G},
#' \code{U}, and \code{N}.
#' The length of the sequence should not be too short (e.g., less than 10 bp).
#'
#' @return A numeric value representing the G4Hunter score for the provided
#' sequence.
#'
#' @seealso \code{\link{G4HunterDetect}} for detecting the G4 sequences in a
#' given \code{DNAStringSet} object.
#'
#' @export
#'
#' @examples
#'
#' sequence <- "GGGTAAGGGATGGGTCGGG"
#' score <- G4HunterScore(sequence)
#' print(score)
#' # A negative value indicates that the G4 sequence
#' # is located on the reverse strand
#' sequence <- "GGGTAAGGGATGGGTCGGG"
#' score <- G4HunterScore(sequence)
#' print(score)
G4HunterScore <- function(seq = NULL) {
  if (!is.character(seq) || length(seq) != 1) {
    stop("'seq' must be a single string.")
  }

  valid_bases <- c("A", "T", "C", "G", "U", "N")
  seq_upper <- toupper(seq)
  if (any(!strsplit(seq_upper, NULL)[[1]] %in% valid_bases)) {
    stop("'seq' must only contain A, T, C, G, U, N characters.")
  }

  avg_score <- sum(G4HTranslate(seq_upper)) / length(G4HTranslate(seq_upper))

  return(signif(avg_score, 3))
}

test_that("test G4HunterScore function", {

  seq1 <- "GGGATCGGGTACGGGGACTGGGACTGGGCA"
  seq2 <- "ccctgatcccgatcgccctagcgcgagtcgactg"
  seq3 <- "gycuqgwy12ygd2yudy3wqdqwdqw"

  expect_equal(G4HunterScore(seq1), 1.57)
  expect_equal(G4HunterScore(seq2), -0.676)

  expect_error(G4HunterScore(seq3),
               "'seq' must only contain A, T, C, G, U, N characters.")

  expect_error(G4HunterScore(12345),
               "'seq' must be a single string.")

  expect_equal(G4HunterScore("gggtaagggatgggtcggg"),
               G4HunterScore("GGGTAAGGGATGGGTCGGG"))

})

#' Evaluate the Impact of Variants (SNVs, Indels, MNVs) on G4 Sequences
#'
#' This function evaluates the impact of variants (SNVs, indels, and MNVs) on
#' G4 formation.
#'
#' @param G4 A \code{GRanges} object representing the G4 regions. This object
#' must include a \code{sequence} metadata column containing the G4 sequences.
#' It should come from the output of the \code{G4HunterDetect} function with
#' \code{include_sequences = TRUE}.
#' @param variants A \code{GRanges} object representing the variants. This
#' object must include metadata columns for reference and alternative alleles.
#' @param ref_col A \code{character} string specifying the name of the column
#' in \code{variants} that contains the reference alleles.
#' Default is \code{"ref"}.
#' @param alt_col A \code{character} string specifying the name of the column
#' in \code{variants} that contains the alternative alleles.
#' Default is \code{"alt"}.
#' @param mode A \code{character} string indicating the mode of operation.
#' Set to \code{"s"} to evaluate the impact of individual variants on G4
#' regions one at a time (single-variant mode).
#' Set to \code{"m"} to assess the combined impact of multiple variants that
#' overlap the same G4 region within a sample (multi-variant Mode).
#' If using \code{"m"} mode, you must specify \code{sampleid_col}.
#' @param sampleid_col A \code{character} string specifying the name of the
#' column in \code{variants} that contains the sample IDs.
#' Required when \code{mode} is \code{"m"};
#' ignored if \code{mode} is \code{"s"}.
#'
#' @return A \code{GRanges} object with variant impact results:
#' \describe{
#'   \item{Mode "s" (single-variant mode )}{
#'     For each variant-G4 overlap:
#'     \itemize{
#'       \item Original G4 metadata (\code{G4.info.*})
#'       \item Variant information (\code{variant.info.*})
#'       \item Mutated sequence (\code{mutated.G4.seq})
#'       \item Annotated mutation sequence (\code{mutated.G4.anno.seq})
#'       \item New G4Hunter max_score (\code{mutated.max_score})
#'       \item Score difference (\code{score.diff})
#'     }
#'   }
#'   \item{Mode "m" (multi-variant mode)}{
#'     For each sample-G4 combination:
#'     \itemize{
#'       \item Original G4 metadata (\code{G4.info.*})
#'       \item Combined variant information (\code{variant.info.*})
#'       \item Mutated sequence with all variants incorporated
#'       (\code{mutated.G4.seq})
#'       \item Annotated mutation sequence (\code{mutated.G4.anno.seq})
#'       \item New G4Hunter max_score (\code{mutated.max_score})
#'       \item Score difference (\code{score.diff})
#'     }
#'   }
#' }
#'
#' @seealso \code{\link{G4HunterDetect}} for detecting the G4 sequences in a
#'                given \code{DNAStringSet} object.
#'          \code{\link{filterVarImpact}} for filtering out variants with
#'                significant impact.
#'
#' @import GenomicRanges
#' @importFrom S4Vectors queryHits subjectHits mcols
#' @importFrom IRanges start findOverlaps
#' @export
#' @examples
#'
#' if (!requireNamespace("BiocManager", quietly = TRUE)) {
#'   install.packages("BiocManager")
#' }
#'
#' if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
#'   BiocManager::install("GenomicRanges")
#' }
#'
#' if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
#'   BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
#' }
#'
#' library(GenomicRanges)
#' library(BSgenome.Hsapiens.UCSC.hg19)
#'
#' # Load sequence for chromosome 21 (hg19)
#' hg19 <- BSgenome.Hsapiens.UCSC.hg19
#' chr21_seq <- DNAStringSet(hg19$chr21)
#' # Chromosome name is needed
#' names(chr21_seq) <- "chr21"
#'
#' # Detect G4s in human chromosome 21
#' G4 <- G4HunterDetect(chr21_seq)
#'
#' # Load variants
#' data(snv_gr)
#'
#' # 's' mode; single-variant mode ('s')
#' # evaluating each variant individually.
#' res_snv_s <- G4VarImpact(G4,
#'                          snv_gr,
#'                          ref_col = "ref",
#'                          alt_col = "alt")
#' print(res_snv_s)
#'
#' # 'm' mode; multi-variant mode ('m')
#' # evaluating the combined impact of variants on G4s.
#' # Grouped by the sample IDs specified in the 'sampleid_col' column.
#' res_snv_m <- G4VarImpact(G4,
#'                          snv_gr,
#'                          ref_col = "ref",
#'                          alt_col = "alt",
#'                          mode = "m",
#'                          sampleid_col = "sampleid"
#' )
#' print(res_snv_m)
G4VarImpact <- function(G4 = NULL,
                        variants = NULL,
                        ref_col = NULL,
                        alt_col = NULL,
                        mode = "s",
                        sampleid_col = NULL) {

  validateG4VarImpactInputs(G4,
                            variants,
                            ref_col,
                            alt_col,
                            mode,
                            sampleid_col)

  message("Start processing...")

  mcols(variants)[[alt_col]] <- vapply(
    mcols(variants)[[alt_col]],
    function(x) gsub(" ", "", paste0(as.character(x), collapse = ",")),
    character(1)
  )

  if (mode == "s") {

    variants <- splitAltVariants(variants, alt_col)

    overlaps <- findOverlaps(G4, variants)
    if (length(overlaps) == 0) {
      message("No overlaps found between variants and G4 regions")
      return(GRanges())
    }
    g4_hits <- queryHits(overlaps)
    variant_hits <- subjectHits(overlaps)

    var_G4 <- variantImpactG4ModeS(G4,
                                   variants,
                                   ref_col = ref_col,
                                   alt_col = alt_col,
                                   g4_hits, variant_hits)

  } else if (mode == "m") {

    checkSampleID(variants, sampleid_col)
    variants <- assignVarIDs(variants, ref_col, sampleid_col)
    variants <- splitAltVariants(variants, alt_col)

    overlaps <- findOverlaps(G4, variants)
    if (length(overlaps) == 0) {
      message("No overlaps found between variants and G4 regions")
      return(GRanges())
    }
    g4_hits <- queryHits(overlaps)
    variant_hits <- subjectHits(overlaps)

    var_G4 <- variantImpactG4ModeM(G4,
                                   variants,
                                   ref_col = ref_col,
                                   alt_col = alt_col,
                                   sampleid_col = sampleid_col,
                                   variantid_col = "G4SNVHunter_Var_ID",
                                   g4_hits, variant_hits)

  }

  metadata(var_G4) <-  metadata(G4)

  message("Processing finished!")

  return(var_G4)
}

#' validateG4VarImpactInputs
#'
#' This function validates the input arguments for the \code{G4VarImpact}
#' function.
#'
#' @param G4 A \code{GRanges} object representing the G4 regions. This object
#' must include a \code{sequence} metadata column containing the G4 sequences.
#' It should come from the output of the \code{G4HunterDetect} function with
#' \code{include_sequences = TRUE}.
#' @param variants A \code{GRanges} object representing the variants. This
#' object must include metadata columns for reference and alternative alleles.
#' @param ref_col A \code{character} string specifying the name of the column
#' in \code{variants} that contains the reference alleles.
#' Default is \code{"ref"}.
#' @param alt_col A \code{character} string specifying the name of the column
#' in \code{variants} that contains the alternative alleles.
#' Default is \code{"alt"}.
#' @param mode A \code{character} string indicating the mode of operation.
#' Set to \code{"s"} to evaluate the impact of individual variants on G4
#' regions one at a time (single-variant mode).
#' Set to \code{"m"} to assess the combined impact of multiple variants that
#' overlap the same G4 region within a sample (multi-variant mode).
#' If using \code{"m"} mode, you must specify \code{sampleid_col}.
#' @param sampleid_col A \code{character} string specifying the name of the
#' column in \code{variants} that contains the sample IDs.
#' Required when \code{mode} is \code{"m"};
#' ignored if \code{mode} is \code{"s"}.
#'
#' @return NULL
#'
#' @noRd
validateG4VarImpactInputs <- function(G4,
                                      variants,
                                      ref_col,
                                      alt_col,
                                      mode,
                                      sampleid_col) {

  if (is.null(G4) || !inherits(G4, "GRanges") || length(G4) == 0) {
    stop("'G4' must be a non-null, non-empty GRanges object.")
  }

  if (!"sequence" %in% colnames(mcols(G4))) {
    stop("'G4' must contain a 'sequence' column. Please use the ",
         "'G4HunterDetect' function with 'include_sequences = TRUE'.")
  }

  if (is.null(variants) ||
      !inherits(variants, "GRanges") ||
      length(variants) == 0) {
    stop("'variants' must be a non-null, non-empty GRanges object.")
  }

  if (is.null(ref_col) || nchar(ref_col) == 0) {
    stop("'ref_col' cannot be NULL or empty")
  }

  if (is.null(alt_col) || nchar(alt_col) == 0) {
    stop("'alt_col' cannot be NULL or empty.")
  }

  if (!mode %in% c("s", "m")) {
    stop("Invalid mode. Choose 's' for single-variant mode or ",
         "'m' for multi-variant mode.")
  }

  if (!alt_col %in% colnames(mcols(variants))) {
    stop("variants must contain a column named '", alt_col, "' in mcols.")
  }

  if (mode == "m" &&
      (is.null(sampleid_col) ||
       !sampleid_col %in% colnames(mcols(variants)))) {
    stop("When mode is 'm', 'sampleid_col' must be specified and present ",
         "in 'variants'.")
  }

  if (mode == "s" && !is.null(sampleid_col)) {
    stop("When mode is 's', 'sampleid_col' should not be specified.")
  }

}

#' assignVarIDs
#'
#' Assign IDs to the \code{GRanges} variants
#'
#' @param gr A \code{GRanges} object containing variants.
#' @param ref_col Name of reference allele column.
#' @param sampleid_col A string specifying the column for sample identifier.
#'
#' @return A \code{GRanges} object with an additional ID metadata column
#' \code{G4SNVHunter_Var_ID}.
#'
#' @noRd
assignVarIDs <- function(gr, ref_col = NULL, sampleid_col = NULL) {

  gr_df <- as.data.frame(gr)

  sample_id_vec <- gr_df[[sampleid_col]]
  ref_vec <- gr_df[[ref_col]]

  gr_df$variant_key <- paste(sample_id_vec, gr_df$seqnames,
                             gr_df$start, ref_vec, sep = "_")

  unique_keys <- unique(gr_df$variant_key)
  key_to_id <- setNames(paste0("ID_", seq_along(unique_keys)), unique_keys)
  gr_df$id <- key_to_id[gr_df$variant_key]

  mcols(gr)$G4SNVHunter_Var_ID <- gr_df$id

  return(gr)
}


#' variantImpactG4ModeS
#'
#' Processes variant impacts on G4 formation in single-variant mode.
#'
#' @param G4  A \code{GRanges} object containing G4s.
#' @param variants  A \code{GRanges} object containing variants.
#' @param ref_col Name of reference allele column.
#' @param alt_col Name of alternate allele column.
#' @param g4_hits An integer vector of row indices in G4 object that overlap
#' with variants.
#' @param variant_hits An integer vector of row indices in variant object that
#' overlap with G4s.
#'
#' @return A \code{GRanges} object with mutated G4 sequences and associated
#' variant information.
#'
#' @import GenomicRanges
#' @import S4Vectors
#'
#' @noRd
variantImpactG4ModeS <- function(G4 = NULL,
                                 variants = NULL,
                                 ref_col = NULL,
                                 alt_col = NULL,
                                 g4_hits, variant_hits) {

  G4_filtered <- G4[g4_hits]

  results <- lapply(seq_along(G4_filtered), function(i) {
    G4_item <- G4_filtered[i]
    affected_var <- variants[variant_hits[i]]
    mut_result <- applyVariantsToG4(G4_item, affected_var, ref_col, alt_col)

    mut_G4_pred <- mutG4MaxScore(
      mut_seq = as.character(mut_result$mut_g4_seq),
      G4_orig_start = start(G4_item),
      threshold = metadata(G4_item)$threshold,
      window_size = metadata(G4_item)$window_size,
      strands = metadata(G4_item)$strands
    )

    GRanges(
      G4_item,
      G4.info = mcols(G4_item),
      variant.info.ranges.start = start(affected_var),
      variant.info.ranges.end = end(affected_var),
      variant.info.ranges.width = width(affected_var),
      variant.info = mcols(affected_var),
      mutated.G4.seq = mut_G4_pred$mut_G4_seq,
      mutated.G4.anno.seq = as.character(mut_result$mut_g4_anno_seq),
      mutated.max_score = mut_G4_pred$max_score,
      score.diff =
        signif(abs(mut_G4_pred$max_score) - abs(G4_item$max_score), 3)
    )

  })

  final_var_G4 <- do.call(c, results)

  return(final_var_G4)
}

#' variantImpactG4ModeM
#'
#' Processes variant impacts on G4 formation in multiple variant mode.
#'
#' @param G4 A \code{GRanges} object containing G4s.
#' @param variants A \code{GRanges} object containing variants.
#' @param ref_col Name of reference allele column.
#' @param alt_col Name of alternate allele column.
#' @param sampleid_col Name of sample ID column.
#' @param variantid_col  Name of variant ID column.
#' @param g4_hits An integer vector of row indices in G4 object that overlap
#' with variants.
#' @param variant_hits An integer vector of row indices in variant object that
#' overlap with G4s.
#'
#' @return A \code{GRanges} object with mutated G4 sequences and associated
#' variant information.
#'
#' @import GenomicRanges
#' @import S4Vectors
#'
#' @noRd
variantImpactG4ModeM <- function(G4 = NULL,
                                 variants = NULL,
                                 ref_col = NULL,
                                 alt_col = NULL,
                                 sampleid_col = NULL,
                                 variantid_col = NULL,
                                 g4_hits, variant_hits) {

  G4_filtered <- G4[unique(g4_hits)]

  results <- lapply(seq_along(G4_filtered), function(i) {
    G4_item <- G4_filtered[i]
    idx <- unique(g4_hits)[i]

    affected_var <- variants[variant_hits[g4_hits == idx]]

    #var_groups <- split(affected_var, mcols(affected_var)[[sampleid_col]])
    var_groups <- processVariants(affected_var,
                                  alt_col,
                                  sampleid_col,
                                  variantid_col)

    sample_results <- lapply(names(var_groups), function(sample) {
      var_s <- var_groups[[sample]]

      mut_result <- applyVariantsToG4(G4_item, var_s, ref_col, alt_col)

      mut_G4_pred <- mutG4MaxScore(
        mut_seq = as.character(mut_result$mut_g4_seq),
        G4_orig_start = start(G4_item),
        threshold = metadata(G4_item)$threshold,
        window_size = metadata(G4_item)$window_size,
        strands = metadata(G4_item)$strands
      )

      #meta_cols <- setdiff(colnames(mcols(var_s)),
      #                     c(sampleid_col, variantid_col))

      meta_cols <- setdiff(colnames(mcols(var_s)), c(variantid_col))

      variant_info <- do.call(DataFrame, c(
        list(variant.info.ranges.start = paste(as.character(start(var_s)),
                                               collapse = ";"),
             variant.info.ranges.end = paste(as.character(end(var_s)),
                                               collapse = ";"),
             variant.info.ranges.width = paste(as.character(width(var_s)),
                                               collapse = ";")),
        setNames(lapply(meta_cols, function(col) paste(mcols(var_s)[[col]],
                                                       collapse = ";")),
                 paste0("variant.info.", meta_cols))
      ))

      tmp_gr <- GRanges(
        G4_item,
        G4.info = mcols(G4_item)#,
        #variant.ids = paste(mcols(var_s)[[variantid_col]], collapse = ";"),
        #sample.ids = sub("_[^_]+$", "", sample)
      )

      mcols(tmp_gr) <- c(mcols(tmp_gr), variant_info, DataFrame(
        mutated.G4.seq = mut_G4_pred$mut_G4_seq,
        mutated.G4.anno.seq = as.character(mut_result$mut_g4_anno_seq),
        mutated.max_score = mut_G4_pred$max_score,
        score.diff =
          signif(abs(mut_G4_pred$max_score) - abs(G4_item$max_score), 3)
      ))

      sample_colname <- paste0("variant.info.", sampleid_col)
      sample_id_all <- mcols(tmp_gr)[[sample_colname]]
      cleaned_sample_id <- paste(unique(
        strsplit(sample_id_all, ";", fixed = TRUE)[[1]]), collapse = ";")
      mcols(tmp_gr)[[sample_colname]] <- cleaned_sample_id

      return(tmp_gr)

    })

    do.call(c, sample_results)
  })

  final_var_G4 <- do.call(c, results)

  return(final_var_G4)
}

#' applyVariantsToG4
#'
#' Introduces variants into a G4 sequence, and generate both the G4 mutated
#' sequence and the annotated sequence indicating the mutations.
#'
#' @param g4 A \code{GRanges} object containing a G4 sequence.
#' @param variants A \code{GRanges} object containing variants overlapping the
#' G4.
#' @param ref_col Name of reference allele column.
#' @param alt_col Name of alternate allele column.
#'
#' @return A \code{list} containing mutated sequence and annotated sequence
#' with mutations in \code{[ref>alt]} format.
#'
#' @import GenomicRanges
#' @import S4Vectors
#'
#' @noRd
applyVariantsToG4 <- function(g4,
                              variants,
                              ref_col = NULL,
                              alt_col = NULL) {

  end(variants) <- start(variants) + nchar(mcols(variants)[[ref_col]]) - 1

  g4_seq <- g4$sequence[1]
  g4_start <- start(g4)[1]
  g4_end <- end(g4)[1]

  variants <- variants[order(start(variants)), ]
  ref_values <- mcols(variants)[[ref_col]]
  #alt_values <- mcols(variants)[[alt_col]]
  alt_values <- gsub(" ", "", mcols(variants)[[alt_col]])
  is_insert <- ref_values == "-"

  rel_starts <- pmax(1, start(variants) - g4_start + 1)
  rel_ends <- pmin(end(variants) - g4_start + 1, nchar(g4_seq))

  cut_points <- sort(unique(c(1, rel_starts, rel_ends + 1, nchar(g4_seq) + 1)))
  seg_starts <- cut_points[-length(cut_points)]
  seg_ends <- cut_points[-1] - 1
  segments <- substring(g4_seq, seg_starts, seg_ends)

  is_var_segment <- rowSums(outer(seg_starts, rel_starts, `>=`) &
                              outer(seg_starts, rel_ends, `<=`)) > 0
  match_idx <- max.col(outer(seg_starts, rel_starts, `>=`) &
                         outer(seg_starts, rel_ends, `<=`),
                       "first")
  match_idx[!is_var_segment] <- NA

  var_idx <- match_idx[is_var_segment]
  #pos_refs <-
  #  str_sub(g4_seq, seg_starts[is_var_segment], seg_starts[is_var_segment])
  pos_refs <-
    substr(g4_seq, seg_starts[is_var_segment], seg_starts[is_var_segment])
  is_insert_segment <- is_insert[var_idx]

  modified_segments <- segments
  modified_segments[is_var_segment] <- ifelse(
    is_insert_segment,
    paste0(pos_refs, alt_values[var_idx]),
    alt_values[var_idx]
  )

  mut_anno_seq <- segments
  mut_anno_seq[is_var_segment] <- ifelse(
    is_insert_segment,
    paste0(pos_refs, "[->", alt_values[var_idx], "]"),
    paste0("[", ref_values[var_idx], ">",
           alt_values[var_idx], "]")
  )

  list(
    mut_g4_seq =
      gsub("-", "", paste0(modified_segments, collapse = ""), fixed = TRUE),
    mut_g4_anno_seq = paste0(mut_anno_seq, collapse = "")
  )
}

#' mutG4MaxScore
#'
#' Recalculates the maximum G4Hunter score for a mutated G4 sequence.
#'
#' @param mut_seq A \code{character} string of the mutated G4 sequence.
#' @param G4_orig_start An \code{integer} specifying the original start
#' position of the G4.
#' @param threshold A \code{numeric} value specifying the G4Hunter score
#' threshold used for prediction.
#' @param window_size An \code{integer} specifying the window size used in
#' G4Hunter prediction.
#' @param strands A \code{character} string indicating the strand(s) to
#' predict on.
#'
#' @return A \code{list} containing the maximum G4Hunter score identified
#' after mutation and the G4-forming sequence corresponding to that score.
#'
#' @import S4Vectors
#'
#' @noRd
mutG4MaxScore <- function(mut_seq = "",
                          G4_orig_start = 25,
                          threshold = 1.5,
                          window_size = 25,
                          strands = "b"
                             ) {

  if (nchar(mut_seq) < window_size) {
    diff_nuc <- window_size - nchar(mut_seq)
    left_pad <- ifelse(ceiling(diff_nuc / 2) > G4_orig_start,
                       G4_orig_start - 1, ceiling(diff_nuc / 2))
    right_pad <- diff_nuc - left_pad
  } else {
    left_pad <- 0
    right_pad <- 0
  }

  tmp_G4 <- G4HunterAnalysis(paste0(strrep("N", left_pad),
                                    as.character(mut_seq),
                                    strrep("N", right_pad)),
                             "CHR", threshold, window_size,
                             include_sequences = T, strands)

  if (length(tmp_G4) == 0) {
    # the G4 score fell below the threshold after the mutation
    # Re-predict at threshold 0 to find the most high-scoring G4s (unstable)
    tmp_G4 <- G4HunterAnalysis(paste0(strrep("N", left_pad),
                                      as.character(mut_seq),
                                      strrep("N", right_pad)),
                               "CHR", 0, window_size,
                               include_sequences = T, strands)
  }

  tmp_pred <- list()
  if (length(tmp_G4) == 0) {
    # No any Gs/Cs can be found
    tmp_pred$max_score <- 0
    tmp_pred$mut_G4_seq <- mut_seq
  } else {
    idx <- which.max(abs(mcols(tmp_G4)$max_score))[1]
    tmp_G4 <- tmp_G4[idx]
    tmp_pred$max_score <- tmp_G4$max_score
    tmp_pred$mut_G4_seq <- tmp_G4$sequence
  }

  return(tmp_pred)
}

#' splitAltVariants
#'
#' Split multi-allelic variants into separate ones.
#'
#' @param gr A \code{GRanges} object containing variant data.
#' @param alt_col A \code{character} string specifying the column name for
#' alternative alleles.
#'
#' @return A \code{GRanges} object, where each record represents a
#' non-multiallelic variant
#'
#' @importFrom dplyr mutate
#' @importFrom tidyr unnest
#' @importFrom data.table :=
#' @importFrom magrittr %>%
#'
#' @noRd
splitAltVariants <- function(gr, alt_col = NULL) {

  df <- as.data.frame(gr)

  if (!any(grepl(",", df[[alt_col]], fixed = TRUE))) {
    return(gr)
  } else {
    df_split <- df %>%
      mutate(!!alt_col := strsplit(as.character(.data[[alt_col]]), ",")) %>%
      unnest(!!alt_col) %>% mutate(!!alt_col := as.character(.data[[alt_col]]))

    metadata_cols <- df_split[, !colnames(df_split) %in%
                                c("seqnames", "start", "end",
                                  "width", "strand")]
    new_gr <- GRanges(
      seqnames = df_split$seqnames,
      ranges = IRanges(start = df_split$start, end = df_split$end),
      strand = df_split$strand,
      metadata_cols)

    return(new_gr)
  }
}

#' checkSampleID
#'
#' Validate if sample IDs are missing.
#'
#' @param gr A \code{GRanges} object containing variant data.
#' @param id_col A \code{character} string specifying the column name for
#' sample ID.
#'
#' @return No return value. Stops execution if invalid IDs
#' (missing or empty) are found.
#'
#' @noRd
checkSampleID <- function(gr, id_col = NULL) {
  id_values <- mcols(gr)[[id_col]]

  bad_idx <- which(is.na(id_values) | id_values == "")

  if (length(bad_idx) > 0) {
    message("Invalid sample ID first found at:\n")
    print(gr[bad_idx[1]])
    stop("Sample ID contains missing (NA) or empty values.")
  }
}

#' processVariants
#'
#' Generate all possible multi-allelic combinations within the G4 region for
#' each sample.
#'
#' @param gr A \code{GRanges} object containing variant data.
#' @param alt_col A \code{character} string specifying the name of the column
#' in \code{gr} that contains the alternative alleles.
#' @param sampleid_col A \code{character} string specifying the name of the
#' column in \code{gr} that contains the sample IDs.
#' @param variantid_col A \code{character} string specifying the name of the
#' column in \code{gr} that contains variant IDs.
#'
#' A \code{GRangesList}, where each element represents a combination of
#' multiallelic variants for a specific sample.
#'
#' @importFrom data.table CJ
#' @importFrom GenomicRanges GRangesList
#'
#' @noRd
processVariants <- function(gr, alt_col, sampleid_col, variantid_col) {

  samples <- split(gr, mcols(gr)[[sampleid_col]])

  result_list <- lapply(names(samples), function(sample_name) {
    sample_data <- samples[[sample_name]]

    variant_options <- split(mcols(sample_data)[[alt_col]],
                             mcols(sample_data)[[variantid_col]])
    variant_options <- lapply(variant_options, unique)

    all_combinations <- do.call(CJ, c(variant_options, unique = TRUE))

    gr_list <- lapply(seq_len(nrow(all_combinations)), function(i) {
      new_gr <-
        GRangesList(lapply(names(variant_options), function(variant_id) {
        #alt_value <- all_combinations[i, ..variant_id]
        alt_value <- all_combinations[i, get(variant_id)]
        pos_data <-
          sample_data[mcols(sample_data)[[variantid_col]] == variant_id][1]
        mcols(pos_data)[[alt_col]] <- as.character(alt_value)
        pos_data
      }))
      unlist(new_gr, use.names = FALSE)
    })

    names(gr_list) <- paste0(sample_name, "_", seq_along(gr_list))

    gr_list
  })

  gr_split <- do.call(c, result_list)

  return(gr_split)
}

test_seq <- paste0("ATCATCATGGGAGGGAGGGAGGGATGGGTATCAGTCGATGCTATAGCATCGACTA",
                  "GGGGCAGGGCATGCAGGGATGGGCATGCTAGCATCAGCTACGATCGGGATCAGCTA",
                  "GCACATCAGATCCCTACCCGCCCTCCCCTGATCGCACTATCTAC")
test_seq <- DNAStringSet(test_seq)
names(test_seq) <- "CHR"
G4 <- G4HunterDetect(test_seq)

variants<- GRanges(
  seqnames = c("CHR", "CHR", "CHR", "CHR", "CHR", "CHR"),
  ranges = IRanges(start = c(5, 33, 12, 11, 124, 389),
                   end = c(10, 33, 15, 11,124, 390)),
  strand = "*",
  rsid = c("ID1", "ID2", "ID3", "ID4", "ID5", "ID6"),
  sid = "ts1",
  ref = c("TCATGG", "A", "AGGG", "G", "C", "AG"),
  alt = c("T", "ACCCGACCATCGCACCCCCC, G", "A,T,GGG", "C", "G", "A"),
  other_info = c("test1", "test2", "test3", "test4", "test5", "test6"),
  other_info2 = c("msg1", "msg2", "msg3", "msg4", "msg5", "msg6")
)

test_that("Input validation works correctly", {

  expect_error(G4VarImpact(NULL, variants),
               "'G4' must be a non-null, non-empty GRanges object.")
  expect_error(G4VarImpact(G4, NULL),
               "'variants' must be a non-null, non-empty GRanges object.")

  expect_error(G4VarImpact(GRanges(), variants),
               "'G4' must be a non-null, non-empty GRanges object.")

  test_g4 <- G4
  test_g4$sequence <- NULL
  expect_error(G4VarImpact(test_g4, variants),
               "'G4' must contain a 'sequence' column")
  expect_error(G4VarImpact(G4, variants, ref_col = NULL),
               "'ref_col' cannot be NULL or empty")
  expect_error(G4VarImpact(G4, variants, ref_col = "ref", alt_col = NULL),
               "'alt_col' cannot be NULL or empty")

  expect_error(G4VarImpact(G4, variants,
                           ref_col = "ref",
                           alt_col = "alt",
                           mode = "x"),
               "Invalid mode. Choose 's' for single-variant mode")

  expect_error(
    G4VarImpact(G4, variants,
                ref_col = "ref",
                alt_col = "alt",
                mode = "m"),
    "When mode is 'm', 'sampleid_col' must be specified"
  )
})

test_that("single-variant mode (s) works correctly", {

  result <- G4VarImpact(G4, variants,
                        ref_col = "ref", alt_col = "alt",
                        mode = "s")

  expect_s4_class(result, "GRanges")
  expect_true(nrow(mcols(result)) > 0)
  expect_true("mutated.max_score" %in% colnames(mcols(result)))
  expect_true("score.diff" %in% colnames(mcols(result)))

  no_overlap_vars <- GRanges(
    seqnames = "chr2",
    ranges = IRanges(start = 1000, width = 1),
    ref = "G",
    alt = "A"
  )
  empty_result <- G4VarImpact(
    G4,
    no_overlap_vars,
    ref_col = "ref",
    alt_col = "alt",
    mode = "s"
  )
  expect_equal(length(empty_result), 0)
})

test_that("multi-variant mode (m) works correctly", {

  result <- G4VarImpact(G4, variants,
                        ref_col = "ref", alt_col = "alt",
                        mode = "m",
                        sampleid_col = "sid")

  expect_s4_class(result, "GRanges")
  expect_true(nrow(mcols(result)) > 0)
})

test_that("Special cases are handled properly", {

  ins_var <- GRanges(
    seqnames = "CHR",
    ranges = IRanges(start = 20, width = 1),
    ref = "-",
    alt = "GGGGGGGGGGGG",
    sampleid = "sample1",
    variantid = "var1"
  )
  expect_s4_class(
    G4VarImpact(
      G4,
      ins_var,
      ref_col = "ref",
      alt_col = "alt",
      mode = "s"
    ),
    "GRanges"
  )

  del_var <- GRanges(
    seqnames = "CHR",
    ranges = IRanges(start = 17, width = 3),
    ref = "GGG",
    alt = "-",
    sampleid = "sample1",
    variantid = "var1"
  )
  expect_s4_class(
    G4VarImpact(
      G4,
      del_var,
      ref_col = "ref",
      alt_col = "alt",
      mode = "m",
      sampleid_col = "sampleid"),
    "GRanges")
})

fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
seq <- loadSequence(seq_path = fa_path)
G4_detected <- G4HunterDetect(seq)

mut <- GRanges(
  seqnames = c("seq1", "seq5"),
  ranges = IRanges(start = c(81, 11), end = c(88, 11)),
  strand = "*",
  ref = c("GGGTAGGG", "A"),
  alt = c("G", "AGGGGGGGGGGGGGGGG")
)

mut_G4 <- G4VarImpact(G4_detected, mut, ref_col = "ref", alt_col = "alt")

test_that("Basic input validation", {

  expect_error(filterVarImpact(GRanges()),
               "Input GRanges object 'mut_G4' is empty.")

  expect_error(filterVarImpact(data.frame()),
               "The input object 'mut_G4' must be a GRanges object.")

  expect_error(filterVarImpact(mut_G4),
               "All thresholds.*are NULL")
})

test_that("Threshold parameter validation", {

  expect_error(filterVarImpact(mut_G4, raw_score_threshold = -1),
               "must be a numeric value between 0 and 4")
  expect_error(filterVarImpact(mut_G4, raw_score_threshold = 5),
               "must be a numeric value between 0 and 4")

  expect_error(filterVarImpact(mut_G4, mut_score_threshold = -1),
               "must be a numeric value between 0 and 4")

  expect_error(filterVarImpact(mut_G4, score_diff_threshold = 1),
               "must be a numeric value between -4 and 0")
  expect_error(filterVarImpact(mut_G4, score_diff_threshold = -5),
               "must be a numeric value between -4 and 0")
})

test_that("Metadata storage", {

  filtered <- filterVarImpact(mut_G4,
                              raw_score_threshold = 2,
                              mut_score_threshold = 1.9,
                              score_diff_threshold = -0.3)

  expect_equal(metadata(filtered)$raw_score_threshold, 2)
  expect_equal(metadata(filtered)$mut_score_threshold, 1.9)
  expect_equal(metadata(filtered)$score_diff_threshold, -0.3)

  filtered_partial <- filterVarImpact(mut_G4, score_diff_threshold = -2)
  expect_null(metadata(filtered_partial)$raw_score_threshold)
  expect_equal(metadata(filtered_partial)$score_diff_threshold, -2)

})

#' Single Nucleotide Variant GRanges Object
#'
#' This dataset contains a \code{GRanges} object storing single nucleotide 
#' variants (SNVs).
#'
#' @docType data
#' @keywords datasets
#' @name snv_gr
#' @format A \code{GRanges} object with 50,000 SNVs.
#' @usage data(snv_gr)
NULL

#' Single Nucleotide Polymorphisms GRanges Object
#'
#' This dataset contains a \code{GRanges} object storing single nucleotide 
#' polymorphisms (SNPs).
#'
#' @docType data
#' @keywords datasets
#' @name snp_gr
#' @format A \code{GRanges} object with 30,000 SNPs.
#' @usage data(snp_gr)
NULL

#' Detect G4 Sequences Using the G4Hunter Algorithm
#'
#' This function detects G4 sequences from a given \code{DNAStringSet} object
#' using the G4Hunter algorithm.
#'
#' @param sequences A \code{DNAStringSet} object containing the input sequences
#' to be analyzed.
#' @param threshold A numeric value specifying the threshold for the G4Hunter
#' score (absolute value). Default is 1.5. It is not recommended to set the
#' threshold below 1.2.
#' @param window_size An integer specifying the window size (bp) used for
#' prediction. Default is 25. Another commonly used window size is 20.
#' However, 25 is generally preferred.
#' @param include_sequences A logical value (\code{TRUE}/\code{FALSE})
#' indicating whether to include the predicted G4 sequences in the output.
#' Default is \code{TRUE}. Setting this parameter to \code{FALSE} can reduce
#' memory usage, which may be beneficial for extremely large genomes. However,
#' we strongly recommend retaining the sequence information in the output, as
#' it is indispensable for subsequent analysis of the impact of variants on
#' G4 formation.
#' @param strands A character string specifying which strand(s) to consider:
#' \code{"b"} for both strands or \code{"p"} for the positive strand only.
#' Default is \code{"b"}.
#'
#' @return A \code{GRanges} object containing the predicted G4 sequences.
#' The \code{GRanges} object includes the following metadata columns:
#' \describe{
#'   \item{\code{score}}{The final G4Hunter score of the predicted G4 sequence
#'   after merging and pruning.}
#'   \item{\code{max_score}}{The maximum G4Hunter score observed across all
#'   sliding windows covering the G4.}
#'   \item{\code{sequence}}{The sequence of the predicted G4 (if
#'   \code{include_sequences = TRUE}).}
#' }
#' Additionally, the following parameters used during detection are stored
#' in the \code{metadata()} of the returned \code{GRanges} object:
#' \describe{
#'   \item{\code{threshold}}{The G4Hunter score threshold used.}
#'   \item{\code{window_size}}{The window size used.}
#'   \item{\code{include_sequences}}{Whether sequences were included.}
#'   \item{\code{strands}}{The strand(s) considered.}
#' }
#' If no G4 sequences are detected, an empty \code{GRanges} object is returned.
#'
#' @seealso \code{\link{loadSequence}} for loading genome sequences into a
#' \code{DNAStringSet} object.
#'
#' @import GenomicRanges
#' @import IRanges
#' @import progress
#' @importFrom Seqinfo seqlevels
#' @export
#'
#' @examples
#'
#' if (!requireNamespace("BiocManager", quietly = TRUE)) {
#'   install.packages("BiocManager")
#' }
#'
#' if (!requireNamespace("Biostrings", quietly = TRUE)) {
#'   BiocManager::install("Biostrings")
#' }
#'
#' library(Biostrings)
#' sequences <- DNAStringSet(c(
#'   "AGTGAATGGGATGGGAGGAGGGACGGGGTAGTACAGCATAGCATG",
#'   "TAGGTAGCTACGACACCCTGCCCTACCCCTACCCCTATCTA"
#' ))
#' names(sequences) <- c("seq1", "seq2")
#'
#' G4s <- G4HunterDetect(sequences, threshold = 1.5, window_size = 25)
#' print(G4s)
#'
#' seq_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
#' G4s <- G4HunterDetect(loadSequence(seq_path = seq_path))
#' print(G4s)
#'
#' seq_path <- system.file("extdata", "seq.txt", package = "G4SNVHunter")
#' G4s <- G4HunterDetect(loadSequence(seq_path = seq_path))
#' print(G4s)
G4HunterDetect <- function(sequences = NULL,
                           threshold = 1.5,
                           window_size = 25,
                           include_sequences = TRUE,
                           strands = "b") {

  validateG4HunterDetectInputs(sequences,
                               threshold,
                               window_size,
                               include_sequences,
                               strands)

  message("Start processing...")

  if (length(sequences) > 2) {
    pb <- progress_bar$new(
      format = "Processing seq No.:current/:total [:bar] :percent in :elapsed",
      total = length(sequences),
      clear = FALSE, width = 60
    )
  } else {
    pb <- NULL
  }

  results <- lapply(seq_along(sequences), function(i) {
    seq <- toupper(as.character(sequences[[i]]))
    chr <- names(sequences)[i]
    if (length(sequences) <= 2) message("Now process: ", chr, ".")
    if (nchar(seq) < window_size) {
      message("The sequence length of ", chr,
              " is less than the window size. ",
              "Skipping this sequence.")
      result <- NULL
    } else {
      result <- G4HunterAnalysis(seq, chr, threshold, window_size,
                                 include_sequences, strands)
    }
    if (!is.null(pb)) {
      pb$tick()
    }

    return(result)
  })

  results <- Filter(Negate(is.null), results)

  if (length(results) == 0) {
    message("No G4 sequences were found.")

    G4_detected <- GRanges()
    metadata(G4_detected)$threshold <- threshold
    metadata(G4_detected)$window_size <- window_size
    metadata(G4_detected)$include_sequences <- include_sequences
    metadata(G4_detected)$strands <- strands

    message("Done!")

    return(G4_detected)
  } else {
    message("Merging predicted G4s...")
    seqlevels_union <- Reduce(union, lapply(results, seqlevels))
    results <- lapply(results, function(gr) {
      seqlevels(gr) <- seqlevels_union
      gr
    })
    G4_detected <- do.call(c, results)

    metadata(G4_detected)$threshold <- threshold
    metadata(G4_detected)$window_size <- window_size
    metadata(G4_detected)$include_sequences <- include_sequences
    metadata(G4_detected)$strands <- strands

    message("Done!")

    return(G4_detected)
  }
}

fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
seq <- loadSequence(seq_path = fa_path)
G4_detected <- G4HunterDetect(seq)

test_that("exportG4 throws error with non-GRanges input", {
  expect_error(exportG4(G4 = "notGRanges", filename = "test.txt"),
               "'G4' must be a GRanges object")
})

test_that("exportG4 throws error with empty GRanges", {
  empty_gr <- GRanges()
  expect_error(exportG4(G4 = empty_gr, filename = "test.txt"),
               "'G4' cannot be NULL")
})

test_that("exportG4 throws error with invalid filename extension", {
  expect_error(exportG4(G4_detected, filename = "res.txt2"),
               "File extension must be one of: txt, csv, xlsx.", fixed = TRUE)
})

test_that("exportG4 throws error with invalid include_metadata", {
  expect_error(exportG4(G4_detected, filename = "res.txt",
                        include_metadata = "hello"),
               "'include_metadata' must be a single logical value.", fixed = TRUE)
})

test_that("exportG4 throws error with invalid revcomp_antisense", {
  expect_error(exportG4(G4_detected, filename = "res.txt",
                        revcomp_antisense = "hello-world"),
               "'revcomp_antisense' must be a single logical value.", fixed = TRUE)
})

seq <- DNAStringSet(
  c("TGTACGGGGTACGGGACTGGGCATGGGCATGCACCCTGCCCTACCCCTCCCATTACGTAGCTAGCTACGAT",
    "AGTGATCGACTAGCTAGCTAGCGACTAGCAGCTAGCGCCCTACGCCCTGACCCCTACGCCCCATACTAG",
    "GATCGATCGATCAGCGTAGCTAGCATCGATCGATCAGCATGCAG"))
names(seq) <- c("seq1", "seq2", "seq3")
G4_gr <- G4HunterDetect(seq)

snp_gr <- GRanges(seqnames = c("seq1", "seq2", "seq1"),
                  ranges = IRanges(start = c(20, 39, 7), width = 1),
                  strand = "*")
mcols(snp_gr)$REF <- c("G", "C", "G")
mcols(snp_gr)$ALT <- c("C", "G", "A")
mcols(snp_gr)$snp_id <- c("snp1", "snp2", "snp3")
mcols(snp_gr)$sampleid <- c("sample1", "sample2", "sample1")

test_that("test SNVImpactG4 function", {

  result <- SNVImpactG4(G4_gr, snp_gr, alt_col = "ALT", mode = "s")

  expect_s4_class(result, "GRanges")
  expect_equal(length(result), 3)
  expect_true("mut.G4.seq" %in% colnames(mcols(result)))
  expect_true("score.diff" %in% colnames(mcols(result)))

  result <- SNVImpactG4(G4_gr, snp_gr, alt_col = "ALT", mode = "m",
                        sampleid_col = "sampleid", snvid_col = "snp_id")

  expect_s4_class(result, "GRanges")
  expect_equal(length(result), 2)
  expect_true("snv.ids" %in% colnames(mcols(result)))
  expect_true("sample.ids" %in% colnames(mcols(result)))

})

test_that("test SNVImpactG4 function with errors", {
  expect_error(SNVImpactG4(NULL, NULL),
               "'G4' must be a non-null, non-empty GRanges object.")
  expect_error(SNVImpactG4(NULL, GRanges()),
               "'G4' must be a non-null, non-empty GRanges object.")
  expect_error(SNVImpactG4(NULL, snp_gr),
               "'G4' must be a non-null, non-empty GRanges object.")
  expect_error(SNVImpactG4(GRanges(), NULL),
               "'G4' must be a non-null, non-empty GRanges object.")
  expect_error(SNVImpactG4(G4_gr, NULL),
               "'snvs' must be a non-null, non-empty GRanges object.")
  expect_error(SNVImpactG4(G4_gr, GRanges()),
               "'snvs' must be a non-null, non-empty GRanges object.")
  expect_error(
    SNVImpactG4(G4_gr, snp_gr),
    "snvs must contain a column named 'alt' in mcols.")
  expect_error(
    SNVImpactG4(G4_gr, snp_gr, mode = "x"),
    paste0("Invalid mode. Choose 's' for variant-centric mode ",
           "or 'm' for sample-centric mode."))
  expect_error(
    SNVImpactG4(G4_gr, snp_gr, alt_col = "ALT", sampleid_col = "test"),
    "When mode is 's', 'sampleid_col' should not be specified.")
  expect_error(
    SNVImpactG4(G4_gr, snp_gr, alt_col = "ALT", snvid_col = "test"),
    "When mode is 's', 'snvid_col' should not be specified.")
  expect_error(SNVImpactG4(G4_gr, snp_gr, alt_col = "hello-world"),
               "snvs must contain a column named 'hello-world' in mcols.")
  expect_error(SNVImpactG4(G4_gr, snp_gr, mode = "m",
                           sampleid_col = "test"),
               "snvs must contain a column named 'alt' in mcols.")
  expect_error(
    SNVImpactG4(G4_gr, snp_gr, mode = "m",
                alt_col = "ALT", sampleid_col = "test"),
    "When mode is 'm', 'sampleid_col' must be specified and present in 'snvs'.")
  expect_error(
    SNVImpactG4(G4_gr, snp_gr, mode = "m",
                alt_col = "ALT", sampleid_col = "sampleid"),
    "When mode is 'm', 'snvid_col' must be specified and present in 'snvs'.")
  expect_error(
    SNVImpactG4(G4_gr,
                snp_gr, mode = "m", alt_col = "ALT",
                sampleid_col = "sampleid", snvid_col = "test"),
    "When mode is 'm', 'snvid_col' must be specified and present in 'snvs'.")

})

test_that("test G4HunterDetect function (input)", {

  sequences <- DNAStringSet(c(
    "AGTGAATGGGATGGGAGGAGGGACGGGGTAGTACAGCATGGGTGCGGGATGGGCGTGGGATCGGAGCATG",
    "TAGGTAGCTACGACACCCTGCCCTACCCCTACCCCTATCTCCTAGGCCCTAGCCCTACCCTAGCCCA"
  ))
  names(sequences) <- c("seq1", "seq2")

  result <- G4HunterDetect(sequences)
  expect_s4_class(result, "GRanges")

  expect_true(all(c("score", "max_score", "sequence") %in%
                    names(mcols(result))))

  expect_equal(length(result), 3)

  result <- G4HunterDetect(sequences, include_sequences = FALSE)

  expect_equal(c("score", "max_score"), names(mcols(result)))
  expect_equal(length(result), 3)

  result <- G4HunterDetect(sequences, strand = "p", include_sequences = FALSE)

  expect_equal(c("score", "max_score"), names(mcols(result)))
  expect_equal(length(result), 2)
  expect_equal(as.character(unique(strand(result))), "+")


  result <- G4HunterDetect(sequences, strand = "p", include_sequences = TRUE)

  expect_equal(c("score", "max_score", "sequence"), names(mcols(result)))
  expect_equal(length(result), 2)
  expect_equal(as.character(unique(strand(result))), "+")


  sequences_short <- DNAStringSet(c(
    "AGTGACTGACAATGGGAATGCGGGAGGAGGGACGGGGGTAGTACAGCTAGTCGAC",
    "TAGGTAGCTACGACAC",
    "TAGCTAGCTACGTACGATCGTACGATCGATCGATCAGCTAGCTACGTCATTCAG",
    "TAGCTAGCTACGTACGATCGTACGA",
    "GGGTGGGTGGGTGGTTGGGTGGGTG",
    "GGGTGGGTGGGTGGTTGGGTGGGTGG",
    "GGGTGGGTGGGTGGTTGGGTGGGTGA",
    "GGGTGGGTGGGTGGTTGGGTGGGT",
    "ctcccctacccctaccctaccccct"
  ))
  names(sequences_short) <- paste0("seq", 1:9)

  result <- G4HunterDetect(sequences_short, window_size = 25)

  expect_s4_class(result, "GRanges")
  expect_equal(length(result), 5)

  sequences_short <- DNAStringSet(c(
    "ATGCTAGCTAGT",
    "TAGGTAGCTACGACAC",
    "AGCTAGCTA"
  ))
  names(sequences_short) <- paste0("seq", 1:3)

  result <- G4HunterDetect(sequences_short, window_size = 25)

})

test_that("test G4HunterDetect function with errors", {

  expect_error(G4HunterDetect(sequences = "TGTAGCGAGGGGGGGACTAGGGGGGCATGCGGG"),
               "'sequences' must be a DNAStringSet object.")

  sequences <- DNAStringSet(c(
    "TAGTCGAGGGCATGCGATCAGCTACGATCGATCGATCAGTCAGCTAGCATCGATGGGTGGGTGGGTGGGG"
  ))
  names(sequences) <- c("seq1")

  expect_error(G4HunterDetect(sequences, threshold = 0.5),
               "'threshold' is not reasonable. It should be between 1 and 4.")

  expect_error(G4HunterDetect(sequences, window_size = 9),
               "'window_size' must be an integer between 10 and 50.")

  expect_error(G4HunterDetect(sequences, strands = "all"),
               "'strands' should be 'b' or 'p', but got 'all' instead.")
})

#' Load Small Variant Data from VCF or MAF Files
#'
#' This function loads variant data from either a standard VCF or MAF file and
#' filters for small variants, such as SNVs, INDELs, and DELINs.
#'
#' @param variant_file A \code{character} string specifying the path to the
#' variant file, which can be in either VCF or MAF format.
#' @param file_type A \code{character} string specifying the type of the
#' input file: either \code{"vcf"} or \code{"maf"}. This parameter must be
#' provided. If \code{"maf"} is specified, the following columns must be
#' present in the MAF file: "Chromosome", "Start_Position", "Reference_Allele",
#' and "Tumor_Seq_Allele2".
#' @param keep_vcf_id A \code{logical} value indicating whether to keep the
#' original ID field from the VCF file in the output. Default is \code{TRUE}.
#'
#' @return A \code{GRanges} object containing the variants loaded from the
#' specified VCF or MAF file. For VCF files, only the \code{ID}, \code{REF},
#' and \code{ALT} metadata columns are included in the output. For MAF files,
#' all available MAF columns are retained.
#'
#' @export
#'
#' @examples
#' # load the vcf file, please do not forget to specify the file type
#' vcf_path <- system.file("extdata",
#'                         "example_variants_chr16.vcf",
#'                         package = "G4SNVHunter")
#' variants <- loadVariant(vcf_path, file_type = "vcf")
#' # load the maf file
#' maf_path <- system.file("extdata",
#'                         "example_variants_chr16.maf",
#'                         package = "G4SNVHunter")
#' variants <- loadVariant(maf_path, file_type = "maf")
loadVariant <- function(variant_file,
                        file_type = c("vcf", "maf"),
                        keep_vcf_id = TRUE) {

  validateLoadVariantInputs(variant_file,
                            file_type,
                            keep_vcf_id)

  result <- switch(
    match.arg(file_type),
    vcf = loadVcf(variant_file, keep_vcf_id),
    maf = loadMaf(variant_file)
  )

  return(result)

}

#' validateLoadVariantInputs
#'
#' This function validates the input arguments for the \code{loadVariant}
#' function.
#'
#' @param variant_file A \code{character} string specifying the path to the
#' variant file, which can be in either VCF or MAF format.
#' @param file_type A \code{character} string specifying the type of the
#' input file: either \code{"vcf"} or \code{"maf}. This parameter must be
#' provided. If \code{"maf"} is specified, the following columns must be
#' present in the MAF file: "Chromosome", "Start_Position", "Reference_Allele",
#' and "Tumor_Seq_Allele2".
#' @param keep_vcf_id A \code{logical} value indicating whether to keep the
#' original ID field from the VCF file in the output. Default is \code{TRUE}.
#'
#' @return NULL
#'
#' @importFrom tools file_ext
#'
#' @noRd
validateLoadVariantInputs <- function(variant_file,
                                      file_type,
                                      keep_vcf_id) {

  if (length(file_type) != 1) {
    stop("'file_type' must be a single specified value: ",
         "either 'vcf' or 'maf'.")
  }

  if (!file_type %in% c("vcf", "maf")) {
    stop("Invalid file type: must be 'vcf' or 'maf'.")
  }

  file_ext <- tolower(file_ext(variant_file))
  if (file_ext == "vcf" && file_type == "maf") {
    stop("The file looks like a VCF (based on extension), but you ",
         "specified file_type = 'maf'.\nPlease verify and correct ",
         "either the file_type or the file extension.")
  } else if (file_ext == "maf" && file_type == "vcf") {
    stop("The file looks like a MAF (based on extension), but you ",
         "specified file_type = 'maf'.\nPlease verify and correct ",
         "either the file_type or the file extension.")
  }

  if (!file.exists(variant_file)) {
    stop("File does not exist: ", variant_file, ".")
  }

  if (!is.logical(keep_vcf_id) || length(keep_vcf_id) != 1) {
    stop("keep_vcf_id must be a single logical value.")
  }
}

#' loadVcf
#'
#' This function loads small variants such as SNVs, INDELs, and DELINs from a
#' VCF file. Structural variants and other types of variants will be ignored.
#'
#' @param vcf_file A \code{character} string specifying the path to the
#' VCF file.
#' @param keep_vcf_id A \code{logical} value indicating whether to keep the
#' VCF ID column in the output.
#'
#' @return A \code{GRanges} object containing the filtered small variant data.
#'
#' @importFrom VariantAnnotation readVcf isSNV isIndel isDelins
#' @importFrom VariantAnnotation isSubstitution isTransition
#' @importFrom SummarizedExperiment rowRanges
#' @importFrom S4Vectors mcols
#' @importFrom IRanges IRanges
#' @importFrom GenomicRanges GRanges
#'
#' @noRd
loadVcf <- function(vcf_file, keep_vcf_id) {

  vcf <- readVcf(vcf_file)
  initial_count <- length(vcf)
  if (initial_count == 0) stop("No variants found in the VCF file.")

  type_mat <- cbind(
    SNV = isSNV(vcf, singleAltOnly = FALSE),
    Indel = isIndel(vcf, singleAltOnly = FALSE),
    Delins = isDelins(vcf, singleAltOnly = FALSE),
    Substitution = isSubstitution(vcf, singleAltOnly = FALSE),
    Transition = isTransition(vcf, singleAltOnly = FALSE)
  )
  vcf <- vcf[rowSums(type_mat) > 0, ]
  filtered_count <- initial_count - length(vcf)
  if (length(vcf) == 0) stop("No variants passed the type filtering!")

  gr <- rowRanges(vcf)

  if (any(as.character(gr$REF) == "-") |
      any(unlist(as.character(unlist(gr$ALT))) == "-")) stop(
        "Detected MAF-style variants (with '-' notation) in the VCF file.")

  if (keep_vcf_id) {
    mcols(gr)$ID <- names(gr)
  }

  mcols(gr) <- mcols(gr)[, c(if(keep_vcf_id) "ID", "REF", "ALT")]

  message(sprintf(
    "VCF loading summary:
    - Initial variants: %d
    - Filtered out by type: %d
    - Final variants retained: %d",
    initial_count,
    filtered_count,
    length(gr)
  ))

  return(gr)
}

#' loadMaf
#'
#' This function loads small variants from a MAF file.
#' Structural variants and other types of variants will be ignored.
#'
#' @param maf_file A \code{character} string specifying the path to the
#' MAF file.
#'
#' @return A \code{GRanges} object containing the filtered small variant data.
#'
#' @importFrom data.table fread .SD
#' @importFrom IRanges IRanges
#' @importFrom GenomicRanges GRanges
#'
#' @noRd
loadMaf <- function(maf_file) {

  maf <- fread(maf_file, header = TRUE, stringsAsFactors = FALSE)
  initial_count <- nrow(maf)

  required_cols <- c("Chromosome", "Start_Position",
                     "Reference_Allele", "Tumor_Seq_Allele2")
  missing_cols <- setdiff(required_cols, colnames(maf))
  if (length(missing_cols) > 0) {
    stop("Missing required columns in MAF: ",
         paste(missing_cols, collapse = ", "))
  }

  maf <- maf[
    !is.na(get("Chromosome")) & get("Chromosome") != "" &
      !is.na(get("Start_Position")) & get("Start_Position") != "" &
      !is.na(get("Reference_Allele")) & get("Reference_Allele") != "" &
      !is.na(get("Tumor_Seq_Allele2")) & get("Tumor_Seq_Allele2") != "" &
      as.numeric(get("Start_Position")) == floor(
        as.numeric(get("Start_Position"))) &
      grepl("^[ACGTNacgtn-]+$", get("Reference_Allele")) &
      grepl("^[ACGTNacgtn-]+$", get("Tumor_Seq_Allele2")) &

      !startsWith(get("Reference_Allele"), "<") &
      !startsWith(get("Tumor_Seq_Allele2"), "<")
    , ]

  filtered_count <- initial_count - nrow(maf)

  message(sprintf(
    "MAF loading summary:
    - Initial variants: %d
    - Filtered out by type: %d
    - Final variants retained: %d",
    initial_count,
    filtered_count,
    nrow(maf)
  ))

  gr <- GRanges(
    seqnames = maf$Chromosome,
    ranges = IRanges(
      start = maf$Start_Position,
      end = maf$Start_Position + nchar(maf$Reference_Allele) - 1
    ),
    strand = "*",
    Reference_Allele = maf$Reference_Allele,
    Tumor_Seq_Allele2 = maf$Tumor_Seq_Allele2,
    maf[, .SD, .SDcols = setdiff(colnames(maf), required_cols)]
  )

  return(gr)
}

vcf_file <- system.file("extdata",
                        "example_variants_chr16.vcf",
                        package = "G4SNVHunter")

maf_file <- system.file("extdata",
                        "example_variants_chr16.maf",
                        package = "G4SNVHunter")

test_that("loadVariant correctly loads VCF files", {

  result <- loadVariant(vcf_file, "vcf")
  expect_s4_class(result, "GRanges")
  expect_equal(length(result), 5000)
  expect_true(all(c("ID", "REF", "ALT") %in% names(mcols(result))))

  result_no_id <- loadVariant(vcf_file, "vcf", keep_vcf_id = FALSE)
  expect_false("ID" %in% names(mcols(result_no_id)))
})

test_that("loadVariant correctly loads MAF files", {

  result <- loadVariant(maf_file, "maf")
  expect_s4_class(result, "GRanges")
  expect_equal(length(result), 13)
  expect_true(all(c("Reference_Allele", "Tumor_Seq_Allele2") %in%
                    names(mcols(result))))
})

test_that("loadVariant validates inputs correctly", {

  expect_error(loadVariant("test.txt", "invalid_type"),
               "Invalid file type: must be 'vcf' or 'maf'")

  expect_error(loadVariant(vcf_file, "maf"),
               "The file looks like a VCF")
  expect_error(loadVariant(maf_file, "vcf"),
               "The file looks like a MAF")

  expect_error(loadVariant("nonexistent.txt", "vcf"),
               "File does not exist")

  expect_error(loadVariant(vcf_file, "vcf", keep_vcf_id = "not_logical"),
               "keep_vcf_id must be a single logical value")
})

#' Check the validity of SNVs
#'
#' This function is deprecated and will be removed in a future version.
#'
#' This function checks whether the user-provided SNVs are single
#' nucleotide substitutions.
#'
#' @param snv_gr A \code{GRanges} object containing SNV data.
#' @param mode A character string specifying the checks to be performed.
#' \code{w} for checking if all widths (w) are 1,
#' \code{r} for checking if all ref (r) values are A, T, C, or G,
#' \code{a} for checking if all alt (a) values are A, T, C, or G.
#' @param ref_col Column name for the ref bases in \code{snv_gr}.
#' Default is NULL.
#' @param alt_col Column name for the alt bases in \code{snv_gr}.
#' Default is NULL.
#'
#' @return A logical value indicating whether the user-provided SNVs
#' passed all checks.
#'
#' @export
#'
#' @examples
#'
#' if (!requireNamespace("BiocManager", quietly = TRUE)) {
#'   install.packages("BiocManager")
#' }
#'
#' if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
#'   BiocManager::install("GenomicRanges")
#' }
#'
#' library(GenomicRanges)
#' gr1 <- GRanges("chr1", IRanges(start = 100, width = 1))
#' # check width ('w')
#' checkSNV(gr1, mode = "w")
#'
#' gr2 <- GRanges(
#'   seqnames = Rle("seq1"),
#'   ranges = IRanges(c(100, 200, 300), width = 1),
#'   ref = c("A", "C", "G"),
#'   alt = c("T", "T", "A")
#' )
#'
#' # check width ('w'), ref ('r'), and alt ('a')
#' checkSNV(gr2, mode = "wra", ref_col = "ref", alt_col = "alt")
#' # check width ('w') and alt ('a')
#' checkSNV(gr2, mode = "wa", alt_col = "alt")
#'
#' gr3 <- GRanges("chr1", IRanges(start = 100, width = 10))
#' # widths should be all one
#' checkSNV(gr3, mode = "w")
#'
#' gr4 <- GRanges(
#'   seqnames = Rle("seq1"),
#'   ranges = IRanges(start = 100, width = 1),
#'   ref = "AG",
#'   alt = "T"
#' )
#'
#' # ref should be all one
#' checkSNV(gr4, mode = "wr", ref_col = "ref")
#' @section Deprecated:
#' This function is no longer supported.
checkSNV <- function(snv_gr = NULL,
                     mode = "wra",
                     ref_col = NULL,
                     alt_col = NULL) {

  .Deprecated(msg = "checkSNV is deprecated.")

  if (!inherits(snv_gr, "GRanges")) {
    stop("'snv_gr' must be a GRanges object")
  }

  if (length(snv_gr) == 0) {
    stop("'snv_gr' is empty. Please provide a non-empty GRanges object.")
  }

  valid_modes <- c("w", "r", "a")
  if (any(!unlist(strsplit(mode, "")) %in% valid_modes)) {
    stop("Invalid 'mode' specified. Use a combination of 'w', 'r', 'a'.")
  }

  if (grepl("r", mode)) {
    if (is.null(ref_col)) {
      stop("Mode 'r' is specified but 'ref_col' is NULL. ",
           "Please provide a valid column name for 'ref'.")
    }
    if (!ref_col %in% colnames(mcols(snv_gr))) {
      stop("The 'ref_col' (", ref_col, ") does not exist in 'snv_gr'.")
    }
    if (all(is.na(mcols(snv_gr)[[ref_col]]))) {
      stop("The 'ref_col' (", ref_col, ") in 'snv_gr' contains only NAs.")
    }
  }

  if (grepl("a", mode)) {
    if (is.null(alt_col)) {
      stop("Mode 'a' is specified but 'alt_col' is NULL. ",
           "Please provide a valid column name for 'alt'.")
    }
    if (!alt_col %in% colnames(mcols(snv_gr))) {
      stop("The 'alt_col' (", alt_col, ") does not exist in 'snv_gr'.")
    }
    if (all(is.na(mcols(snv_gr)[[alt_col]]))) {
      stop("The 'alt_col' (", alt_col, ") in 'snv_gr' contains only NAs.")
    }
  }

  is_valid_nucleotide <- function(x) {
    if (is.null(x) || length(x) == 0) {
      return(FALSE)
    } else {
      return(all(x %in% c("A", "T", "C", "G")))
    }
  }

  check_results <- logical(3)

  if (grepl("w", mode)) {
    check_results[1] <- all(width(snv_gr) == 1)
    if (!check_results[1]) {
      message("Some SNVs do not have a width of 1.")
    }
  } else {
    check_results[1] <- TRUE
  }

  if (grepl("r", mode)) {
    ref_values <- mcols(snv_gr)[[ref_col]]
    check_results[2] <- is_valid_nucleotide(ref_values)
    if (!check_results[2]) {
      message("Invalid nucleotides found in ref column. Only (A, T, C, G) ",
              "are supported.")
    }
  } else {
    check_results[2] <- TRUE
  }

  if (grepl("a", mode)) {
    alt_values <- mcols(snv_gr)[[alt_col]]
    check_results[3] <- is_valid_nucleotide(alt_values)
    if (!check_results[3]) {
      message("Invalid nucleotides found in alt column. Only (A, T, C, G) ",
              "are supported.")
    }
  } else {
    check_results[3] <- TRUE
  }

  return(all(check_results))
}

test_that("test checkSNV function", {

  gr1 <- GRanges(
    seqnames = Rle("chr1"),
    ranges = IRanges(start = 100, width = 1),
    ref = "A",
    alt = "T"
  )
  expect_true(checkSNV(gr1, mode = "wra", ref_col = "ref", alt_col = "alt"))

  gr2 <- GRanges(
    seqnames = Rle("chr1"),
    ranges = IRanges(start = c(100, 200, 300), width = 1),
    ref = c("A", "C", "G"),
    alt = c("T", "T", "A")
  )
  expect_true(checkSNV(gr2, mode = "wra", ref_col = "ref", alt_col = "alt"))

  gr3 <- GRanges(
    seqnames = Rle("chr1"),
    ranges = IRanges(start = 100, width = 10)
  )
  expect_false(checkSNV(gr3, mode = "w"))

  gr4 <- GRanges(
    seqnames = Rle("chr1"),
    ranges = IRanges(start = 100, width = 1),
    ref = "AG",
    alt = "T"
  )
  expect_false(checkSNV(gr4, mode = "r", ref_col = "ref"))
  expect_error(checkSNV(gr4, mode = "r", ref_col = "reference"))

  gr5 <- GRanges(
    seqnames = Rle("chr1"),
    ranges = IRanges(start = 100, width = 1),
    ref = "A",
    alt = "AG"
  )
  expect_false(checkSNV(gr5, mode = "a", alt_col = "alt"))

  gr6 <- GRanges(
    seqnames = Rle("chr1"),
    ranges = IRanges(start = 100, width = 1)
  )
  expect_error(checkSNV(gr6, mode = "r", ref_col = "ref"))
  expect_error(checkSNV(gr6, mode = "a", alt_col = "alt"))

  expect_error(checkSNV(gr6, mode = "xyz", ref_col = "ref"))
  expect_error(checkSNV(gr6, mode = "xyz", ref_col = "reference"))

})

## ----echo=FALSE, results='asis'-----------------------------------------------
cat("### Please refer to the full documentation at :👉 [this link](https://rongxinzh.github.io/G4SNVHunter/G4SNVHunter.html) 👈  :)")

#' Plot Basic Statistics of G4s Detected by the G4HunterDetect Function
#'
#' This function generates a series of plots to visualize basic statistics of
#' G4 sequences predicted by the G4Hunter algorithm.
#' The function produces the following plots:
#' \itemize{
#'   \item Distribution of max scores (absolute values).
#'   \item Distribution of max scores, split by strand.
#'   \item Distribution of scores (absolute values).
#'   \item Distribution of scores, split by strand.
#'   \item Distribution of sequence lengths (with lengths greater than 50 bp
#'   grouped into a single bin).
#'   \item Distribution of sequence lengths, split by strand (with lengths
#'   greater than 50 bp grouped into a single bin).
#' }
#'
#' @param G4 A GRanges object containing G4 sequences. The object must include
#' at least the following metadata columns:
#' \itemize{
#'   \item \code{score}: Numeric vector of scores for G4 sequences.
#'   \item \code{max_score}: Numeric vector of maximum scores for G4 sequences.
#'   \item \code{sequence}: Character vector of G4 sequence strings.
#' }
#' The GRanges object must include \code{strand} information.
#' It should come from the results returned by the
#' \code{G4HunterDetect} function.
#'
#' @param p_colors A vector of colors to be used for the plots. It should
#' contain nine color values corresponding to the different plot elements.
#'
#' @return A combined plot displays all the generated plots arranged in a
#' grid layout.
#'
#' @seealso \code{\link{G4HunterDetect}} for detecting the G4 sequences in a
#' given \code{DNAStringSet} object.
#'
#' @import ggplot2
#' @import cowplot
#' @import GenomicRanges
#' @export
#'
#' @examples
#'
#' seq_path <- system.file("extdata", "seq.txt", package = "G4SNVHunter")
#' G4 <- G4HunterDetect(loadSequence(seq_path = seq_path))
#' plotG4Info(G4)
plotG4Info <- function(G4,
                       p_colors = c("#6B9ECC", "#D91117", "#0E619C",
                                    "#58AC7B", "#D91117", "#0E619C",
                                    "#F39C40", "#D91117", "#0E619C")) {
  if (length(p_colors) != 9) {
    stop("'p_colors' must be a vector of exactly nine colors.")
  }

  if (!inherits(G4, "GRanges")) {
    stop("The input object G4 must be a GRanges object.")
  }

  df <- as.data.frame(G4)

  p1 <- ggplot(df, aes(x = abs(.data$max_score))) +
    geom_histogram(binwidth = 0.1, color = "black", fill = p_colors[1]) +
    labs(title = "Distribution of max_score (Absolute)",
         x = "max_score", y = "Frequency") +
    theme_minimal()

  p2 <- ggplot(df, aes(x = .data$max_score, fill = .data$strand)) +
    geom_histogram(binwidth = 0.1, position = "identity",
                   alpha = 0.6, color = "black") +
    labs(title = "Distribution of max_score (By Strand)",
         x = "max_score", y = "Frequency") +
    theme_minimal() +
    scale_fill_manual(values = c("+" = p_colors[2], "-" = p_colors[3]))

  p3 <- ggplot(df, aes(x = abs(.data$score))) +
    geom_histogram(binwidth = 0.1, color = "black", fill = p_colors[4]) +
    labs(title = "Distribution of score (Absolute)",
         x = "score", y = "Frequency") +
    theme_minimal()

  p4 <- ggplot(df, aes(x = .data$score, fill = .data$strand)) +
    geom_histogram(binwidth = 0.1, position = "identity",
                   alpha = 0.6, color = "black") +
    labs(title = "Distribution of score (By Strand)",
         x = "score", y = "Frequency") +
    theme_minimal() +
    scale_fill_manual(values = c("+" = p_colors[5], "-" = p_colors[6]))

  df$length_group <- cut(df$width,
    breaks = c(seq(0, 50, by = 5), Inf),
    labels = c(paste0(seq(0, 45, by = 5), "-", seq(5, 50, by = 5)), ">50"),
    right = FALSE
  )

  p5 <- ggplot(df, aes(x = .data$length_group)) +
    geom_bar(color = "black", fill = p_colors[7]) +
    labs(title = "Distribution of Sequence Length",
         x = "Length", y = "Frequency") +
    theme_minimal() +
    theme(axis.text.x = element_text(angle = 90, hjust = 1))

  df$strand_length_group <- interaction(df$strand,
                                        df$length_group,
                                        drop = TRUE)

  p6 <- ggplot(df, aes(x = .data$strand_length_group, fill = .data$strand)) +
    geom_bar(position = "dodge", color = "black", alpha = 0.6) +
    labs(title = "Distribution of Sequence Length (By Strand)",
         x = "Length", y = "Frequency") +
    theme_minimal() +
    theme(axis.text.x = element_text(angle = 90, hjust = 1)) +
    scale_fill_manual(values = c("+" = p_colors[8], "-" = p_colors[9]))

  # Combine all plots

  combined_plot <- plot_grid(p1, p2, p3, p4, p5, p6,
                             ncol = 2, labels = "AUTO", align = 'v')

  title <- ggdraw() +
    draw_label(
      "Basic Statistics of G4s Predicted by G4Hunter Algorithm",
      fontface = 'bold',
      x = 0.5, hjust = 0.5
    )

  final_plot <- plot_grid(title, combined_plot,
                          ncol = 1, rel_heights = c(0.1, 1))

  return(final_plot)
}

#' Load Genome Sequences
#'
#' This function loads genomic sequences from multiple sources, including a
#' FASTA file, a text file with sequence identifiers and corresponding
#' sequences, or a data frame object.
#'
#' @param genome_seq A data frame containing sequence identifiers and
#' corresponding sequences.
#' @param seq_path A character string specifying the file path to a FASTA file
#' (suffixed with \code{.fa}, \code{.fna} or \code{.fasta}) or a text file with
#' sequence identifiers and corresponding sequences (file headers should not be
#' provided). Ignored if \code{genome_seq} is provided.
#'
#' @return A \code{DNAStringSet} object containing the genome sequences.
#'
#' @importFrom data.table fread
#' @importFrom stats setNames
#' @importFrom Biostrings DNAStringSet readDNAStringSet
#' @import Seqinfo
#' @export
#'
#' @examples
#'
#' # File path for sequences in fasta format
#' fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
#' seq <- loadSequence(seq_path = fa_path)
#' print(seq)
#'
#' # Another example
#' # Load sequences from data.frame
#' seq_df <- data.frame(
#'   chr = c("seq1", "seq2"),
#'   sequence = c(
#'     paste0(rep("G", 100), collapse = ""),
#'     paste0(rep("A", 100), collapse = "")
#'   )
#' )
#' seq <- loadSequence(genome_seq = seq_df)
#' print(seq)
loadSequence <- function(genome_seq = NULL, seq_path = NULL) {
  if (!is.null(genome_seq) && !is.null(seq_path)) {
    stop("Both 'genome_seq' and 'seq_path' cannot be provided ",
         "simultaneously. Please specify only one.")
  }

  if (!is.null(genome_seq)) {
    if (is.data.frame(genome_seq)) {
      if (ncol(genome_seq) != 2) {
        stop("The data frame must have two columns: 'chr' and 'sequence'.")
      }
      if (nrow(genome_seq) < 1) {
        stop("The data frame must have at least one record.")
      }
      colnames(genome_seq) <- c("chr", "sequence")
      dna_set <- DNAStringSet(setNames(genome_seq$sequence, genome_seq$chr))
      seqlengths(dna_set) <- nchar(genome_seq$sequence)
      return(dna_set)
    } else {
      stop("'genome_seq' must be a data.frame.")
    }
  } else if (!is.null(seq_path)) {
    if (file.exists(seq_path)) {
      if (grepl("\\.fasta$", seq_path) ||
          grepl("\\.fa$", seq_path) ||
          grepl("\\.fna$", seq_path)) {
        dna_set <- readDNAStringSet(seq_path)
        seqlengths(dna_set) <- width(dna_set)
        return(dna_set)
      } else {
        seq_df <- fread(seq_path, header = FALSE)
        if (ncol(seq_df) != 2) {
          stop("The file must have two columns: 'chr' and 'sequence'.")
        }
        colnames(seq_df) <- c("chr", "sequence")
        dna_set <- DNAStringSet(setNames(seq_df$sequence, seq_df$chr))
        seqlengths(dna_set) <- nchar(seq_df$sequence)

        return(dna_set)
      }
    } else {
      stop("The file specified in 'seq_path' does not exist.")
    }
  } else {
    stop("Either 'genome_seq' or 'seq_path' must be provided.")
  }
}

test_that("test loadSequence function (load from data.frame)", {

  seq_df <- data.frame(
    chr = c("seq1", "seq2"),
    sequence = c(
      paste0(rep("G", 100), collapse = ""),
      paste0(rep("A", 200), collapse = "")
    )
  )

  result <- loadSequence(genome_seq = seq_df)
  expect_s4_class(result, "DNAStringSet")
  expect_equal(length(result), 2)
  expect_equal(names(result), c("seq1", "seq2"))
  expect_equal(width(result), c(100, 200))
})

test_that("test loadSequence function (load from fasta)", {

  fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")

  result <- loadSequence(seq_path = fa_path)
  expect_s4_class(result, "DNAStringSet")
  expect_equal(length(result), 5)
  expect_equal(names(result), paste0("seq", 1:5))
  expect_equal(width(result), c(625, 83, 87, 50, 44))

})

test_that("test loadSequence function (load from txt)", {

  txt_path <- system.file("extdata", "seq.txt", package = "G4SNVHunter")

  result <- loadSequence(seq_path = txt_path)
  expect_s4_class(result, "DNAStringSet")
  expect_equal(length(result), 5)
  expect_equal(names(result), paste0("seq", 1:5))
  expect_equal(width(result), c(625, 83, 87, 50, 44))

})

test_that("test loadSequence function with errors", {

  seq_df <- data.frame(
    chr = c("seq1", "seq2"),
    sequence = c(
      paste0(rep("G", 100), collapse = ""),
      paste0(rep("A", 200), collapse = "")
    )
  )

  fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")

  expect_error(
    loadSequence(genome_seq = seq_df,
                 seq_path = fa_path),
    paste0("Both 'genome_seq' and 'seq_path' cannot be provided ",
           "simultaneously. Please specify only one.")
  )

  expect_error(
    loadSequence(genome_seq = list(chr = "seq1", sequence = "G")),
    "'genome_seq' must be a data.frame."
  )

  bad_df <- data.frame(chr = "seq1", sequence = "GGGG", extra = "hello-world")
  expect_error(
    loadSequence(genome_seq = bad_df),
    "The data frame must have two columns: 'chr' and 'sequence'."
  )

  empty_df <- data.frame(chr = character(), sequence = character())
  expect_error(
    loadSequence(genome_seq = empty_df),
    "The data frame must have at least one record."
  )

  expect_error(
    loadSequence(seq_path = "./non_existent_file_quwdgy2d23gdyt23d23hd34.txt"),
    "The file specified in 'seq_path' does not exist."
  )

})

#' Filter Variant Impact G4 Object Based on User-Defined Thresholds
#'
#' This function filters the G4 (\code{GRanges}) object returned by the
#' \code{G4VarImpact} function based on user-defined thresholds for the
#' \code{G4.info.max_score}, \code{mutated.max_score}, and \code{score.diff}
#' columns.
#' Users are not required to specify all three thresholds;
#' however, at least one must be provided.
#'
#' @param mut_G4 A \code{GRanges} object returned by the \code{G4VarImpact}
#' function.
#' @param raw_score_threshold A positive numeric value (no greater than 4)
#' used as the threshold for the absolute value of \code{G4.info.max_score}.
#' G4s with an absolute G4Hunter max_score exceeding this threshold will be
#' retained.
#' If \code{NULL}, this threshold is not applied.
#' @param mut_score_threshold A positive numeric value (no greater than 4)
#' used as the threshold for the absolute value of \code{mutated.max_score}.
#' Mutated G4s with an absolute G4Hunter max_score below this threshold will be
#' retained.
#' If \code{NULL}, this threshold is not applied.
#' @param score_diff_threshold A negative numeric value (no less than -4)
#' used as the threshold for \code{score.diff}. G4s with a decrease in
#' G4Hunter score greater than this threshold after mutation will be retained.
#' If \code{NULL}, this threshold is not applied.
#'
#' @seealso \code{\link{G4VarImpact}} for assessing the impact of variants on
#' G4 formation.
#'
#' @return A filtered \code{GRanges} object, containing only the records that
#' meet the specified threshold criteria.
#' @import GenomicRanges
#' @export
#'
#' @examples
#'
#' if (!requireNamespace("BiocManager", quietly = TRUE)) {
#'   install.packages("BiocManager")
#' }
#'
#' if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
#'   BiocManager::install("GenomicRanges")
#' }
#'
#' if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
#'   BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
#' }
#'
#' library(GenomicRanges)
#' library(BSgenome.Hsapiens.UCSC.hg19)
#'
#' hg19 <- BSgenome.Hsapiens.UCSC.hg19
#' chr21_seq <- DNAStringSet(hg19$chr21)
#' # Chromosome name is needed
#' names(chr21_seq) <- "chr21"
#'
#' G4 <- G4HunterDetect(chr21_seq)
#'
#' data(snp_gr)
#'
#' res_snp <- G4VarImpact(G4, snp_gr, ref_col = "ref", alt_col = "alt")
#' filtered_var_eff <- filterVarImpact(res_snp,
#'   mut_score_threshold = 1.2,
#'   score_diff_threshold = -0.2
#' )
#' print(filtered_var_eff)
filterVarImpact <- function(mut_G4,
                            raw_score_threshold = NULL,
                            mut_score_threshold = NULL,
                            score_diff_threshold = NULL) {

  if (!inherits(mut_G4, "GRanges")) {
    stop("The input object 'mut_G4' must be a GRanges object.")
  }

  if (length(mut_G4) == 0) {
    stop("Input GRanges object 'mut_G4' is empty.")
  }

  if (is.null(raw_score_threshold) &&
      is.null(mut_score_threshold) &&
      is.null(score_diff_threshold)) {
    stop("Error: All thresholds ",
         "('raw_score_threshold', 'mut_score_threshold',",
         " and 'score_diff_threshold') are NULL. ",
         "Please provide at least one.")
  }

  if (!is.null(raw_score_threshold) &&
      (!is.numeric(raw_score_threshold) ||
       raw_score_threshold < 0 ||
       raw_score_threshold > 4)) {
    stop("'raw_score_threshold' must be a numeric value between 0 and 4.")
  }

  if (!is.null(mut_score_threshold) &&
      (!is.numeric(mut_score_threshold) ||
       mut_score_threshold < 0 ||
       mut_score_threshold > 4)) {
    stop("'mut_score_threshold' must be a numeric value between 0 and 4.")
  }

  if (!is.null(score_diff_threshold) &&
      (!is.numeric(score_diff_threshold) ||
       score_diff_threshold >= 0 ||
       score_diff_threshold < -4)) {
    stop("'score_diff_threshold' must be a numeric value between -4 and 0.")
  }

  if (!is.null(raw_score_threshold)) {
    mut_G4 <- mut_G4[abs(mut_G4$G4.info.max_score) >= raw_score_threshold]
  }

  if (!is.null(mut_score_threshold)) {
    mut_G4 <- mut_G4[abs(mut_G4$mutated.max_score) <= mut_score_threshold]
  }

  if (!is.null(score_diff_threshold)) {
    mut_G4 <- mut_G4[mut_G4$score.diff <= score_diff_threshold]
  }

  if (!is.null(raw_score_threshold)) {
    metadata(mut_G4)$raw_score_threshold <- raw_score_threshold
  }

  if (!is.null(mut_score_threshold)) {
    metadata(mut_G4)$mut_score_threshold <- mut_score_threshold
  }

  if (!is.null(score_diff_threshold)) {
    metadata(mut_G4)$score_diff_threshold <- score_diff_threshold
  }

  return(mut_G4)
}

#' Evaluate the Impact of SNVs on G4 Sequences
#'
#' This function is deprecated and will be removed in a future version.
#'
#' This function evaluates the impact of SNVs on G4 formation.
#'
#' @param G4 A \code{GRanges} object representing the G4 regions. This object
#' must include metadata columns for G4 sequence and G4Hunter scores.
#' It should come from the results returned by the
#' \code{G4HunterDetect} function.
#' @param snvs A \code{GRanges} object representing the SNVs. This object must
#' include metadata columns for SNV alternative alleles and, optionally,
#' SNV IDs.
#' @param alt_col A character string specifying the name of the column in
#' \code{snvs} that contains the alternative alleles for the SNVs.
#' The default is \code{"alt"}.
#' @param mode A character string indicating the mode of operation.
#' Set to \code{"s"} to evaluate the impact of individual SNVs on G4 regions
#' one at a time.
#' Set to \code{"m"} to assess the combined impact of multiple SNVs that
#' overlap the same G4 region within a sample.
#' If using \code{"m"} mode, you must specify the \code{sampleid_col} and
#' \code{snvid_col} parameters.
#' @param sampleid_col A character string specifying the name of the column in
#' \code{snvs} that contains the sample IDs.
#' This parameter is required when \code{mode} is \code{"m"}, and is ignored if
#' \code{mode} is \code{"s"}.
#' @param snvid_col A character string specifying the column name in
#' \code{snvs} that contains SNV IDs.
#' Required when \code{mode} is \code{"m"}.
#' Ignored if \code{mode} is \code{"s"}.
#'
#' @return A \code{GRanges} object depending on the \code{mode} parameter:
#' \describe{
#'   \item{Mode "s":}{
#'     For \code{mode = "s"}, the returned \code{GRanges} object includes the
#'     following metadata columns:
#'     \itemize{
#'       \item \code{seqnames} Identifiers for SNVs.
#'       \item \code{ranges} Position of the SNVs.
#'       \item \code{strand} Strand of the SNVs.
#'       \item \code{SNV.info.*} Metadata columns related to each SNV.
#'       \item \code{G4.info.*} Metadata columns from the original G4 object.
#'       \item \code{mut.G4.seq} The mutated G4 sequence after applying the
#'       SNV change.
#'       \item \code{mut.G4.anno.seq} The mutated G4 sequence, with the
#'       mutated bases annotated using square brackets.
#'       \item \code{mut.score} The G4Hunter score of the mutated sequence.
#'       \item \code{score.diff} The difference between the mutated G4Hunter
#'       score and the original G4Hunter score.
#'       The value is calculated as the absolute value of the mutated G4Hunter
#'       score minus the absolute value of the original G4Hunter score.
#'     }
#'   }
#'   \item{Mode "m":}{
#'     For \code{mode = "m"}, the returned \code{GRanges} object includes the
#'     following metadata columns:
#'     \itemize{
#'       \item \code{seqnames} Identifiers for G4 sequences.
#'       \item \code{ranges} Position of the G4 sequences (start and end).
#'       \item \code{strand} Strand of the G4 sequences.
#'       \item \code{G4.info.*} Metadata columns from the original G4 object.
#'       \item \code{snv.ids} Concatenated SNV IDs for all SNVs affecting the
#'       G4 region.
#'       \item \code{sample.ids} A semicolon-separated list of sample IDs
#'       overlapping each G4 region.
#'       \item \code{mut.G4.seq} The mutated G4 sequence after applying the
#'       combined SNV changes.
#'       \item \code{mut.G4.anno.seq} The mutated G4 sequence, with the
#'       mutated bases annotated using square brackets.
#'       \item \code{mut.score} The G4Hunter score of the mutated sequence.
#'       \item \code{score.diff} The difference between the mutated G4Hunter
#'       score and the original G4Hunter score. This value is calculated as the
#'       absolute value of the mutated G4Hunter score minus the absolute value
#'       of the original G4Hunter score.
#'     }
#'   }
#' }
#'
#' @seealso \code{\link{G4HunterDetect}} for detecting the G4 sequences in a
#'                given \code{DNAStringSet} object.
#'          \code{\link{G4HunterScore}} for calculating the G4Hunter scores
#'                for a given sequence.
#'          \code{\link{filterSNVImpact}} for filtering out SNVs with
#'                significant impact.
#'
#' @import GenomicRanges
#' @importFrom S4Vectors queryHits subjectHits mcols
#' @importFrom IRanges start
#' @importFrom Biostrings DNAString DNAStringSet replaceAt BString
#' @export
#'
#' @examples
#'
#' if (!requireNamespace("BiocManager", quietly = TRUE)) {
#'   install.packages("BiocManager")
#' }
#'
#' if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
#'   BiocManager::install("GenomicRanges")
#' }
#'
#' if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
#'   BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
#' }
#'
#' library(GenomicRanges)
#' library(BSgenome.Hsapiens.UCSC.hg19)
#'
#' # Load sequence for chromosome 21 (hg19)
#' hg19 <- BSgenome.Hsapiens.UCSC.hg19
#' chr21_seq <- DNAStringSet(hg19$chr21)
#' # Chromosome name is needed
#' names(chr21_seq) <- "chr21"
#'
#' # Detect G4s in human chromosome 21
#' G4 <- G4HunterDetect(chr21_seq)
#'
#' # 's' mode
#' # Load SNPs
#' data(snp_gr)
#'
#' # Obtain SNPs that overlap with G4 regions and assess their impact on G4.
#' # In variant-centric mode ('s'), evaluating each SNP individually.
#' res_snp <- SNVImpactG4(G4, snp_gr, alt_col = "alt")
#' print(res_snp)
#'
#' # 'm' mode
#' # Load SNVs
#' data(snv_gr)
#'
#' # Obtain SNVs that overlap with G4 regions and assess their impact on G4.
#' # In sample-centric mode ('m'), evaluate the combined impact of SNVs on G4s.
#' # Grouped by the sample IDs specified in 'sampleid_col'.
#' res_snv <- SNVImpactG4(G4, snv_gr,
#'   alt_col = "alt",
#'   mode = "m", sampleid_col = "sampleid", snvid_col = "snv_id"
#' )
#' print(res_snv)
#' @section Deprecated:
#' This function is no longer supported.
#' Use \code{\link{G4VarImpact}} instead.
SNVImpactG4 <- function(G4 = NULL,
                        snvs = NULL,
                        alt_col = "alt",
                        mode = "s",
                        sampleid_col = NULL,
                        snvid_col = NULL) {

  .Deprecated("G4VarImpact")

  if (is.null(G4) || !inherits(G4, "GRanges") || length(G4) == 0) {
    stop("'G4' must be a non-null, non-empty GRanges object.")
  }

  if (!"sequence" %in% colnames(mcols(G4))) {
    stop("'G4' must contain a 'sequence' column. Please use the ",
         "'G4HunterDetect' function with 'include_sequences = TRUE'.")
  }

  if (is.null(snvs) || !inherits(snvs, "GRanges") || length(snvs) == 0) {
    stop("'snvs' must be a non-null, non-empty GRanges object.")
  }

  if (!mode %in% c("s", "m")) {
    stop("Invalid mode. Choose 's' for variant-centric mode or ",
         "'m' for sample-centric mode.")
  }

  if (!alt_col %in% colnames(mcols(snvs))) {
    stop("snvs must contain a column named '", alt_col, "' in mcols.")
  }

  if (mode == "m" &&
      (is.null(sampleid_col) || !sampleid_col %in% colnames(mcols(snvs)))) {
    stop("When mode is 'm', 'sampleid_col' must be specified and present ",
      "in 'snvs'.")
  }

  if (mode == "s" && !is.null(sampleid_col)) {
    stop("When mode is 's', 'sampleid_col' should not be specified.")
  }

  if (mode == "m" &&
      (is.null(snvid_col) || !snvid_col %in% colnames(mcols(snvs)))) {
    stop("When mode is 'm', 'snvid_col' must be specified and present in ",
         "'snvs'.")
  }

  if (mode == "s" && !is.null(snvid_col)) {
    stop("When mode is 's', 'snvid_col' should not be specified.")
  }

  if (!checkSNV(snvs, mode = "wa", alt_col = alt_col)) {
    stop("Your SNV data did not pass the check.")
  }

  message("Start processing...")

  overlaps <- findOverlaps(snvs, G4)

  snv_hits <- queryHits(overlaps)
  g4_hits <- subjectHits(overlaps)

  if (length(snv_hits) == 0) {
    message("No overlaps found between SNVs and G4 regions")
    return(GRanges())
  }

  if (mode == "s") {
    snv_G4 <- GRanges(snvs[snv_hits],
                      SNV.info = data.frame(mcols(snvs[snv_hits])),
                      G4.info = data.frame(G4[g4_hits]))
    snv_pos_in_g4 <- start(snvs[snv_hits]) - start(G4[g4_hits]) + 1

    snv_seq_replaced <- mapply(function(seq, pos, alt) {
      new_seq <- replaceAt(DNAString(seq), IRanges(pos, pos), DNAString(alt))
      new_score <- G4HunterScore(as.character(new_seq))
      annotated_seq <- replaceAt(BString(seq), IRanges(pos, pos),
                                 paste0("[", alt, "]"))
      list(sequence = as.character(new_seq),
           annotated_sequence = as.character(annotated_seq),
           new_score = new_score)
    },
    mcols(G4[g4_hits])$sequence,
    snv_pos_in_g4,
    mcols(snvs[snv_hits])[[alt_col]],
    SIMPLIFY = FALSE)

    snv_seq_replaced_df <- do.call(rbind,
                                   lapply(snv_seq_replaced, function(res) {
      data.frame(sequence = res$sequence,
                 annotated_sequence = res$annotated_sequence,
                 new_score = res$new_score,
                 stringsAsFactors = FALSE)
    }))

    mcols(snv_G4)$mut.G4.seq <- snv_seq_replaced_df$sequence
    mcols(snv_G4)$mut.G4.anno.seq <- snv_seq_replaced_df$annotated_sequence
    mcols(snv_G4)$mut.score <- snv_seq_replaced_df$new_score
    mcols(snv_G4)$score.diff <-
      signif(
        abs(mcols(snv_G4)$mut.score) - abs(mcols(snv_G4)$G4.info.score), 3)
  } else if (mode == "m") {
    combined_data <- data.frame(
      snv_hits = snv_hits,
      g4_hits = g4_hits,
      sample_id = mcols(snvs[snv_hits])[[sampleid_col]],
      stringsAsFactors = FALSE
    )

    sample_g4_groups <- split(combined_data,
                              list(combined_data$sample_id,
                                   combined_data$g4_hits),
                              drop = TRUE)

    result_list <- lapply(sample_g4_groups, function(df) {
      sample_id <- unique(df$sample_id)
      g4_idx <- unique(df$g4_hits)
      base_g4 <- G4[g4_idx]
      base_seq <- mcols(base_g4)$sequence
      snvs_subset <- snvs[df$snv_hits]
      snv_pos_in_g4 <- start(snvs_subset) - start(base_g4) + 1
      snv_alts <- mcols(snvs_subset)[[alt_col]]

      mutated_dna <- replaceAt(DNAString(base_seq),
                               IRanges(snv_pos_in_g4, snv_pos_in_g4),
                               DNAStringSet(snv_alts))
      anno_mutated_dna <- replaceAt(BString(base_seq),
                                    IRanges(snv_pos_in_g4, snv_pos_in_g4),
                                    paste0("[", snv_alts, "]"))
      final_seq <- as.character(mutated_dna)
      final_anno_seq <- as.character(anno_mutated_dna)
      final_score <- G4HunterScore(final_seq)
      score_diff <- signif(abs(final_score) - abs(mcols(base_g4)$score), 3)

      snv_ids <- paste(mcols(snvs_subset)[[snvid_col]], collapse = ";")

      GRanges(base_g4,
        G4.info = mcols(base_g4),
        snv.ids = snv_ids,
        sample.ids = sample_id,
        mut.G4.seq = final_seq,
        mut.G4.anno.seq = final_anno_seq,
        mut.score = final_score,
        score.diff = score_diff
      )
    })

    snv_G4 <- Reduce(function(x, y) c(x, y), result_list)
  }

  message("Processing finished!")

  return(snv_G4)
}

#' Filter SNV Impact GRanges Object Based on User-Defined Thresholds
#'
#' This function is deprecated and will be removed in a future version.
#'
#' This function filters the SNV Impact \code{GRanges} object returned by the
#' \code{SNVImpactG4} function based on user-defined thresholds for the
#' \code{G4.info.score}, \code{mut.score}, and \code{score.diff} parameters.
#' This function filters SNVs that may significantly impair the formation of G4
#' structures using customizable filtering criteria. You are not required to
#' specify all three threshold parameters. However, at least one threshold
#' parameter must be provided.
#'
#' @param gr A \code{GRanges} object returned by the \code{SNVImpactG4}
#' function.
#' @param raw_score_threshold A positive numeric value no greater than 4 used
#' as the threshold for the absolute value of \code{G4.info.score}. G4s with
#' an absolute G4Hunter score exceeding this threshold will be retained.
#' If \code{NULL}, this threshold is not applied.
#' @param mut_score_threshold A positive numeric value no greater than 4 used
#' as the threshold for the absolute value of \code{mut.score}. Mutated G4s
#' with an absolute G4Hunter score below this threshold will be retained.
#' If \code{NULL}, this threshold is not applied.
#' @param score_diff_threshold A negative numeric value no less than -4 used as
#' the threshold for \code{score.diff}. G4s with a decrease in G4Hunter score
#' greater than this threshold after variation will be retained. If
#' \code{NULL}, this threshold is not applied.
#'
#' @seealso \code{\link{SNVImpactG4}} for assessing the impact of SNVs on
#' G4 formation.
#'
#' @return A filtered \code{GRanges} object, containing only the records that
#' meet the specified threshold criteria.
#' @import GenomicRanges
#' @export
#'
#' @examples
#'
#' if (!requireNamespace("BiocManager", quietly = TRUE)) {
#'   install.packages("BiocManager")
#' }
#'
#' if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
#'   BiocManager::install("GenomicRanges")
#' }
#'
#' if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
#'   BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
#' }
#'
#' library(GenomicRanges)
#' library(BSgenome.Hsapiens.UCSC.hg19)
#'
#' hg19 <- BSgenome.Hsapiens.UCSC.hg19
#' chr21_seq <- DNAStringSet(hg19$chr21)
#' # Chromosome name is needed
#' names(chr21_seq) <- "chr21"
#'
#' G4 <- G4HunterDetect(chr21_seq)
#'
#' data(snp_gr)
#'
#' res_snp <- SNVImpactG4(G4, snp_gr, alt_col = "alt")
#' filtered_snv_eff <- filterSNVImpact(res_snp,
#'   mut_score_threshold = 1.2,
#'   score_diff_threshold = -0.2
#' )
#' print(filtered_snv_eff)
#' @section Deprecated:
#' This function is no longer supported.
#' Use \code{\link{filterVarImpact}} instead.
filterSNVImpact <- function(gr,
                            raw_score_threshold = NULL,
                            mut_score_threshold = NULL,
                            score_diff_threshold = NULL) {

  .Deprecated("filterVarImpact")

  if (!inherits(gr, "GRanges")) {
    stop("The input object 'gr' must be a GRanges object.")
  }

  if (length(gr) == 0) {
    stop("Input GRanges object 'gr' is empty.")
  }

  if (is.null(raw_score_threshold) &&
      is.null(mut_score_threshold) &&
      is.null(score_diff_threshold)) {
    stop("Error: All thresholds ",
         "('raw_score_threshold', 'mut_score_threshold',",
         " and 'score_diff_threshold') are NULL. ",
         "Please provide at least one.")
  }

  if (!is.null(raw_score_threshold) &&
      (!is.numeric(raw_score_threshold) ||
       raw_score_threshold < 0 ||
       raw_score_threshold > 4)) {
    stop("'raw_score_threshold' must be a numeric value between 0 and 4.")
  }

  if (!is.null(mut_score_threshold) &&
      (!is.numeric(mut_score_threshold) ||
       mut_score_threshold < 0 ||
       mut_score_threshold > 4)) {
    stop("'mut_score_threshold' must be a numeric value between 0 and 4.")
  }

  if (!is.null(score_diff_threshold) &&
      (!is.numeric(score_diff_threshold) ||
       score_diff_threshold >= 0 ||
       score_diff_threshold < -4)) {
    stop("'score_diff_threshold' must be a numeric value between -4 and 0.")
  }

  if (!is.null(raw_score_threshold)) {
    gr <- gr[abs(gr$G4.info.score) >= raw_score_threshold]
  }

  if (!is.null(mut_score_threshold)) {
    gr <- gr[abs(gr$mut.score) <= mut_score_threshold]
  }

  if (!is.null(score_diff_threshold)) {
    gr <- gr[gr$score.diff <= score_diff_threshold]
  }

  return(gr)
}

hg19 <- BSgenome.Hsapiens.UCSC.hg19
chr21_seq <- DNAStringSet(hg19$chr21)
names(chr21_seq) <- "chr21"
G4 <- G4HunterDetect(chr21_seq)
data(snp_gr)

res_snv <- SNVImpactG4(G4, snp_gr,
                       alt_col = "alt")

test_that("test filterSNVImpact with input type error ", {
  expect_error(filterSNVImpact("test"),
               "The input object 'gr' must be a GRanges object.")
})

test_that("test filterSNVImpact with empty input", {
  empty_gr <- GRanges()
  expect_error(filterSNVImpact(empty_gr),
               "Input GRanges object 'gr' is empty.")
})

test_that("test filterSNVImpact with all thresholds are NULL", {
  expect_error(filterSNVImpact(res_snv),
               paste0("All thresholds \\('raw_score_threshold', ",
                      "'mut_score_threshold', and 'score_diff_threshold'\\) ",
                      "are NULL. Please provide at least one."))
})

test_that("test filterSNVImpact with a negative raw_score_threshold", {
  expect_error(filterSNVImpact(res_snv, raw_score_threshold = -1),
               paste0("'raw_score_threshold' must ",
                      "be a numeric value between 0 and 4."))
})

test_that("test filterSNVImpact with a raw_score_threshold greater than 4", {
  expect_error(filterSNVImpact(res_snv, raw_score_threshold = 4.2),
               paste0("'raw_score_threshold' must ",
                      "be a numeric value between 0 and 4."))
})

test_that("test filterSNVImpact with a negative mut_score_threshold", {
  expect_error(filterSNVImpact(res_snv, mut_score_threshold = -1),
               paste0("'mut_score_threshold' must be a ",
                      "numeric value between 0 and 4."))
})

test_that("test filterSNVImpact with a mut_score_threshold greater than 4", {
  expect_error(filterSNVImpact(res_snv, mut_score_threshold = 4.8),
               paste0("'mut_score_threshold' must be a ",
                      "numeric value between 0 and 4."))
})

test_that("test filterSNVImpact with a positive score_diff_threshold", {
  expect_error(filterSNVImpact(res_snv, score_diff_threshold = 0.2),
               paste0("'score_diff_threshold' must be a ",
                      "numeric value between -4 and 0."))
})

test_that("test filterSNVImpact with a score_diff_threshold less than -4", {
  expect_error(filterSNVImpact(res_snv, score_diff_threshold = -4.2),
               paste0("'score_diff_threshold' must be a ",
                      "numeric value between -4 and 0."))
})

test_that("test filterSNVImpact with a defined raw_score_threshold", {
  filtered_gr <- filterSNVImpact(res_snv, raw_score_threshold = 2.52)
  expect_equal(length(filtered_gr), 1)
  expect_equal(filtered_gr$G4.info.score, 2.59)
})

test_that("test filterSNVImpact with a defined mut_score_threshold", {
  filtered_gr <- filterSNVImpact(res_snv, mut_score_threshold = 1.2)
  expect_equal(length(filtered_gr), 52)
  expect_equal(sum(filtered_gr$mut.score), -4.153)
})

test_that("test filterSNVImpact with a defined score_diff_threshold", {
  filtered_gr <- filterSNVImpact(res_snv, score_diff_threshold = -0.35)
  expect_equal(length(filtered_gr), 36)
})

test_that("test filterSNVImpact with combined thresholds", {
  filtered_gr <- filterSNVImpact(res_snv, raw_score_threshold = 1.5,
                                 mut_score_threshold = 1.1,
                                 score_diff_threshold = -0.3)
  expect_equal(length(filtered_gr), 2)
  expect_equal(filtered_gr$G4.info.score, c(1.54, 1.52))
})

#' Export Predicted G4s to a File
#'
#' This function exports a \code{GRanges} object containing predicted G4s
#' generated by the \code{G4HunterDetect} function from the G4SNVHunter
#' package, to a file in TXT, CSV, or XLSX format.
#'
#' @param G4 A \code{GRanges} object returned by \code{G4HunterDetect}.
#' @param filename A \code{character} string specifying the output file path.
#' The file extension must be one of:
#' \code{.txt}, \code{.csv}, or \code{.xlsx}.
#' @param include_metadata A \code{logical} value.
#' Whether to include global metadata (G4 prediction parameters) in the output.
#' Default is \code{TRUE}.
#' @param revcomp_antisense A \code{logical} value. Whether to
#' reverse-complement sequences on the antisense (negative) strand.
#' Default is \code{TRUE}.
#'
#' @return Invisibly returns a \code{data.frame} object.
#'
#' @export
#'
#' @importFrom Biostrings reverseComplement DNAStringSet
#' @importFrom tools file_ext
#' @importFrom data.table fwrite
#' @importFrom openxlsx createWorkbook addWorksheet writeData saveWorkbook
#' @importFrom S4Vectors metadata
#'
#' @examples
#' fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
#' seq <- loadSequence(seq_path = fa_path)
#' # Predict G4s
#' G4_detected <- G4HunterDetect(seq)
#'
#' out_xlsx <- file.path(tempdir(), "results.xlsx")
#' out_txt  <- file.path(tempdir(), "results.txt")
#' out_csv  <- file.path(tempdir(), "results.csv")
#'
#' exportG4(G4_detected, out_xlsx)
#' exportG4(G4_detected, out_txt, include_metadata = FALSE)
#' exportG4(G4_detected, out_csv, revcomp_antisense = FALSE)
#'
#' unlink(c(out_xlsx, out_txt, out_csv))
exportG4 <- function(G4 = NULL,
                     filename = NULL,
                     include_metadata = TRUE,
                     revcomp_antisense = TRUE) {

  validateExportG4Inputs(G4,
                         filename,
                         include_metadata,
                         revcomp_antisense)

  df <- as.data.frame(G4, stringsAsFactors = FALSE)

  if (revcomp_antisense) {
    if (!"sequence" %in% names(df)) {
      stop("The 'G4' object must contain a 'sequence' column.")
    }

    df$G_rich_sequence <- df$sequence

    antisense_idx <- which(df$strand == "-")
    if (length(antisense_idx) > 0) {
      df$G_rich_sequence[antisense_idx] <- as.character(
        Biostrings::reverseComplement(
          Biostrings::DNAStringSet(df$sequence[antisense_idx])
        )
      )
    } else {
      message("All G4s are on the positive strand; ",
              "reverse complementation was not applied.")
    }
  }

  if (include_metadata) {
    meta_cols <- names(metadata(G4))
    if (length(meta_cols) > 0) {
      metadata_str <- paste(meta_cols,
                            metadata(G4)[meta_cols],
                            sep = ":",
                            collapse = ";")
      metadata_str <- paste0("#", metadata_str)
    } else {
      message("[Note] No metadata columns available in 'G4'.")
    }
  }

  file_ext <- tolower(file_ext(filename))

  if (file_ext %in% c("txt", "csv")) {

    if (exists("metadata_str")) {
      writeLines(metadata_str, filename)
      append_mode <- TRUE
    } else {
      append_mode <- FALSE
    }

    data.table::fwrite(
      df,
      file = filename,
      sep = ifelse(file_ext == "csv", ",", "\t"),
      quote = FALSE,
      col.names = TRUE,
      append = append_mode
    )
  } else {

    wb <- openxlsx::createWorkbook()
    openxlsx::addWorksheet(wb, sheetName = "G4SNVHunter-G4s")

    if (exists("metadata_str")) {
      openxlsx::writeData(wb, sheet = "G4SNVHunter-G4s",
                          metadata_str, startRow = 1, colNames = FALSE)
      data_start_row <- length(metadata_str) + 1
    } else {
      data_start_row <- 1
    }

    openxlsx::writeData(
      wb,
      sheet = "G4SNVHunter-G4s",
      df,
      startRow = data_start_row,
      colNames = TRUE
    )

    openxlsx::saveWorkbook(wb, filename, overwrite = TRUE)
  }

  invisible(df)
}

#' validateExportG4Inputs
#'
#' This function validates the input arguments for the \code{loadVariant}
#' function.
#'
#' @param G4 A \code{GRanges} object returned by \code{G4HunterDetect}.
#' @param filename A \code{character} string specifying the output file path.
#' The file extension must be one of:
#' \code{.txt}, \code{.csv}, or \code{.xlsx}.
#' @param include_metadata A \code{logical} value.
#' Whether to include global metadata (G4 prediction parameters) in the output.
#' Default is \code{TRUE}.
#' @param revcomp_antisense A \code{logical} value. Whether to
#' reverse-complement sequences on the antisense (negative) strand.
#' Default is \code{TRUE}.
#'
#' @return NULL
#' @importFrom tools file_ext
#' @noRd
validateExportG4Inputs <- function(G4,
                                   filename,
                                   include_metadata,
                                   revcomp_antisense) {

  if (!inherits(G4, "GRanges")) {
    stop("'G4' must be a GRanges object.")
  }

  if (length(G4) == 0) {
    stop("'G4' cannot be NULL. Please check your input.")
  }

  if (!is.character(filename) || length(filename) != 1) {
    stop("'filename' should be a single character string.")
  }

  valid_exts <- c("txt", "csv", "xlsx")
  file_ext <- tolower(file_ext(filename))
  if (!file_ext %in% valid_exts) {
    stop(sprintf("File extension must be one of: %s",
                 paste(valid_exts, collapse = ", ")), ".")
  }

  if (!is.logical(include_metadata) || length(include_metadata) != 1) {
    stop("'include_metadata' must be a single logical value.")
  }

  if (!is.logical(revcomp_antisense) || length(revcomp_antisense) != 1) {
    stop("'revcomp_antisense' must be a single logical value.")
  }

}

---
title: "Introduction to G4SNVHunter"
package: G4SNVHunter
output: 
  html_document:
    theme: cerulean
    toc: false
vignette: >
  %\VignetteIndexEntry{Introduction to G4SNVHunter}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}  
---

<img src="images/logo.png" width="230" border="0"/>

```{r, echo=FALSE, results='asis'}
cat("### Please refer to the full documentation at :👉 [this link](https://rongxinzh.github.io/G4SNVHunter/G4SNVHunter.html) 👈  :)")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/G4Hunter.R
\name{G4HunterDetect}
\alias{G4HunterDetect}
\title{Detect G4 Sequences Using the G4Hunter Algorithm}
\usage{
G4HunterDetect(
  sequences = NULL,
  threshold = 1.5,
  window_size = 25,
  include_sequences = TRUE,
  strands = "b"
)
}
\arguments{
\item{sequences}{A \code{DNAStringSet} object containing the input sequences
to be analyzed.}

\item{threshold}{A numeric value specifying the threshold for the G4Hunter
score (absolute value). Default is 1.5. It is not recommended to set the
threshold below 1.2.}

\item{window_size}{An integer specifying the window size (bp) used for
prediction. Default is 25. Another commonly used window size is 20.
However, 25 is generally preferred.}

\item{include_sequences}{A logical value (\code{TRUE}/\code{FALSE})
indicating whether to include the predicted G4 sequences in the output.
Default is \code{TRUE}. Setting this parameter to \code{FALSE} can reduce
memory usage, which may be beneficial for extremely large genomes. However,
we strongly recommend retaining the sequence information in the output, as
it is indispensable for subsequent analysis of the impact of variants on
G4 formation.}

\item{strands}{A character string specifying which strand(s) to consider:
\code{"b"} for both strands or \code{"p"} for the positive strand only.
Default is \code{"b"}.}
}
\value{
A \code{GRanges} object containing the predicted G4 sequences.
The \code{GRanges} object includes the following metadata columns:
\describe{
  \item{\code{score}}{The final G4Hunter score of the predicted G4 sequence
  after merging and pruning.}
  \item{\code{max_score}}{The maximum G4Hunter score observed across all
  sliding windows covering the G4.}
  \item{\code{sequence}}{The sequence of the predicted G4 (if
  \code{include_sequences = TRUE}).}
}
Additionally, the following parameters used during detection are stored
in the \code{metadata()} of the returned \code{GRanges} object:
\describe{
  \item{\code{threshold}}{The G4Hunter score threshold used.}
  \item{\code{window_size}}{The window size used.}
  \item{\code{include_sequences}}{Whether sequences were included.}
  \item{\code{strands}}{The strand(s) considered.}
}
If no G4 sequences are detected, an empty \code{GRanges} object is returned.
}
\description{
This function detects G4 sequences from a given \code{DNAStringSet} object
using the G4Hunter algorithm.
}
\examples{

if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}

if (!requireNamespace("Biostrings", quietly = TRUE)) {
  BiocManager::install("Biostrings")
}

library(Biostrings)
sequences <- DNAStringSet(c(
  "AGTGAATGGGATGGGAGGAGGGACGGGGTAGTACAGCATAGCATG",
  "TAGGTAGCTACGACACCCTGCCCTACCCCTACCCCTATCTA"
))
names(sequences) <- c("seq1", "seq2")

G4s <- G4HunterDetect(sequences, threshold = 1.5, window_size = 25)
print(G4s)

seq_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
G4s <- G4HunterDetect(loadSequence(seq_path = seq_path))
print(G4s)

seq_path <- system.file("extdata", "seq.txt", package = "G4SNVHunter")
G4s <- G4HunterDetect(loadSequence(seq_path = seq_path))
print(G4s)
}
\seealso{
\code{\link{loadSequence}} for loading genome sequences into a
\code{DNAStringSet} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/G4VarImpact.R
\name{G4VarImpact}
\alias{G4VarImpact}
\title{Evaluate the Impact of Variants (SNVs, Indels, MNVs) on G4 Sequences}
\usage{
G4VarImpact(
  G4 = NULL,
  variants = NULL,
  ref_col = NULL,
  alt_col = NULL,
  mode = "s",
  sampleid_col = NULL
)
}
\arguments{
\item{G4}{A \code{GRanges} object representing the G4 regions. This object
must include a \code{sequence} metadata column containing the G4 sequences.
It should come from the output of the \code{G4HunterDetect} function with
\code{include_sequences = TRUE}.}

\item{variants}{A \code{GRanges} object representing the variants. This
object must include metadata columns for reference and alternative alleles.}

\item{ref_col}{A \code{character} string specifying the name of the column
in \code{variants} that contains the reference alleles.
Default is \code{"ref"}.}

\item{alt_col}{A \code{character} string specifying the name of the column
in \code{variants} that contains the alternative alleles.
Default is \code{"alt"}.}

\item{mode}{A \code{character} string indicating the mode of operation.
Set to \code{"s"} to evaluate the impact of individual variants on G4
regions one at a time (single-variant mode).
Set to \code{"m"} to assess the combined impact of multiple variants that
overlap the same G4 region within a sample (multi-variant Mode).
If using \code{"m"} mode, you must specify \code{sampleid_col}.}

\item{sampleid_col}{A \code{character} string specifying the name of the
column in \code{variants} that contains the sample IDs.
Required when \code{mode} is \code{"m"};
ignored if \code{mode} is \code{"s"}.}
}
\value{
A \code{GRanges} object with variant impact results:
\describe{
  \item{Mode "s" (single-variant mode )}{
    For each variant-G4 overlap:
    \itemize{
      \item Original G4 metadata (\code{G4.info.*})
      \item Variant information (\code{variant.info.*})
      \item Mutated sequence (\code{mutated.G4.seq})
      \item Annotated mutation sequence (\code{mutated.G4.anno.seq})
      \item New G4Hunter max_score (\code{mutated.max_score})
      \item Score difference (\code{score.diff})
    }
  }
  \item{Mode "m" (multi-variant mode)}{
    For each sample-G4 combination:
    \itemize{
      \item Original G4 metadata (\code{G4.info.*})
      \item Combined variant information (\code{variant.info.*})
      \item Mutated sequence with all variants incorporated
      (\code{mutated.G4.seq})
      \item Annotated mutation sequence (\code{mutated.G4.anno.seq})
      \item New G4Hunter max_score (\code{mutated.max_score})
      \item Score difference (\code{score.diff})
    }
  }
}
}
\description{
This function evaluates the impact of variants (SNVs, indels, and MNVs) on
G4 formation.
}
\examples{

if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}

if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
  BiocManager::install("GenomicRanges")
}

if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
  BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
}

library(GenomicRanges)
library(BSgenome.Hsapiens.UCSC.hg19)

# Load sequence for chromosome 21 (hg19)
hg19 <- BSgenome.Hsapiens.UCSC.hg19
chr21_seq <- DNAStringSet(hg19$chr21)
# Chromosome name is needed
names(chr21_seq) <- "chr21"

# Detect G4s in human chromosome 21
G4 <- G4HunterDetect(chr21_seq)

# Load variants
data(snv_gr)

# 's' mode; single-variant mode ('s')
# evaluating each variant individually.
res_snv_s <- G4VarImpact(G4,
                         snv_gr,
                         ref_col = "ref",
                         alt_col = "alt")
print(res_snv_s)

# 'm' mode; multi-variant mode ('m')
# evaluating the combined impact of variants on G4s.
# Grouped by the sample IDs specified in the 'sampleid_col' column.
res_snv_m <- G4VarImpact(G4,
                         snv_gr,
                         ref_col = "ref",
                         alt_col = "alt",
                         mode = "m",
                         sampleid_col = "sampleid"
)
print(res_snv_m)
}
\seealso{
\code{\link{G4HunterDetect}} for detecting the G4 sequences in a
               given \code{DNAStringSet} object.
         \code{\link{filterVarImpact}} for filtering out variants with
               significant impact.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snp_gr.R
\docType{data}
\name{snp_gr}
\alias{snp_gr}
\title{Single Nucleotide Polymorphisms GRanges Object}
\format{
A \code{GRanges} object with 30,000 SNPs.
}
\usage{
data(snp_gr)
}
\description{
This dataset contains a \code{GRanges} object storing single nucleotide 
polymorphisms (SNPs).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportG4.R
\name{exportG4}
\alias{exportG4}
\title{Export Predicted G4s to a File}
\usage{
exportG4(
  G4 = NULL,
  filename = NULL,
  include_metadata = TRUE,
  revcomp_antisense = TRUE
)
}
\arguments{
\item{G4}{A \code{GRanges} object returned by \code{G4HunterDetect}.}

\item{filename}{A \code{character} string specifying the output file path.
The file extension must be one of:
\code{.txt}, \code{.csv}, or \code{.xlsx}.}

\item{include_metadata}{A \code{logical} value.
Whether to include global metadata (G4 prediction parameters) in the output.
Default is \code{TRUE}.}

\item{revcomp_antisense}{A \code{logical} value. Whether to
reverse-complement sequences on the antisense (negative) strand.
Default is \code{TRUE}.}
}
\value{
Invisibly returns a \code{data.frame} object.
}
\description{
This function exports a \code{GRanges} object containing predicted G4s
generated by the \code{G4HunterDetect} function from the G4SNVHunter
package, to a file in TXT, CSV, or XLSX format.
}
\examples{
fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
seq <- loadSequence(seq_path = fa_path)
# Predict G4s
G4_detected <- G4HunterDetect(seq)

out_xlsx <- file.path(tempdir(), "results.xlsx")
out_txt  <- file.path(tempdir(), "results.txt")
out_csv  <- file.path(tempdir(), "results.csv")

exportG4(G4_detected, out_xlsx)
exportG4(G4_detected, out_txt, include_metadata = FALSE)
exportG4(G4_detected, out_csv, revcomp_antisense = FALSE)

unlink(c(out_xlsx, out_txt, out_csv))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snv_gr.R
\docType{data}
\name{snv_gr}
\alias{snv_gr}
\title{Single Nucleotide Variant GRanges Object}
\format{
A \code{GRanges} object with 50,000 SNVs.
}
\usage{
data(snv_gr)
}
\description{
This dataset contains a \code{GRanges} object storing single nucleotide 
variants (SNVs).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LoadVariant.R
\name{loadVariant}
\alias{loadVariant}
\title{Load Small Variant Data from VCF or MAF Files}
\usage{
loadVariant(variant_file, file_type = c("vcf", "maf"), keep_vcf_id = TRUE)
}
\arguments{
\item{variant_file}{A \code{character} string specifying the path to the
variant file, which can be in either VCF or MAF format.}

\item{file_type}{A \code{character} string specifying the type of the
input file: either \code{"vcf"} or \code{"maf"}. This parameter must be
provided. If \code{"maf"} is specified, the following columns must be
present in the MAF file: "Chromosome", "Start_Position", "Reference_Allele",
and "Tumor_Seq_Allele2".}

\item{keep_vcf_id}{A \code{logical} value indicating whether to keep the
original ID field from the VCF file in the output. Default is \code{TRUE}.}
}
\value{
A \code{GRanges} object containing the variants loaded from the
specified VCF or MAF file. For VCF files, only the \code{ID}, \code{REF},
and \code{ALT} metadata columns are included in the output. For MAF files,
all available MAF columns are retained.
}
\description{
This function loads variant data from either a standard VCF or MAF file and
filters for small variants, such as SNVs, INDELs, and DELINs.
}
\examples{
# load the vcf file, please do not forget to specify the file type
vcf_path <- system.file("extdata",
                        "example_variants_chr16.vcf",
                        package = "G4SNVHunter")
variants <- loadVariant(vcf_path, file_type = "vcf")
# load the maf file
maf_path <- system.file("extdata",
                        "example_variants_chr16.maf",
                        package = "G4SNVHunter")
variants <- loadVariant(maf_path, file_type = "maf")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportMutG4s.R
\name{exportMutG4}
\alias{exportMutG4}
\title{Export G4s and Associated Variant Information to File}
\usage{
exportMutG4(
  mut_G4 = NULL,
  filename = NULL,
  include_metadata = TRUE,
  revcomp_G4_seq = TRUE,
  revcomp_mutG4_seq = TRUE
)
}
\arguments{
\item{mut_G4}{A \code{GRanges} object returned by \code{filterVarImpact}.
Both the \code{G4.info.sequence} and the \code{mutated.G4.seq} columns
must be included.}

\item{filename}{A \code{character} string specifying the output file path.
The file extension must be one of:
\code{.txt}, \code{.csv}, or \code{.xlsx}.}

\item{include_metadata}{A \code{logical} value.
Whether to include global metadata (e.g., G4 prediction parameters) in the
output.
Default is \code{TRUE}.}

\item{revcomp_G4_seq}{A \code{logical} value. Whether to
reverse-complement G4 sequences on the antisense (negative) strand.
Default is \code{TRUE}.}

\item{revcomp_mutG4_seq}{A \code{logical} value. Whether to
reverse-complement mutated G4 sequences on the antisense (negative) strand.
Default is \code{TRUE}.}
}
\value{
Invisibly returns a \code{data.frame} object.
}
\description{
This function exports a \code{GRanges} object containing predicted G4
regions affected by variants to a TXT, CSV, or XLSX file. The output
includes both the original and mutated G-rich sequences, along with
detailed information about the associated variants.
}
\examples{
library(GenomicRanges)
fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
seq <- loadSequence(seq_path = fa_path)
# Predict G4s
G4_detected <- G4HunterDetect(seq)
# create mut granges object
mut <- data.frame(
  chr = c("seq1", "seq5"),
  pos = c(81, 11),
  ref = c("GGGTAGGG", "A"),
  alt = c("G", "AGGGGGGGGGGGGGGGG")
)

mut <- GRanges(
  seqnames = mut$chr,
  ranges = IRanges(start = mut$pos, end = mut$pos),
  strand = "*",
  ref = mut$ref,
  alt = mut$alt
)

mut_G4 <- G4VarImpact(G4_detected, mut, ref_col = "ref", alt_col = "alt")
filtered_mut_G4 <- filterVarImpact(mut_G4, score_diff_threshold = -0.2)
exportMutG4(filtered_mut_G4, "./result_mut.txt")
exportMutG4(filtered_mut_G4, "./result_mut.xlsx", include_metadata = FALSE)
exportMutG4(filtered_mut_G4, "./result_mut.csv", revcomp_mutG4_seq = FALSE)

# remove all exported files
unlink("./result_mut.txt")
unlink("./result_mut.xlsx")
unlink("./result_mut.csv")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/G4SNVImpact.R
\name{SNVImpactG4}
\alias{SNVImpactG4}
\title{Evaluate the Impact of SNVs on G4 Sequences}
\usage{
SNVImpactG4(
  G4 = NULL,
  snvs = NULL,
  alt_col = "alt",
  mode = "s",
  sampleid_col = NULL,
  snvid_col = NULL
)
}
\arguments{
\item{G4}{A \code{GRanges} object representing the G4 regions. This object
must include metadata columns for G4 sequence and G4Hunter scores.
It should come from the results returned by the
\code{G4HunterDetect} function.}

\item{snvs}{A \code{GRanges} object representing the SNVs. This object must
include metadata columns for SNV alternative alleles and, optionally,
SNV IDs.}

\item{alt_col}{A character string specifying the name of the column in
\code{snvs} that contains the alternative alleles for the SNVs.
The default is \code{"alt"}.}

\item{mode}{A character string indicating the mode of operation.
Set to \code{"s"} to evaluate the impact of individual SNVs on G4 regions
one at a time.
Set to \code{"m"} to assess the combined impact of multiple SNVs that
overlap the same G4 region within a sample.
If using \code{"m"} mode, you must specify the \code{sampleid_col} and
\code{snvid_col} parameters.}

\item{sampleid_col}{A character string specifying the name of the column in
\code{snvs} that contains the sample IDs.
This parameter is required when \code{mode} is \code{"m"}, and is ignored if
\code{mode} is \code{"s"}.}

\item{snvid_col}{A character string specifying the column name in
\code{snvs} that contains SNV IDs.
Required when \code{mode} is \code{"m"}.
Ignored if \code{mode} is \code{"s"}.}
}
\value{
A \code{GRanges} object depending on the \code{mode} parameter:
\describe{
  \item{Mode "s":}{
    For \code{mode = "s"}, the returned \code{GRanges} object includes the
    following metadata columns:
    \itemize{
      \item \code{seqnames} Identifiers for SNVs.
      \item \code{ranges} Position of the SNVs.
      \item \code{strand} Strand of the SNVs.
      \item \code{SNV.info.*} Metadata columns related to each SNV.
      \item \code{G4.info.*} Metadata columns from the original G4 object.
      \item \code{mut.G4.seq} The mutated G4 sequence after applying the
      SNV change.
      \item \code{mut.G4.anno.seq} The mutated G4 sequence, with the
      mutated bases annotated using square brackets.
      \item \code{mut.score} The G4Hunter score of the mutated sequence.
      \item \code{score.diff} The difference between the mutated G4Hunter
      score and the original G4Hunter score.
      The value is calculated as the absolute value of the mutated G4Hunter
      score minus the absolute value of the original G4Hunter score.
    }
  }
  \item{Mode "m":}{
    For \code{mode = "m"}, the returned \code{GRanges} object includes the
    following metadata columns:
    \itemize{
      \item \code{seqnames} Identifiers for G4 sequences.
      \item \code{ranges} Position of the G4 sequences (start and end).
      \item \code{strand} Strand of the G4 sequences.
      \item \code{G4.info.*} Metadata columns from the original G4 object.
      \item \code{snv.ids} Concatenated SNV IDs for all SNVs affecting the
      G4 region.
      \item \code{sample.ids} A semicolon-separated list of sample IDs
      overlapping each G4 region.
      \item \code{mut.G4.seq} The mutated G4 sequence after applying the
      combined SNV changes.
      \item \code{mut.G4.anno.seq} The mutated G4 sequence, with the
      mutated bases annotated using square brackets.
      \item \code{mut.score} The G4Hunter score of the mutated sequence.
      \item \code{score.diff} The difference between the mutated G4Hunter
      score and the original G4Hunter score. This value is calculated as the
      absolute value of the mutated G4Hunter score minus the absolute value
      of the original G4Hunter score.
    }
  }
}
}
\description{
This function is deprecated and will be removed in a future version.
}
\details{
This function evaluates the impact of SNVs on G4 formation.
}
\section{Deprecated}{

This function is no longer supported.
Use \code{\link{G4VarImpact}} instead.
}

\examples{

if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}

if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
  BiocManager::install("GenomicRanges")
}

if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
  BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
}

library(GenomicRanges)
library(BSgenome.Hsapiens.UCSC.hg19)

# Load sequence for chromosome 21 (hg19)
hg19 <- BSgenome.Hsapiens.UCSC.hg19
chr21_seq <- DNAStringSet(hg19$chr21)
# Chromosome name is needed
names(chr21_seq) <- "chr21"

# Detect G4s in human chromosome 21
G4 <- G4HunterDetect(chr21_seq)

# 's' mode
# Load SNPs
data(snp_gr)

# Obtain SNPs that overlap with G4 regions and assess their impact on G4.
# In variant-centric mode ('s'), evaluating each SNP individually.
res_snp <- SNVImpactG4(G4, snp_gr, alt_col = "alt")
print(res_snp)

# 'm' mode
# Load SNVs
data(snv_gr)

# Obtain SNVs that overlap with G4 regions and assess their impact on G4.
# In sample-centric mode ('m'), evaluate the combined impact of SNVs on G4s.
# Grouped by the sample IDs specified in 'sampleid_col'.
res_snv <- SNVImpactG4(G4, snv_gr,
  alt_col = "alt",
  mode = "m", sampleid_col = "sampleid", snvid_col = "snv_id"
)
print(res_snv)
}
\seealso{
\code{\link{G4HunterDetect}} for detecting the G4 sequences in a
               given \code{DNAStringSet} object.
         \code{\link{G4HunterScore}} for calculating the G4Hunter scores
               for a given sequence.
         \code{\link{filterSNVImpact}} for filtering out SNVs with
               significant impact.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkSNV.R
\name{checkSNV}
\alias{checkSNV}
\title{Check the validity of SNVs}
\usage{
checkSNV(snv_gr = NULL, mode = "wra", ref_col = NULL, alt_col = NULL)
}
\arguments{
\item{snv_gr}{A \code{GRanges} object containing SNV data.}

\item{mode}{A character string specifying the checks to be performed.
\code{w} for checking if all widths (w) are 1,
\code{r} for checking if all ref (r) values are A, T, C, or G,
\code{a} for checking if all alt (a) values are A, T, C, or G.}

\item{ref_col}{Column name for the ref bases in \code{snv_gr}.
Default is NULL.}

\item{alt_col}{Column name for the alt bases in \code{snv_gr}.
Default is NULL.}
}
\value{
A logical value indicating whether the user-provided SNVs
passed all checks.
}
\description{
This function is deprecated and will be removed in a future version.
}
\details{
This function checks whether the user-provided SNVs are single
nucleotide substitutions.
}
\section{Deprecated}{

This function is no longer supported.
}

\examples{

if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}

if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
  BiocManager::install("GenomicRanges")
}

library(GenomicRanges)
gr1 <- GRanges("chr1", IRanges(start = 100, width = 1))
# check width ('w')
checkSNV(gr1, mode = "w")

gr2 <- GRanges(
  seqnames = Rle("seq1"),
  ranges = IRanges(c(100, 200, 300), width = 1),
  ref = c("A", "C", "G"),
  alt = c("T", "T", "A")
)

# check width ('w'), ref ('r'), and alt ('a')
checkSNV(gr2, mode = "wra", ref_col = "ref", alt_col = "alt")
# check width ('w') and alt ('a')
checkSNV(gr2, mode = "wa", alt_col = "alt")

gr3 <- GRanges("chr1", IRanges(start = 100, width = 10))
# widths should be all one
checkSNV(gr3, mode = "w")

gr4 <- GRanges(
  seqnames = Rle("seq1"),
  ranges = IRanges(start = 100, width = 1),
  ref = "AG",
  alt = "T"
)

# ref should be all one
checkSNV(gr4, mode = "wr", ref_col = "ref")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/G4SNVImpact.R
\name{filterSNVImpact}
\alias{filterSNVImpact}
\title{Filter SNV Impact GRanges Object Based on User-Defined Thresholds}
\usage{
filterSNVImpact(
  gr,
  raw_score_threshold = NULL,
  mut_score_threshold = NULL,
  score_diff_threshold = NULL
)
}
\arguments{
\item{gr}{A \code{GRanges} object returned by the \code{SNVImpactG4}
function.}

\item{raw_score_threshold}{A positive numeric value no greater than 4 used
as the threshold for the absolute value of \code{G4.info.score}. G4s with
an absolute G4Hunter score exceeding this threshold will be retained.
If \code{NULL}, this threshold is not applied.}

\item{mut_score_threshold}{A positive numeric value no greater than 4 used
as the threshold for the absolute value of \code{mut.score}. Mutated G4s
with an absolute G4Hunter score below this threshold will be retained.
If \code{NULL}, this threshold is not applied.}

\item{score_diff_threshold}{A negative numeric value no less than -4 used as
the threshold for \code{score.diff}. G4s with a decrease in G4Hunter score
greater than this threshold after variation will be retained. If
\code{NULL}, this threshold is not applied.}
}
\value{
A filtered \code{GRanges} object, containing only the records that
meet the specified threshold criteria.
}
\description{
This function is deprecated and will be removed in a future version.
}
\details{
This function filters the SNV Impact \code{GRanges} object returned by the
\code{SNVImpactG4} function based on user-defined thresholds for the
\code{G4.info.score}, \code{mut.score}, and \code{score.diff} parameters.
This function filters SNVs that may significantly impair the formation of G4
structures using customizable filtering criteria. You are not required to
specify all three threshold parameters. However, at least one threshold
parameter must be provided.
}
\section{Deprecated}{

This function is no longer supported.
Use \code{\link{filterVarImpact}} instead.
}

\examples{

if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}

if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
  BiocManager::install("GenomicRanges")
}

if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
  BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
}

library(GenomicRanges)
library(BSgenome.Hsapiens.UCSC.hg19)

hg19 <- BSgenome.Hsapiens.UCSC.hg19
chr21_seq <- DNAStringSet(hg19$chr21)
# Chromosome name is needed
names(chr21_seq) <- "chr21"

G4 <- G4HunterDetect(chr21_seq)

data(snp_gr)

res_snp <- SNVImpactG4(G4, snp_gr, alt_col = "alt")
filtered_snv_eff <- filterSNVImpact(res_snp,
  mut_score_threshold = 1.2,
  score_diff_threshold = -0.2
)
print(filtered_snv_eff)
}
\seealso{
\code{\link{SNVImpactG4}} for assessing the impact of SNVs on
G4 formation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotSNVImpact.R
\name{plotImpactSeq}
\alias{plotImpactSeq}
\title{Visualize the variants in G4 sequence}
\usage{
plotImpactSeq(filtered_gr, ncol = 1)
}
\arguments{
\item{filtered_gr}{A \code{GRanges} object containing sequence data and
G4Hunter scores. The object must have metadata columns named
\code{G4.info.score}, \code{mut.score}, \code{G4.info.sequence}, and
\code{mut.G4.seq}.}

\item{ncol}{An integer specifying the number of columns in the output plot
grid. Default is 1.}
}
\value{
A plot that displays the grid of sequence logos, showing the
differences between the original and mutated sequences.
}
\description{
This function is deprecated and will be removed in a future version.
}
\details{
This function plot sequence logos to visualize sequence variants caused by
SNVs or SNPs, with the location of the variants highlighted by rectangles
and arrows.
}
\section{Deprecated}{

This function is no longer supported.
Use \code{\link{plotImpactedG4}} instead.
}

\examples{

if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
  BiocManager::install("GenomicRanges")
}

library(GenomicRanges)

seq <- data.frame(chr = c("seq1", "seq2"),
                  seq = c("ATTTGGGAGGGAGGGAGGGATGATGAAAATTTTATTTATTTTATTTA",
                          "TTTATACTATTCCCTTACCCTCCCATCCCCATACGGCATCTAGATC"))
seq_gr <- loadSequence(seq)
G4 <- G4HunterDetect(seq_gr)

snv_gr <- GRanges(seqnames = c("seq1", "seq2"),
                  ranges = IRanges(start = c(18, 23), width = 1),
                  ref = c("G", "C"),
                  alt = c("C", "G"))

effect <- SNVImpactG4(G4, snv_gr, alt_col = "alt")
plotImpactSeq(effect, ncol = 2)
}
\seealso{
\code{\link{SNVImpactG4}} for evaluating the impact of SNVs on G4
formation, and \code{\link{filterSNVImpact}} for filtering G4s that are
significantly affected by SNVs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotVarImpact.R
\name{plotImpactedG4}
\alias{plotImpactedG4}
\title{Sequence Logo of the G4 Sequence and Its Mutant Counterpart}
\usage{
plotImpactedG4(mut_G4 = NULL, keep_gstrand = TRUE)
}
\arguments{
\item{mut_G4}{A \code{GRanges} object containing the mutant G4.
If multiple G4s are provided, only the first is plotted.}

\item{keep_gstrand}{Logical. If \code{TRUE} (default), negative
strand (C-rich) G4 will be displayed in its complementary reversed form.}
}
\value{
A ggplot2-based object containing the G4 sequence logo.
}
\description{
This function visualizes and compares the G4 and its mutant counterpart
using sequence logos.
}
\examples{
if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}

if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
  BiocManager::install("GenomicRanges")
}

if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
  BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
}

library(GenomicRanges)
library(BSgenome.Hsapiens.UCSC.hg19)

hg19 <- BSgenome.Hsapiens.UCSC.hg19
chr21_seq <- DNAStringSet(hg19$chr21)
# Chromosome name is needed
names(chr21_seq) <- "chr21"

G4 <- G4HunterDetect(chr21_seq)

# Load SNVs
data(snp_gr)

res_snp <- G4VarImpact(G4, snp_gr, ref_col = "ref", alt_col = "alt")
plotImpactedG4(res_snp[1])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LoadSequence.R
\name{loadSequence}
\alias{loadSequence}
\title{Load Genome Sequences}
\usage{
loadSequence(genome_seq = NULL, seq_path = NULL)
}
\arguments{
\item{genome_seq}{A data frame containing sequence identifiers and
corresponding sequences.}

\item{seq_path}{A character string specifying the file path to a FASTA file
(suffixed with \code{.fa}, \code{.fna} or \code{.fasta}) or a text file with
sequence identifiers and corresponding sequences (file headers should not be
provided). Ignored if \code{genome_seq} is provided.}
}
\value{
A \code{DNAStringSet} object containing the genome sequences.
}
\description{
This function loads genomic sequences from multiple sources, including a
FASTA file, a text file with sequence identifiers and corresponding
sequences, or a data frame object.
}
\examples{

# File path for sequences in fasta format
fa_path <- system.file("extdata", "seq.fa", package = "G4SNVHunter")
seq <- loadSequence(seq_path = fa_path)
print(seq)

# Another example
# Load sequences from data.frame
seq_df <- data.frame(
  chr = c("seq1", "seq2"),
  sequence = c(
    paste0(rep("G", 100), collapse = ""),
    paste0(rep("A", 100), collapse = "")
  )
)
seq <- loadSequence(genome_seq = seq_df)
print(seq)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotSNVImpact.R
\name{plotSNVImpact}
\alias{plotSNVImpact}
\title{Plot the Impact of SNVs on G4Hunter Scores}
\usage{
plotSNVImpact(gr, p_colors = c("#b22d2d", "#6ca4d6", "#2d69b0", "#1f77b4"))
}
\arguments{
\item{gr}{A \code{GRanges} object containing the G4Hunter scores and
associated metadata. The object should include the following columns:
\itemize{
  \item \code{G4.info.score}: Numeric vector of the original G4Hunter
  scores.
  \item \code{mut.score}: Numeric vector of the G4Hunter scores after
  mutation.
  \item \code{score.diff}: Numeric vector of the differences between the
  original and mutant G4Hunter scores.
}}

\item{p_colors}{A vector of four colors used for plotting.}
}
\value{
A combined plot (using \code{plot_grid}) containing two subplots:
        - Density scatter plot comparing original vs mutant G4Hunter scores.
        - Density plot of the G4Hunter score differences.
}
\description{
This function is deprecated and will be removed in a future version.
}
\details{
This function generates two plots for visualizing the impact of SNVs on
G4 formation:
1. A scatter plot with density shading comparing the original G4Hunter score
and the mutant G4Hunter score.
2. A density plot showing the distribution of score changes between the
original and mutant G4 sequences.
}
\section{Deprecated}{

This function is no longer supported.
Use \code{\link{plotVarImpact}} instead.
}

\examples{

if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}

if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
  BiocManager::install("GenomicRanges")
}

if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
  BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
}

library(GenomicRanges)
library(BSgenome.Hsapiens.UCSC.hg19)

hg19 <- BSgenome.Hsapiens.UCSC.hg19
chr21_seq <- DNAStringSet(hg19$chr21)
# Chromosome name is needed
names(chr21_seq) <- "chr21"

G4 <- G4HunterDetect(chr21_seq)

# Load SNVs
data(snp_gr)

res_snp <- SNVImpactG4(G4, snp_gr, alt_col = "alt")
plotSNVImpact(res_snp)
}
\seealso{
\code{\link{SNVImpactG4}} for evaluating the impact of SNVs on G4
formation, and \code{\link{filterSNVImpact}} for filtering G4s that are
significantly affected by SNVs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotG4Info.R
\name{plotG4Info}
\alias{plotG4Info}
\title{Plot Basic Statistics of G4s Detected by the G4HunterDetect Function}
\usage{
plotG4Info(
  G4,
  p_colors = c("#6B9ECC", "#D91117", "#0E619C", "#58AC7B", "#D91117", "#0E619C",
    "#F39C40", "#D91117", "#0E619C")
)
}
\arguments{
\item{G4}{A GRanges object containing G4 sequences. The object must include
at least the following metadata columns:
\itemize{
  \item \code{score}: Numeric vector of scores for G4 sequences.
  \item \code{max_score}: Numeric vector of maximum scores for G4 sequences.
  \item \code{sequence}: Character vector of G4 sequence strings.
}
The GRanges object must include \code{strand} information.
It should come from the results returned by the
\code{G4HunterDetect} function.}

\item{p_colors}{A vector of colors to be used for the plots. It should
contain nine color values corresponding to the different plot elements.}
}
\value{
A combined plot displays all the generated plots arranged in a
grid layout.
}
\description{
This function generates a series of plots to visualize basic statistics of
G4 sequences predicted by the G4Hunter algorithm.
The function produces the following plots:
\itemize{
  \item Distribution of max scores (absolute values).
  \item Distribution of max scores, split by strand.
  \item Distribution of scores (absolute values).
  \item Distribution of scores, split by strand.
  \item Distribution of sequence lengths (with lengths greater than 50 bp
  grouped into a single bin).
  \item Distribution of sequence lengths, split by strand (with lengths
  greater than 50 bp grouped into a single bin).
}
}
\examples{

seq_path <- system.file("extdata", "seq.txt", package = "G4SNVHunter")
G4 <- G4HunterDetect(loadSequence(seq_path = seq_path))
plotG4Info(G4)
}
\seealso{
\code{\link{G4HunterDetect}} for detecting the G4 sequences in a
given \code{DNAStringSet} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotVarImpact.R
\name{plotVarImpact}
\alias{plotVarImpact}
\title{Plot the Impact of Variants on G4 Formation}
\usage{
plotVarImpact(
  gr,
  p_colors = c(fill_2d = "#376597", diagonal = "#b22d2d", density_fill = "#6ca4d6",
    density_line = "#2d69b0", vline = "#1f77b4")
)
}
\arguments{
\item{gr}{A \code{GRanges} object containing G4Hunter scores and
associated metadata.
Required metadata columns include:
\itemize{
  \item \code{G4.info.max_score}: Numeric vector of
  original maximum G4Hunter scores.
  \item \code{mutated.max_score}: Numeric vector of
  maximum G4Hunter scores after mutation.
  \item \code{score.diff}: Numeric vector of score differences.
}}

\item{p_colors}{A character vector of five colors used for plotting.
The vector must include:
\describe{
  \item{\code{"fill_2d"}}{Fill Color for the 2D density plot.}
  \item{\code{"diagonal"}}{Color for the diagonal line in the 2D
  density plot.}
  \item{\code{"density_fill"}}{Fill color for the density plot.}
  \item{\code{"density_line"}}{Line color for the density plot.}
  \item{\code{"vline"}}{Color for the vertical line at zero
  in the density plot.}
}}
}
\value{
A \code{ggplot} object combining two subplots:
  \enumerate{
    \item 2D density plot of the absolute maximum G4Hunter scores
    before and after mutation.
    \item Density plot of the maximum G4Hunter score differences.
  }
}
\description{
This function generates two panels to visualize the impact of variants on
G4 formation:
\enumerate{
  \item A 2D density plot of absolute maximum G4Hunter scores before and
  after mutation.
  \item A density plot showing the distribution of score differences.
}
}
\examples{

if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}

if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
  BiocManager::install("GenomicRanges")
}

if (!requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE)) {
  BiocManager::install("BSgenome.Hsapiens.UCSC.hg19")
}

library(GenomicRanges)
library(BSgenome.Hsapiens.UCSC.hg19)

hg19 <- BSgenome.Hsapiens.UCSC.hg19
chr21_seq <- DNAStringSet(hg19$chr21)
# Chromosome name is needed
names(chr21_seq) <- "chr21"

G4 <- G4HunterDetect(chr21_seq)

# Load SNVs
data(snp_gr)

res_snp <- G4VarImpact(G4, snp_gr, ref_col = "ref", alt_col = "alt")
plotVarImpact(res_snp)
}
\seealso{
\code{\link{G4VarImpact}} for computing the effects of variants
on G4 formation,
and \code{\link{filterVarImpact}} for screening G4s whose formation
propensity may be impaired by variants.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/G4HunterScore.R
\name{G4HunterScore}
\alias{G4HunterScore}
\title{Calculate the G4Hunter Score for a Given Sequence}
\usage{
G4HunterScore(seq = NULL)
}
\arguments{
\item{seq}{A single character string representing the nucleotide sequence.
Must contain only the characters \code{A}, \code{T}, \code{C}, \code{G},
\code{U}, and \code{N}.
The length of the sequence should not be too short (e.g., less than 10 bp).}
}
\value{
A numeric value representing the G4Hunter score for the provided
sequence.
}
\description{
This function calculates the G4Hunter score for a given nucleotide sequence,
which reflects the ability of that sequence to form a G4 structure.
}
\examples{

sequence <- "GGGTAAGGGATGGGTCGGG"
score <- G4HunterScore(sequence)
print(score)
# A negative value indicates that the G4 sequence
# is located on the reverse strand
sequence <- "GGGTAAGGGATGGGTCGGG"
score <- G4HunterScore(sequence)
print(score)
}
\seealso{
\code{\link{G4HunterDetect}} for detecting the G4 sequences in a
given \code{DNAStringSet} object.
}
